@@ -5,8 +5,8 @@
 /*                           GODOT ENGINE                                */
 /*                      https://godotengine.org                          */
 /*************************************************************************/
-/* Copyright (c) 2007-2018 Juan Linietsky, Ariel Manzur.                 */
-/* Copyright (c) 2014-2018 Godot Engine contributors (cf. AUTHORS.md)    */
+/* Copyright (c) 2007-2022 Juan Linietsky, Ariel Manzur.                 */
+/* Copyright (c) 2014-2022 Godot Engine contributors (cf. AUTHORS.md).   */
 /*                                                                       */
 /* Permission is hereby granted, free of charge, to any person obtaining */
 /* a copy of this software and associated documentation files (the       */
@@ -30,7 +30,7 @@
 
 #include "shader_gles3.h"
 
-#include "print_string.h"
+#include "core/print_string.h"
 
 //#define DEBUG_OPENGL
 
@@ -49,7 +49,7 @@
 
 #endif
 
-ShaderGLES3 *ShaderGLES3::active = NULL;
+ShaderGLES3 *ShaderGLES3::active = nullptr;
 
 //#define DEBUG_SHADER
 
@@ -64,7 +64,6 @@ ShaderGLES3 *ShaderGLES3::active = NULL;
 #endif
 
 void ShaderGLES3::bind_uniforms() {
-
 	if (!uniforms_dirty) {
 		return;
 	};
@@ -89,7 +88,6 @@ void ShaderGLES3::bind_uniforms() {
 
 	const Map<uint32_t, CameraMatrix>::Element *C = uniform_cameras.front();
 	while (C) {
-
 		int location = version->uniform_location[C->key()];
 		if (location < 0) {
 			C = C->next();
@@ -104,52 +102,43 @@ void ShaderGLES3::bind_uniforms() {
 }
 
 GLint ShaderGLES3::get_uniform_location(int p_index) const {
-
 	ERR_FAIL_COND_V(!version, -1);
 
 	return version->uniform_location[p_index];
 }
 
 bool ShaderGLES3::bind() {
-
 	if (active != this || !version || new_conditional_version.key != conditional_version.key) {
 		conditional_version = new_conditional_version;
 		version = get_current_version();
 	} else {
-
 		return false;
 	}
 
 	ERR_FAIL_COND_V(!version, false);
 
+	if (!version->ok) { //broken, unable to bind (do not throw error, you saw it before already when it failed compilation).
+		glUseProgram(0);
+		return false;
+	}
+
 	glUseProgram(version->id);
 
 	DEBUG_TEST_ERROR("Use Program");
 
 	active = this;
 	uniforms_dirty = true;
-	/*
- *	why on earth is this code here?
-	for (int i=0;i<texunit_pair_count;i++) {
-
-		glUniform1i(texunit_pairs[i].location, texunit_pairs[i].index);
-		DEBUG_TEST_ERROR("Uniform 1 i");
-	}
-
-*/
 	return true;
 }
 
 void ShaderGLES3::unbind() {
-
-	version = NULL;
+	version = nullptr;
 	glUseProgram(0);
 	uniforms_dirty = true;
-	active = NULL;
+	active = nullptr;
 }
 
 static void _display_error_with_code(const String &p_error, const Vector<const char *> &p_code) {
-
 	int line = 1;
 	String total_code;
 
@@ -160,37 +149,35 @@ static void _display_error_with_code(const String &p_error, const Vector<const c
 	Vector<String> lines = String(total_code).split("\n");
 
 	for (int j = 0; j < lines.size(); j++) {
-
-		print_line(itos(line) + ": " + lines[j]);
+		print_line(vformat("%4d | %s", line, lines[j]));
 		line++;
 	}
 
-	ERR_PRINTS(p_error);
+	ERR_PRINT(p_error);
 }
 
 ShaderGLES3::Version *ShaderGLES3::get_current_version() {
-
 	Version *_v = version_map.getptr(conditional_version);
 
 	if (_v) {
-
 		if (conditional_version.code_version != 0) {
 			CustomCode *cc = custom_code_map.getptr(conditional_version.code_version);
 			ERR_FAIL_COND_V(!cc, _v);
-			if (cc->version == _v->code_version)
+			if (cc->version == _v->code_version) {
 				return _v;
+			}
 		} else {
 			return _v;
 		}
 	}
 
-	if (!_v)
+	if (!_v) {
 		version_map[conditional_version] = Version();
+	}
 
 	Version &v = version_map[conditional_version];
 
 	if (!_v) {
-
 		v.uniform_location = memnew_arr(GLint, uniform_count);
 
 	} else {
@@ -214,20 +201,18 @@ ShaderGLES3::Version *ShaderGLES3::get_current_version() {
 	strings.push_back("#version 300 es\n");
 #endif
 
-	int define_line_ofs = 1;
+#ifdef ANDROID_ENABLED
+	strings.push_back("#define ANDROID_ENABLED\n");
+#endif
 
 	for (int i = 0; i < custom_defines.size(); i++) {
-
 		strings.push_back(custom_defines[i].get_data());
-		define_line_ofs++;
+		strings.push_back("\n");
 	}
 
 	for (int j = 0; j < conditional_count; j++) {
-
 		bool enable = ((1 << j) & conditional_version.version);
 		strings.push_back(enable ? conditional_defines[j] : "");
-		if (enable)
-			define_line_ofs++;
 
 		if (enable) {
 			DEBUG_PRINT(conditional_defines[j]);
@@ -240,30 +225,26 @@ ShaderGLES3::Version *ShaderGLES3::get_current_version() {
 	CharString code_globals;
 	CharString material_string;
 
-	//print_line("code version? "+itos(conditional_version.code_version));
-
-	CustomCode *cc = NULL;
+	CustomCode *cc = nullptr;
 
 	if (conditional_version.code_version > 0) {
 		//do custom code related stuff
 
-		ERR_FAIL_COND_V(!custom_code_map.has(conditional_version.code_version), NULL);
+		ERR_FAIL_COND_V(!custom_code_map.has(conditional_version.code_version), nullptr);
 		cc = &custom_code_map[conditional_version.code_version];
 		v.code_version = cc->version;
-		define_line_ofs += 2;
 	}
 
 	/* CREATE PROGRAM */
 
 	v.id = glCreateProgram();
 
-	ERR_FAIL_COND_V(v.id == 0, NULL);
+	ERR_FAIL_COND_V(v.id == 0, nullptr);
 
 	/* VERTEX SHADER */
 
 	if (cc) {
 		for (int i = 0; i < cc->custom_defines.size(); i++) {
-
 			strings.push_back(cc->custom_defines[i].get_data());
 			DEBUG_PRINT("CD #" + itos(i) + ": " + String(cc->custom_defines[i]));
 		}
@@ -306,13 +287,12 @@ ShaderGLES3::Version *ShaderGLES3::get_current_version() {
 
 	DEBUG_PRINT("\nVertex Code:\n\n" + String(code_string.get_data()));
 	for (int i = 0; i < strings.size(); i++) {
-
 		//print_line("vert strings "+itos(i)+":"+String(strings[i]));
 	}
 #endif
 
 	v.vert_id = glCreateShader(GL_VERTEX_SHADER);
-	glShaderSource(v.vert_id, strings.size(), &strings[0], NULL);
+	glShaderSource(v.vert_id, strings.size(), &strings[0], nullptr);
 	glCompileShader(v.vert_id);
 
 	GLint status;
@@ -324,16 +304,13 @@ ShaderGLES3::Version *ShaderGLES3::get_current_version() {
 		glGetShaderiv(v.vert_id, GL_INFO_LOG_LENGTH, &iloglen);
 
 		if (iloglen < 0) {
-
 			glDeleteShader(v.vert_id);
 			glDeleteProgram(v.id);
 			v.id = 0;
 
 			ERR_PRINT("Vertex shader compilation failed with empty log");
 		} else {
-
 			if (iloglen == 0) {
-
 				iloglen = 4096; //buggy driver (Adreno 220+....)
 			}
 
@@ -351,7 +328,7 @@ ShaderGLES3::Version *ShaderGLES3::get_current_version() {
 			v.id = 0;
 		}
 
-		ERR_FAIL_V(NULL);
+		ERR_FAIL_V(nullptr);
 	}
 
 	//_display_error_with_code("pepo", strings);
@@ -401,13 +378,12 @@ ShaderGLES3::Version *ShaderGLES3::get_current_version() {
 	DEBUG_PRINT("\nFragment Globals:\n\n" + String(code_globals.get_data()));
 	DEBUG_PRINT("\nFragment Code:\n\n" + String(code_string2.get_data()));
 	for (int i = 0; i < strings.size(); i++) {
-
 		//print_line("frag strings "+itos(i)+":"+String(strings[i]));
 	}
 #endif
 
 	v.frag_id = glCreateShader(GL_FRAGMENT_SHADER);
-	glShaderSource(v.frag_id, strings.size(), &strings[0], NULL);
+	glShaderSource(v.frag_id, strings.size(), &strings[0], nullptr);
 	glCompileShader(v.frag_id);
 
 	glGetShaderiv(v.frag_id, GL_COMPILE_STATUS, &status);
@@ -417,16 +393,13 @@ ShaderGLES3::Version *ShaderGLES3::get_current_version() {
 		glGetShaderiv(v.frag_id, GL_INFO_LOG_LENGTH, &iloglen);
 
 		if (iloglen < 0) {
-
 			glDeleteShader(v.frag_id);
 			glDeleteShader(v.vert_id);
 			glDeleteProgram(v.id);
 			v.id = 0;
 			ERR_PRINT("Fragment shader compilation failed with empty log");
 		} else {
-
 			if (iloglen == 0) {
-
 				iloglen = 4096; //buggy driver (Adreno 220+....)
 			}
 
@@ -446,7 +419,7 @@ ShaderGLES3::Version *ShaderGLES3::get_current_version() {
 			v.id = 0;
 		}
 
-		ERR_FAIL_V(NULL);
+		ERR_FAIL_V(nullptr);
 	}
 
 	glAttachShader(v.id, v.frag_id);
@@ -454,7 +427,6 @@ ShaderGLES3::Version *ShaderGLES3::get_current_version() {
 
 	// bind attributes before linking
 	for (int i = 0; i < attribute_pair_count; i++) {
-
 		glBindAttribLocation(v.id, attribute_pairs[i].index, attribute_pairs[i].name);
 	}
 
@@ -463,7 +435,6 @@ ShaderGLES3::Version *ShaderGLES3::get_current_version() {
 	if (feedback_count) {
 		Vector<const char *> feedback;
 		for (int i = 0; i < feedback_count; i++) {
-
 			if (feedbacks[i].conditional == -1 || (1 << feedbacks[i].conditional) & conditional_version.version) {
 				//conditional for this feedback is enabled
 				feedback.push_back(feedbacks[i].name);
@@ -485,16 +456,14 @@ ShaderGLES3::Version *ShaderGLES3::get_current_version() {
 		glGetProgramiv(v.id, GL_INFO_LOG_LENGTH, &iloglen);
 
 		if (iloglen < 0) {
-
 			glDeleteShader(v.frag_id);
 			glDeleteShader(v.vert_id);
 			glDeleteProgram(v.id);
 			v.id = 0;
-			ERR_FAIL_COND_V(iloglen <= 0, NULL);
+			ERR_FAIL_COND_V(iloglen < 0, nullptr);
 		}
 
 		if (iloglen == 0) {
-
 			iloglen = 4096; //buggy driver (Adreno 220+....)
 		}
 
@@ -513,7 +482,7 @@ ShaderGLES3::Version *ShaderGLES3::get_current_version() {
 		glDeleteProgram(v.id);
 		v.id = 0;
 
-		ERR_FAIL_V(NULL);
+		ERR_FAIL_V(nullptr);
 	}
 
 	/* UNIFORMS */
@@ -522,20 +491,17 @@ ShaderGLES3::Version *ShaderGLES3::get_current_version() {
 
 	//print_line("uniforms:  ");
 	for (int j = 0; j < uniform_count; j++) {
-
 		v.uniform_location[j] = glGetUniformLocation(v.id, uniform_names[j]);
 		//print_line("uniform "+String(uniform_names[j])+" location "+itos(v.uniform_location[j]));
 	}
 
 	// set texture uniforms
 	for (int i = 0; i < texunit_pair_count; i++) {
-
 		GLint loc = glGetUniformLocation(v.id, texunit_pairs[i].name);
 		if (loc >= 0) {
 			if (texunit_pairs[i].index < 0) {
 				glUniform1i(loc, max_image_units + texunit_pairs[i].index); //negative, goes down
 			} else {
-
 				glUniform1i(loc, texunit_pairs[i].index);
 			}
 		}
@@ -543,18 +509,16 @@ ShaderGLES3::Version *ShaderGLES3::get_current_version() {
 
 	// assign uniform block bind points
 	for (int i = 0; i < ubo_count; i++) {
-
 		GLint loc = glGetUniformBlockIndex(v.id, ubo_pairs[i].name);
-		if (loc >= 0)
+		if (loc >= 0) {
 			glUniformBlockBinding(v.id, loc, ubo_pairs[i].index);
+		}
 	}
 
 	if (cc) {
-
 		v.texture_uniform_locations.resize(cc->texture_uniforms.size());
 		for (int i = 0; i < cc->texture_uniforms.size(); i++) {
-
-			v.texture_uniform_locations[i] = glGetUniformLocation(v.id, String(cc->texture_uniforms[i]).ascii().get_data());
+			v.texture_uniform_locations.write[i] = glGetUniformLocation(v.id, String(cc->texture_uniforms[i]).ascii().get_data());
 			glUniform1i(v.texture_uniform_locations[i], i + base_material_tex_index);
 		}
 	}
@@ -562,18 +526,19 @@ ShaderGLES3::Version *ShaderGLES3::get_current_version() {
 	glUseProgram(0);
 
 	v.ok = true;
+	if (cc) {
+		cc->versions.insert(conditional_version.version);
+	}
 
 	return &v;
 }
 
 GLint ShaderGLES3::get_uniform_location(const String &p_name) const {
-
 	ERR_FAIL_COND_V(!version, -1);
 	return glGetUniformLocation(version->id, p_name.ascii().get_data());
 }
 
 void ShaderGLES3::setup(const char **p_conditional_defines, int p_conditional_count, const char **p_uniform_names, int p_uniform_count, const AttributePair *p_attribute_pairs, int p_attribute_count, const TexUnitPair *p_texunit_pairs, int p_texunit_pair_count, const UBOPair *p_ubo_pairs, int p_ubo_pair_count, const Feedback *p_feedback, int p_feedback_count, const char *p_vertex_code, const char *p_fragment_code, int p_vertex_code_start, int p_fragment_code_start) {
-
 	ERR_FAIL_COND(version);
 	conditional_version.key = 0;
 	new_conditional_version.key = 0;
@@ -612,7 +577,6 @@ void ShaderGLES3::setup(const char **p_conditional_defines, int p_conditional_co
 			if (cpos == -1) {
 				vertex_code1 = code.ascii();
 			} else {
-
 				vertex_code1 = code.substr(0, cpos).ascii();
 				String code2 = code.substr(cpos + globals_tag.length(), code.length());
 
@@ -620,7 +584,6 @@ void ShaderGLES3::setup(const char **p_conditional_defines, int p_conditional_co
 				if (cpos == -1) {
 					vertex_code2 = code2.ascii();
 				} else {
-
 					vertex_code2 = code2.substr(0, cpos).ascii();
 					vertex_code3 = code2.substr(cpos + code_tag.length(), code2.length()).ascii();
 				}
@@ -646,7 +609,6 @@ void ShaderGLES3::setup(const char **p_conditional_defines, int p_conditional_co
 			if (cpos == -1) {
 				fragment_code1 = code.ascii();
 			} else {
-
 				fragment_code1 = code.substr(0, cpos).ascii();
 				//print_line("CODE1:\n"+String(fragment_code1.get_data()));
 
@@ -656,7 +618,6 @@ void ShaderGLES3::setup(const char **p_conditional_defines, int p_conditional_co
 				if (cpos == -1) {
 					fragment_code2 = code2.ascii();
 				} else {
-
 					fragment_code2 = code2.substr(0, cpos).ascii();
 					//print_line("CODE2:\n"+String(fragment_code2.get_data()));
 
@@ -666,7 +627,6 @@ void ShaderGLES3::setup(const char **p_conditional_defines, int p_conditional_co
 					if (cpos == -1) {
 						fragment_code3 = code3.ascii();
 					} else {
-
 						fragment_code3 = code3.substr(0, cpos).ascii();
 						//print_line("CODE3:\n"+String(fragment_code3.get_data()));
 						fragment_code4 = code3.substr(cpos + code_tag.length(), code3.length()).ascii();
@@ -681,10 +641,8 @@ void ShaderGLES3::setup(const char **p_conditional_defines, int p_conditional_co
 }
 
 void ShaderGLES3::finish() {
-
-	const VersionKey *V = NULL;
+	const VersionKey *V = nullptr;
 	while ((V = version_map.next(V))) {
-
 		Version &v = version_map[*V];
 		glDeleteShader(v.vert_id);
 		glDeleteShader(v.frag_id);
@@ -694,10 +652,8 @@ void ShaderGLES3::finish() {
 }
 
 void ShaderGLES3::clear_caches() {
-
-	const VersionKey *V = NULL;
+	const VersionKey *V = nullptr;
 	while ((V = version_map.next(V))) {
-
 		Version &v = version_map[*V];
 		glDeleteShader(v.vert_id);
 		glDeleteShader(v.frag_id);
@@ -708,20 +664,18 @@ void ShaderGLES3::clear_caches() {
 	version_map.clear();
 
 	custom_code_map.clear();
-	version = NULL;
+	version = nullptr;
 	last_custom_code = 1;
 	uniforms_dirty = true;
 }
 
 uint32_t ShaderGLES3::create_custom_shader() {
-
 	custom_code_map[last_custom_code] = CustomCode();
 	custom_code_map[last_custom_code].version = 1;
 	return last_custom_code++;
 }
 
 void ShaderGLES3::set_custom_shader_code(uint32_t p_code_id, const String &p_vertex, const String &p_vertex_globals, const String &p_fragment, const String &p_light, const String &p_fragment_globals, const String &p_uniforms, const Vector<StringName> &p_texture_uniforms, const Vector<CharString> &p_custom_defines) {
-
 	ERR_FAIL_COND(!custom_code_map.has(p_code_id));
 	CustomCode *cc = &custom_code_map[p_code_id];
 
@@ -737,39 +691,46 @@ void ShaderGLES3::set_custom_shader_code(uint32_t p_code_id, const String &p_ver
 }
 
 void ShaderGLES3::set_custom_shader(uint32_t p_code_id) {
-
 	new_conditional_version.code_version = p_code_id;
 }
 
 void ShaderGLES3::free_custom_shader(uint32_t p_code_id) {
-
-	/*  if (! custom_code_map.has( p_code_id )) {
-        print_line("no code id "+itos(p_code_id));
-    } else {
-        print_line("freed code id "+itos(p_code_id));
-
-    }*/
-
 	ERR_FAIL_COND(!custom_code_map.has(p_code_id));
-	if (conditional_version.code_version == p_code_id)
-		conditional_version.code_version = 0; //bye
+	if (conditional_version.code_version == p_code_id) {
+		conditional_version.code_version = 0; //do not keep using a version that is going away
+		unbind();
+	}
+
+	VersionKey key;
+	key.code_version = p_code_id;
+	for (Set<uint32_t>::Element *E = custom_code_map[p_code_id].versions.front(); E; E = E->next()) {
+		key.version = E->get();
+		ERR_CONTINUE(!version_map.has(key));
+		Version &v = version_map[key];
+
+		glDeleteShader(v.vert_id);
+		glDeleteShader(v.frag_id);
+		glDeleteProgram(v.id);
+		memdelete_arr(v.uniform_location);
+		v.id = 0;
+
+		version_map.erase(key);
+	}
 
 	custom_code_map.erase(p_code_id);
 }
 
 void ShaderGLES3::set_base_material_tex_index(int p_idx) {
-
 	base_material_tex_index = p_idx;
 }
 
 ShaderGLES3::ShaderGLES3() {
-	version = NULL;
+	version = nullptr;
 	last_custom_code = 1;
 	uniforms_dirty = true;
 	base_material_tex_index = 0;
 }
 
 ShaderGLES3::~ShaderGLES3() {
-
 	finish();
 }
