@@ -5,8 +5,8 @@
 /*                           GODOT ENGINE                                */
 /*                      https://godotengine.org                          */
 /*************************************************************************/
-/* Copyright (c) 2007-2018 Juan Linietsky, Ariel Manzur.                 */
-/* Copyright (c) 2014-2018 Godot Engine contributors (cf. AUTHORS.md)    */
+/* Copyright (c) 2007-2022 Juan Linietsky, Ariel Manzur.                 */
+/* Copyright (c) 2014-2022 Godot Engine contributors (cf. AUTHORS.md).   */
 /*                                                                       */
 /* Permission is hereby granted, free of charge, to any person obtaining */
 /* a copy of this software and associated documentation files (the       */
@@ -30,32 +30,38 @@
 
 #include "create_dialog.h"
 
-#include "class_db.h"
+#include "core/class_db.h"
+#include "core/os/keyboard.h"
+#include "core/print_string.h"
+#include "editor_feature_profile.h"
 #include "editor_help.h"
 #include "editor_node.h"
+#include "editor_scale.h"
 #include "editor_settings.h"
-#include "os/keyboard.h"
-#include "print_string.h"
 #include "scene/gui/box_container.h"
 
-void CreateDialog::popup_create(bool p_dont_clear, bool p_replace_mode) {
+void CreateDialog::popup_create(bool p_dont_clear, bool p_replace_mode, const String &p_select_type) {
+	type_list.clear();
+	ClassDB::get_class_list(&type_list);
+	ScriptServer::get_global_class_list(&type_list);
+	type_list.sort_custom<StringName::AlphCompare>();
 
 	recent->clear();
 
 	FileAccess *f = FileAccess::open(EditorSettings::get_singleton()->get_project_settings_dir().plus_file("create_recent." + base_type), FileAccess::READ);
 
 	if (f) {
-
 		TreeItem *root = recent->create_item();
 
+		String icon_fallback = has_icon(base_type, "EditorIcons") ? base_type : "Object";
+
 		while (!f->eof_reached()) {
 			String l = f->get_line().strip_edges();
-
-			if (l != String()) {
-
+			String name = l.split(" ")[0];
+			if ((ClassDB::class_exists(name) || ScriptServer::is_global_class(name)) && !_is_class_disabled_by_feature_profile(name)) {
 				TreeItem *ti = recent->create_item(root);
 				ti->set_text(0, l);
-				ti->set_icon(0, _get_editor_icon(l));
+				ti->set_icon(0, EditorNode::get_singleton()->get_class_icon(name, icon_fallback));
 			}
 		}
 
@@ -69,7 +75,6 @@ void CreateDialog::popup_create(bool p_dont_clear, bool p_replace_mode) {
 	favorite_list.clear();
 
 	if (f) {
-
 		while (!f->eof_reached()) {
 			String l = f->get_line().strip_edges();
 
@@ -81,13 +86,14 @@ void CreateDialog::popup_create(bool p_dont_clear, bool p_replace_mode) {
 		memdelete(f);
 	}
 
-	_update_favorite_list();
+	_save_and_update_favorite_list();
 
 	// Restore valid window bounds or pop up at default size.
-	if (EditorSettings::get_singleton()->has_setting("interface/dialogs/create_new_node_bounds")) {
-		popup(EditorSettings::get_singleton()->get("interface/dialogs/create_new_node_bounds"));
+	Rect2 saved_size = EditorSettings::get_singleton()->get_project_metadata("dialog_bounds", "create_new_node", Rect2());
+	if (saved_size != Rect2()) {
+		popup(saved_size);
 	} else {
-		popup_centered_ratio();
+		popup_centered_clamped(Size2(900, 700) * EDSCALE, 0.8);
 	}
 
 	if (p_dont_clear) {
@@ -100,19 +106,10 @@ void CreateDialog::popup_create(bool p_dont_clear, bool p_replace_mode) {
 
 	_update_search();
 
-	bool enable_rl = EditorSettings::get_singleton()->get("docks/scene_tree/draw_relationship_lines");
-	Color rl_color = EditorSettings::get_singleton()->get("docks/scene_tree/relationship_line_color");
-
-	if (enable_rl) {
-		search_options->add_constant_override("draw_relationship_lines", 1);
-		search_options->add_color_override("relationship_line_color", rl_color);
-	} else {
-		search_options->add_constant_override("draw_relationship_lines", 0);
-	}
-
 	is_replace_mode = p_replace_mode;
 
 	if (p_replace_mode) {
+		select_type(p_select_type);
 		set_title(vformat(TTR("Change %s Type"), base_type));
 		get_ok()->set_text(TTR("Change"));
 	} else {
@@ -122,83 +119,116 @@ void CreateDialog::popup_create(bool p_dont_clear, bool p_replace_mode) {
 }
 
 void CreateDialog::_text_changed(const String &p_newtext) {
-
 	_update_search();
 }
 
 void CreateDialog::_sbox_input(const Ref<InputEvent> &p_ie) {
-
 	Ref<InputEventKey> k = p_ie;
-	if (k.is_valid() && (k->get_scancode() == KEY_UP ||
-								k->get_scancode() == KEY_DOWN ||
-								k->get_scancode() == KEY_PAGEUP ||
-								k->get_scancode() == KEY_PAGEDOWN)) {
-
+	if (k.is_valid() && (k->get_scancode() == KEY_UP || k->get_scancode() == KEY_DOWN || k->get_scancode() == KEY_PAGEUP || k->get_scancode() == KEY_PAGEDOWN)) {
 		search_options->call("_gui_input", k);
 		search_box->accept_event();
 	}
 }
 
-Ref<Texture> CreateDialog::_get_editor_icon(const String &p_type) const {
-
-	if (has_icon(p_type, "EditorIcons")) {
-		return get_icon(p_type, "EditorIcons");
+void CreateDialog::add_type(const String &p_type, HashMap<String, TreeItem *> &p_types, TreeItem *p_root, TreeItem **to_select) {
+	if (p_types.has(p_type)) {
+		return;
 	}
 
-	const Map<String, Vector<EditorData::CustomType> > &p_map = EditorNode::get_editor_data().get_custom_types();
-	for (const Map<String, Vector<EditorData::CustomType> >::Element *E = p_map.front(); E; E = E->next()) {
-		const Vector<EditorData::CustomType> &ct = E->value();
-		for (int i = 0; i < ct.size(); ++i) {
-			if (ct[i].name == p_type) {
-				if (ct[i].icon.is_valid()) {
-					return ct[i].icon;
-				} else {
-					return get_icon("Object", "EditorIcons");
-				}
+	bool cpp_type = ClassDB::class_exists(p_type);
+	EditorData &ed = EditorNode::get_editor_data();
+
+	if (p_type == base_type) {
+		return;
+	}
+
+	if (cpp_type) {
+		if (!ClassDB::is_parent_class(p_type, base_type)) {
+			return;
+		}
+	} else {
+		if (!search_loaded_scripts.has(p_type)) {
+			search_loaded_scripts[p_type] = ed.script_class_load_script(p_type);
+		}
+
+		if (!ScriptServer::is_global_class(p_type) || !ed.script_class_is_parent(p_type, base_type)) {
+			return;
+		}
+
+		String script_path = ScriptServer::get_global_class_path(p_type);
+		if (script_path.find("res://addons/", 0) != -1) {
+			if (!EditorNode::get_singleton()->is_addon_plugin_enabled(script_path.get_slicec('/', 3))) {
+				return;
 			}
 		}
 	}
 
-	return get_icon("Object", "EditorIcons");
-}
-
-void CreateDialog::add_type(const String &p_type, HashMap<String, TreeItem *> &p_types, TreeItem *p_root, TreeItem **to_select) {
-
-	if (p_types.has(p_type))
-		return;
-	if (!ClassDB::is_parent_class(p_type, base_type) || p_type == base_type)
-		return;
-
-	String inherits = ClassDB::get_parent_class(p_type);
+	String inherits = cpp_type ? ClassDB::get_parent_class(p_type) : ed.script_class_get_base(p_type);
 
 	TreeItem *parent = p_root;
 
 	if (inherits.length()) {
-
 		if (!p_types.has(inherits)) {
-
 			add_type(inherits, p_types, p_root, to_select);
 		}
 
-		if (p_types.has(inherits))
+		if (p_types.has(inherits)) {
 			parent = p_types[inherits];
+		} else if (ScriptServer::is_global_class(inherits)) {
+			return;
+		}
 	}
 
+	bool can_instance = (cpp_type && ClassDB::can_instance(p_type)) || ScriptServer::is_global_class(p_type);
+
 	TreeItem *item = search_options->create_item(parent);
-	item->set_text(0, p_type);
-	if (!ClassDB::can_instance(p_type)) {
+	if (cpp_type) {
+		item->set_text(0, p_type);
+	} else {
+		item->set_metadata(0, p_type);
+		item->set_text(0, p_type + " (" + ScriptServer::get_global_class_path(p_type).get_file() + ")");
+	}
+	if (!can_instance) {
 		item->set_custom_color(0, get_color("disabled_font_color", "Editor"));
 		item->set_selectable(0, false);
-	} else {
-		bool is_search_subsequence = search_box->get_text().is_subsequence_ofi(p_type);
-		String to_select_type = *to_select ? (*to_select)->get_text(0) : "";
-		bool current_item_is_preffered = ClassDB::is_parent_class(p_type, preferred_search_result_type) && !ClassDB::is_parent_class(to_select_type, preferred_search_result_type);
-		if (*to_select && p_type.length() < (*to_select)->get_text(0).length()) {
-			current_item_is_preffered = true;
-		}
+	} else if (!(*to_select && (*to_select)->get_text(0) == search_box->get_text())) {
+		String search_term = search_box->get_text().to_lower();
 
-		if (((!*to_select || current_item_is_preffered) && is_search_subsequence) || search_box->get_text() == p_type) {
+		// if the node name matches exactly as the search, the node should be selected.
+		// this also fixes when the user clicks on recent nodes.
+		if (p_type.to_lower() == search_term) {
 			*to_select = item;
+		} else {
+			bool current_type_prefered = _is_type_prefered(p_type);
+			bool selected_type_prefered = *to_select ? _is_type_prefered((*to_select)->get_text(0).split(" ")[0]) : false;
+
+			bool is_subsequence_of_type = search_box->get_text().is_subsequence_ofi(p_type);
+			bool is_substring_of_type = p_type.to_lower().find(search_term) >= 0;
+			bool is_substring_of_selected = false;
+			bool is_subsequence_of_selected = false;
+			bool is_selected_equal = false;
+
+			if (*to_select) {
+				String name = (*to_select)->get_text(0).split(" ")[0].to_lower();
+				is_substring_of_selected = name.find(search_term) >= 0;
+				is_subsequence_of_selected = search_term.is_subsequence_of(name);
+				is_selected_equal = name == search_term;
+			}
+
+			if (is_subsequence_of_type && !is_selected_equal) {
+				if (is_substring_of_type) {
+					if (!is_substring_of_selected || (current_type_prefered && !selected_type_prefered)) {
+						*to_select = item;
+					}
+				} else {
+					// substring results weigh more than subsequences, so let's make sure we don't override them
+					if (!is_substring_of_selected) {
+						if (!is_subsequence_of_selected || (current_type_prefered && !selected_type_prefered)) {
+							*to_select = item;
+						}
+					}
+				}
+			}
 		}
 	}
 
@@ -210,76 +240,134 @@ void CreateDialog::add_type(const String &p_type, HashMap<String, TreeItem *> &p
 		// don't collapse the root node
 		collapse &= (item != p_root);
 		// don't collapse abstract nodes on the first tree level
-		collapse &= ((parent != p_root) || (ClassDB::can_instance(p_type)));
+		collapse &= ((parent != p_root) || (can_instance));
 		item->set_collapsed(collapse);
 	}
 
-	const String &description = EditorHelp::get_doc_data()->class_list[p_type].brief_description;
+	const String &description = DTR(EditorHelp::get_doc_data()->class_list[p_type].brief_description);
 	item->set_tooltip(0, description);
 
-	if (has_icon(p_type, "EditorIcons")) {
-
-		item->set_icon(0, get_icon(p_type, "EditorIcons"));
-	}
+	String icon_fallback = has_icon(base_type, "EditorIcons") ? base_type : "Object";
+	item->set_icon(0, EditorNode::get_singleton()->get_class_icon(p_type, icon_fallback));
 
 	p_types[p_type] = item;
 }
 
-void CreateDialog::_update_search() {
+bool CreateDialog::_is_type_prefered(const String &type) {
+	bool cpp_type = ClassDB::class_exists(type);
+	EditorData &ed = EditorNode::get_editor_data();
 
+	if (cpp_type) {
+		return ClassDB::is_parent_class(type, preferred_search_result_type);
+	}
+	return ed.script_class_is_parent(type, preferred_search_result_type);
+}
+
+bool CreateDialog::_is_class_disabled_by_feature_profile(const StringName &p_class) {
+	Ref<EditorFeatureProfile> profile = EditorFeatureProfileManager::get_singleton()->get_current_profile();
+	if (profile.is_null()) {
+		return false;
+	}
+
+	return profile->is_class_disabled(p_class);
+}
+
+void CreateDialog::select_type(const String &p_type) {
+	TreeItem *to_select;
+	if (search_options_types.has(p_type)) {
+		to_select = search_options_types[p_type];
+	} else {
+		to_select = search_options->get_root();
+	}
+
+	// uncollapse from selected type to top level
+	// TODO: should this be in tree?
+	TreeItem *cur = to_select;
+	while (cur) {
+		cur->set_collapsed(false);
+		cur = cur->get_parent();
+	}
+
+	to_select->select(0);
+
+	search_options->scroll_to_item(to_select);
+}
+
+void CreateDialog::_update_search() {
 	search_options->clear();
 	favorite->set_disabled(true);
 
-	help_bit->set_text("");
-	/*
-	TreeItem *root = search_options->create_item();
-	_parse_fs(EditorFileSystem::get_singleton()->get_filesystem());
-*/
+	help_bit->set_text(vformat(TTR("No results for \"%s\"."), search_box->get_text()));
+	help_bit->get_rich_text()->set_self_modulate(Color(1, 1, 1, 0.5));
 
-	HashMap<String, TreeItem *> types;
+	search_options_types.clear();
 
 	TreeItem *root = search_options->create_item();
+	EditorData &ed = EditorNode::get_editor_data();
 
 	root->set_text(0, base_type);
-	if (has_icon(base_type, "EditorIcons")) {
-		root->set_icon(0, get_icon(base_type, "EditorIcons"));
-	}
+	String base_icon = has_icon(base_type, "EditorIcons") ? base_type : "Object";
+	root->set_icon(0, get_icon(base_icon, "EditorIcons"));
 
-	List<StringName>::Element *I = type_list.front();
-	TreeItem *to_select = search_box->get_text() == base_type ? root : NULL;
-
-	for (; I; I = I->next()) {
+	TreeItem *to_select = search_box->get_text() == base_type ? root : nullptr;
 
+	for (List<StringName>::Element *I = type_list.front(); I; I = I->next()) {
 		String type = I->get();
 
-		if (base_type == "Node" && type.begins_with("Editor"))
-			continue; // do not show editor nodes
-
-		if (base_type == "Resource" && ClassDB::is_parent_class(type, "PluginScript"))
-			// PluginScript must be initialized before use, which is not possible here
+		if (_is_class_disabled_by_feature_profile(type)) {
 			continue;
+		}
+		bool cpp_type = ClassDB::class_exists(type);
 
-		if (!ClassDB::can_instance(type))
+		if (base_type == "Node" && type.begins_with("Editor")) {
+			continue; // do not show editor nodes
+		}
+
+		if (cpp_type && !ClassDB::can_instance(type)) {
 			continue; // can't create what can't be instanced
+		}
+
+		if (cpp_type) {
+			bool skip = false;
+
+			for (Set<StringName>::Element *E = type_blacklist.front(); E && !skip; E = E->next()) {
+				if (ClassDB::is_parent_class(type, E->get())) {
+					skip = true;
+				}
+			}
+			if (skip) {
+				continue;
+			}
+		}
 
 		if (search_box->get_text() == "") {
-			add_type(type, types, root, &to_select);
+			add_type(type, search_options_types, root, &to_select);
 		} else {
-
 			bool found = false;
-			String type = I->get();
-			while (type != "" && ClassDB::is_parent_class(type, base_type) && type != base_type) {
-				if (search_box->get_text().is_subsequence_ofi(type)) {
+			String type2 = type;
+			bool cpp_type2 = cpp_type;
 
+			if (!cpp_type && !search_loaded_scripts.has(type)) {
+				search_loaded_scripts[type] = ed.script_class_load_script(type);
+			}
+
+			while (type2 != "" && (cpp_type2 ? ClassDB::is_parent_class(type2, base_type) : ed.script_class_is_parent(type2, base_type)) && type2 != base_type) {
+				if (search_box->get_text().is_subsequence_ofi(type2)) {
 					found = true;
 					break;
 				}
 
-				type = ClassDB::get_parent_class(type);
+				type2 = cpp_type2 ? ClassDB::get_parent_class(type2) : ed.script_class_get_base(type2);
+				cpp_type2 = cpp_type2 || ClassDB::class_exists(type2); // Built-in class can't inherit from custom type, so we can skip the check if it's already true.
+
+				if (!cpp_type2 && !search_loaded_scripts.has(type2)) {
+					search_loaded_scripts[type2] = ed.script_class_load_script(type2);
+				}
 			}
 
-			if (found)
-				add_type(I->get(), types, root, &to_select);
+			if (found) {
+				add_type(type, search_options_types, root, &to_select);
+			}
 		}
 
 		if (EditorNode::get_editor_data().get_custom_types().has(type) && ClassDB::is_parent_class(type, base_type)) {
@@ -287,27 +375,27 @@ void CreateDialog::_update_search() {
 
 			const Vector<EditorData::CustomType> &ct = EditorNode::get_editor_data().get_custom_types()[type];
 			for (int i = 0; i < ct.size(); i++) {
-
 				bool show = search_box->get_text().is_subsequence_ofi(ct[i].name);
 
-				if (!show)
+				if (!show) {
 					continue;
+				}
 
-				if (!types.has(type))
-					add_type(type, types, root, &to_select);
+				if (!search_options_types.has(type)) {
+					add_type(type, search_options_types, root, &to_select);
+				}
 
 				TreeItem *ti;
-				if (types.has(type))
-					ti = types[type];
-				else
+				if (search_options_types.has(type)) {
+					ti = search_options_types[type];
+				} else {
 					ti = search_options->get_root();
+				}
 
 				TreeItem *item = search_options->create_item(ti);
 				item->set_metadata(0, type);
 				item->set_text(0, ct[i].name);
-				if (ct[i].icon.is_valid()) {
-					item->set_icon(0, ct[i].icon);
-				}
+				item->set_icon(0, ct[i].icon.is_valid() ? ct[i].icon : get_icon(base_icon, "EditorIcons"));
 
 				if (!to_select || ct[i].name == search_box->get_text()) {
 					to_select = item;
@@ -327,26 +415,26 @@ void CreateDialog::_update_search() {
 		favorite->set_pressed(favorite_list.find(to_select->get_text(0)) != -1);
 	}
 
-	get_ok()->set_disabled(root->get_children() == NULL);
+	get_ok()->set_disabled(root->get_children() == nullptr);
 }
 
 void CreateDialog::_confirmed() {
-
 	TreeItem *ti = search_options->get_selected();
-	if (!ti)
+	if (!ti) {
 		return;
+	}
 
 	FileAccess *f = FileAccess::open(EditorSettings::get_singleton()->get_project_settings_dir().plus_file("create_recent." + base_type), FileAccess::WRITE);
 
 	if (f) {
 		f->store_line(get_selected_type());
 		TreeItem *t = recent->get_root();
-		if (t)
+		if (t) {
 			t = t->get_children();
+		}
 		int count = 0;
 		while (t) {
 			if (t->get_text(0) != get_selected_type()) {
-
 				f->store_line(t->get_text(0));
 			}
 
@@ -366,10 +454,11 @@ void CreateDialog::_confirmed() {
 }
 
 void CreateDialog::_notification(int p_what) {
-
 	switch (p_what) {
 		case NOTIFICATION_ENTER_TREE: {
 			connect("confirmed", this, "_confirmed");
+			search_box->set_right_icon(get_icon("Search", "EditorIcons"));
+			search_box->set_clear_button_enabled(true);
 			favorite->set_icon(get_icon("Favorites", "EditorIcons"));
 		} break;
 		case NOTIFICATION_EXIT_TREE: {
@@ -382,24 +471,24 @@ void CreateDialog::_notification(int p_what) {
 			}
 		} break;
 		case NOTIFICATION_POPUP_HIDE: {
-			EditorSettings::get_singleton()->set("interface/dialogs/create_new_node_bounds", get_rect());
+			EditorSettings::get_singleton()->get_project_metadata("dialog_bounds", "create_new_node", get_rect());
+			search_loaded_scripts.clear();
 		} break;
 	}
 }
 
 void CreateDialog::set_base_type(const String &p_base) {
-
 	base_type = p_base;
-	if (is_replace_mode)
+	if (is_replace_mode) {
 		set_title(vformat(TTR("Change %s Type"), p_base));
-	else
+	} else {
 		set_title(vformat(TTR("Create New %s"), p_base));
+	}
 
 	_update_search();
 }
 
 String CreateDialog::get_base_type() const {
-
 	return base_type;
 }
 
@@ -408,83 +497,81 @@ void CreateDialog::set_preferred_search_result_type(const String &p_preferred_ty
 }
 
 String CreateDialog::get_preferred_search_result_type() {
-
 	return preferred_search_result_type;
 }
 String CreateDialog::get_selected_type() {
-
 	TreeItem *selected = search_options->get_selected();
-	if (selected)
+	if (selected) {
 		return selected->get_text(0);
-	else
+	} else {
 		return String();
+	}
 }
 
-Object *CreateDialog::instance_selected() {
-
+Variant CreateDialog::instance_selected() {
 	TreeItem *selected = search_options->get_selected();
 
 	if (selected) {
-
 		Variant md = selected->get_metadata(0);
 
 		String custom;
-		if (md.get_type() != Variant::NIL)
+		if (md.get_type() != Variant::NIL) {
 			custom = md;
+		}
 
 		if (custom != String()) {
-			if (EditorNode::get_editor_data().get_custom_types().has(custom)) {
-
-				for (int i = 0; i < EditorNode::get_editor_data().get_custom_types()[custom].size(); i++) {
-					if (EditorNode::get_editor_data().get_custom_types()[custom][i].name == selected->get_text(0)) {
-						Ref<Texture> icon = EditorNode::get_editor_data().get_custom_types()[custom][i].icon;
-						Ref<Script> script = EditorNode::get_editor_data().get_custom_types()[custom][i].script;
-						String name = selected->get_text(0);
-
-						Object *ob = ClassDB::instance(custom);
-						ERR_FAIL_COND_V(!ob, NULL);
-						if (ob->is_class("Node")) {
-							ob->call("set_name", name);
-						}
-						ob->set_script(script.get_ref_ptr());
-						if (icon.is_valid())
-							ob->set_meta("_editor_icon", icon);
-						return ob;
-					}
+			if (ScriptServer::is_global_class(custom)) {
+				Variant obj = EditorNode::get_editor_data().script_class_instance(custom);
+				Node *n = Object::cast_to<Node>(obj);
+				if (n) {
+					n->set_name(custom);
 				}
+				return obj;
 			}
+			return EditorNode::get_editor_data().instance_custom_type(selected->get_text(0), custom);
 		} else {
 			return ClassDB::instance(selected->get_text(0));
 		}
 	}
 
-	return NULL;
+	return Variant();
 }
 
 void CreateDialog::_item_selected() {
-
 	TreeItem *item = search_options->get_selected();
-	if (!item)
+	if (!item) {
 		return;
+	}
 
 	String name = item->get_text(0);
 
 	favorite->set_disabled(false);
 	favorite->set_pressed(favorite_list.find(name) != -1);
 
-	if (!EditorHelp::get_doc_data()->class_list.has(name))
+	if (!EditorHelp::get_doc_data()->class_list.has(name)) {
 		return;
+	}
 
-	help_bit->set_text(EditorHelp::get_doc_data()->class_list[name].brief_description);
+	const String brief_desc = DTR(EditorHelp::get_doc_data()->class_list[name].brief_description);
+	if (!brief_desc.empty()) {
+		// Display both class name and description, since the help bit may be displayed
+		// far away from the location (especially if the dialog was resized to be taller).
+		help_bit->set_text(vformat("[b]%s[/b]: %s", name, brief_desc));
+		help_bit->get_rich_text()->set_self_modulate(Color(1, 1, 1, 1));
+	} else {
+		// Use nested `vformat()` as translators shouldn't interfere with BBCode tags.
+		help_bit->set_text(vformat(TTR("No description available for %s."), vformat("[b]%s[/b]", name)));
+		help_bit->get_rich_text()->set_self_modulate(Color(1, 1, 1, 0.5));
+	}
 
 	get_ok()->set_disabled(false);
 }
 
 void CreateDialog::_favorite_toggled() {
-
 	TreeItem *item = search_options->get_selected();
-	if (!item)
+	if (!item) {
 		return;
+	}
 
 	String name = item->get_text(0);
 
@@ -496,70 +583,79 @@ void CreateDialog::_favorite_toggled() {
 		favorite->set_pressed(false);
 	}
 
-	_save_favorite_list();
-	_update_favorite_list();
+	_save_and_update_favorite_list();
 }
 
 void CreateDialog::_save_favorite_list() {
-
 	FileAccess *f = FileAccess::open(EditorSettings::get_singleton()->get_project_settings_dir().plus_file("favorites." + base_type), FileAccess::WRITE);
 
 	if (f) {
-
 		for (int i = 0; i < favorite_list.size(); i++) {
-
-			f->store_line(favorite_list[i]);
+			String l = favorite_list[i];
+			String name = l.split(" ")[0];
+			if (!(ClassDB::class_exists(name) || ScriptServer::is_global_class(name))) {
+				continue;
+			}
+			f->store_line(l);
 		}
 		memdelete(f);
 	}
 }
 
 void CreateDialog::_update_favorite_list() {
-
 	favorites->clear();
+
 	TreeItem *root = favorites->create_item();
+
+	String icon_fallback = has_icon(base_type, "EditorIcons") ? base_type : "Object";
+
 	for (int i = 0; i < favorite_list.size(); i++) {
-		TreeItem *ti = favorites->create_item(root);
 		String l = favorite_list[i];
+		String name = l.split(" ")[0];
+		if (!((ClassDB::class_exists(name) || ScriptServer::is_global_class(name)) && !_is_class_disabled_by_feature_profile(name))) {
+			continue;
+		}
+
+		TreeItem *ti = favorites->create_item(root);
 		ti->set_text(0, l);
-		ti->set_icon(0, _get_editor_icon(l));
+		ti->set_icon(0, EditorNode::get_singleton()->get_class_icon(name, icon_fallback));
 	}
+	emit_signal("favorites_updated");
 }
 
 void CreateDialog::_history_selected() {
-
 	TreeItem *item = recent->get_selected();
-	if (!item)
+	if (!item) {
 		return;
+	}
 
-	search_box->set_text(item->get_text(0));
+	search_box->set_text(item->get_text(0).get_slicec(' ', 0));
+	favorites->deselect_all();
 	_update_search();
 }
 
 void CreateDialog::_favorite_selected() {
-
 	TreeItem *item = favorites->get_selected();
-	if (!item)
+	if (!item) {
 		return;
+	}
 
-	search_box->set_text(item->get_text(0));
+	search_box->set_text(item->get_text(0).get_slicec(' ', 0));
+	recent->deselect_all();
 	_update_search();
 }
 
 void CreateDialog::_history_activated() {
-
 	_history_selected();
 	_confirmed();
 }
 
 void CreateDialog::_favorite_activated() {
-
 	_favorite_selected();
 	_confirmed();
 }
 
 Variant CreateDialog::get_drag_data_fw(const Point2 &p_point, Control *p_from) {
-
 	TreeItem *ti = favorites->get_item_at_position(p_point);
 	if (ti) {
 		Dictionary d;
@@ -578,7 +674,6 @@ Variant CreateDialog::get_drag_data_fw(const Point2 &p_point, Control *p_from) {
 }
 
 bool CreateDialog::can_drop_data_fw(const Point2 &p_point, const Variant &p_data, Control *p_from) const {
-
 	Dictionary d = p_data;
 	if (d.has("type") && String(d["type"]) == "create_favorite_drag") {
 		favorites->set_drop_mode_flags(Tree::DROP_MODE_INBETWEEN);
@@ -588,25 +683,27 @@ bool CreateDialog::can_drop_data_fw(const Point2 &p_point, const Variant &p_data
 	return false;
 }
 void CreateDialog::drop_data_fw(const Point2 &p_point, const Variant &p_data, Control *p_from) {
-
 	Dictionary d = p_data;
 
 	TreeItem *ti = favorites->get_item_at_position(p_point);
-	if (!ti)
+	if (!ti) {
 		return;
+	}
 
 	String drop_at = ti->get_text(0);
 	int ds = favorites->get_drop_section_at_position(p_point);
 
 	int drop_idx = favorite_list.find(drop_at);
-	if (drop_idx < 0)
+	if (drop_idx < 0) {
 		return;
+	}
 
 	String type = d["class"];
 
 	int from_idx = favorite_list.find(type);
-	if (from_idx < 0)
+	if (from_idx < 0) {
 		return;
+	}
 
 	if (drop_idx == from_idx) {
 		ds = -1; //cause it will be gone
@@ -626,12 +723,15 @@ void CreateDialog::drop_data_fw(const Point2 &p_point, const Variant &p_data, Co
 		}
 	}
 
+	_save_and_update_favorite_list();
+}
+
+void CreateDialog::_save_and_update_favorite_list() {
 	_save_favorite_list();
 	_update_favorite_list();
 }
 
 void CreateDialog::_bind_methods() {
-
 	ClassDB::bind_method(D_METHOD("_text_changed"), &CreateDialog::_text_changed);
 	ClassDB::bind_method(D_METHOD("_confirmed"), &CreateDialog::_confirmed);
 	ClassDB::bind_method(D_METHOD("_sbox_input"), &CreateDialog::_sbox_input);
@@ -641,21 +741,19 @@ void CreateDialog::_bind_methods() {
 	ClassDB::bind_method(D_METHOD("_favorite_selected"), &CreateDialog::_favorite_selected);
 	ClassDB::bind_method(D_METHOD("_history_activated"), &CreateDialog::_history_activated);
 	ClassDB::bind_method(D_METHOD("_favorite_activated"), &CreateDialog::_favorite_activated);
+	ClassDB::bind_method(D_METHOD("_save_and_update_favorite_list"), &CreateDialog::_save_and_update_favorite_list);
 
 	ClassDB::bind_method("get_drag_data_fw", &CreateDialog::get_drag_data_fw);
 	ClassDB::bind_method("can_drop_data_fw", &CreateDialog::can_drop_data_fw);
 	ClassDB::bind_method("drop_data_fw", &CreateDialog::drop_data_fw);
 
 	ADD_SIGNAL(MethodInfo("create"));
+	ADD_SIGNAL(MethodInfo("favorites_updated"));
 }
 
 CreateDialog::CreateDialog() {
-
 	is_replace_mode = false;
 
-	ClassDB::get_class_list(&type_list);
-	type_list.sort_custom<StringName::AlphCompare>();
-
 	set_resizable(true);
 
 	HSplitContainer *hsc = memnew(HSplitContainer);
@@ -673,9 +771,11 @@ CreateDialog::CreateDialog() {
 	fav_vb->add_margin_child(TTR("Favorites:"), favorites, true);
 	favorites->set_hide_root(true);
 	favorites->set_hide_folding(true);
+	favorites->set_allow_reselect(true);
 	favorites->connect("cell_selected", this, "_favorite_selected");
 	favorites->connect("item_activated", this, "_favorite_activated");
 	favorites->set_drag_forwarding(this);
+	favorites->add_constant_override("draw_guides", 1);
 
 	VBoxContainer *rec_vb = memnew(VBoxContainer);
 	vsc->add_child(rec_vb);
@@ -686,8 +786,10 @@ CreateDialog::CreateDialog() {
 	rec_vb->add_margin_child(TTR("Recent:"), recent, true);
 	recent->set_hide_root(true);
 	recent->set_hide_folding(true);
+	recent->set_allow_reselect(true);
 	recent->connect("cell_selected", this, "_history_selected");
 	recent->connect("item_activated", this, "_history_activated");
+	recent->add_constant_override("draw_guides", 1);
 
 	VBoxContainer *vbc = memnew(VBoxContainer);
 	hsc->add_child(vbc);
@@ -718,4 +820,7 @@ CreateDialog::CreateDialog() {
 	help_bit = memnew(EditorHelpBit);
 	vbc->add_margin_child(TTR("Description:"), help_bit);
 	help_bit->connect("request_hide", this, "_closed");
+
+	type_blacklist.insert("PluginScript"); // PluginScript must be initialized before use, which is not possible here
+	type_blacklist.insert("ScriptCreateDialog"); // This is an exposed editor Node that doesn't have an Editor prefix.
 }
