@@ -5,8 +5,8 @@
 /*                           GODOT ENGINE                                */
 /*                      https://godotengine.org                          */
 /*************************************************************************/
-/* Copyright (c) 2007-2018 Juan Linietsky, Ariel Manzur.                 */
-/* Copyright (c) 2014-2018 Godot Engine contributors (cf. AUTHORS.md)    */
+/* Copyright (c) 2007-2022 Juan Linietsky, Ariel Manzur.                 */
+/* Copyright (c) 2014-2022 Godot Engine contributors (cf. AUTHORS.md).   */
 /*                                                                       */
 /* Permission is hereby granted, free of charge, to any person obtaining */
 /* a copy of this software and associated documentation files (the       */
@@ -30,16 +30,13 @@
 
 #include "dependency_editor.h"
 
+#include "core/io/resource_loader.h"
+#include "core/os/file_access.h"
 #include "editor_node.h"
-#include "io/resource_loader.h"
-#include "os/file_access.h"
+#include "editor_scale.h"
 #include "scene/gui/margin_container.h"
 
-void DependencyEditor::_notification(int p_what) {
-}
-
 void DependencyEditor::_searched(const String &p_path) {
-
 	Map<String, String> dep_rename;
 	dep_rename[replacing] = p_path;
 
@@ -50,9 +47,7 @@ void DependencyEditor::_searched(const String &p_path) {
 }
 
 void DependencyEditor::_load_pressed(Object *p_item, int p_cell, int p_button) {
-
 	TreeItem *ti = Object::cast_to<TreeItem>(p_item);
-	String fname = ti->get_text(0);
 	replacing = ti->get_text(1);
 
 	search->set_title(TTR("Search Replacement For:") + " " + replacing.get_file());
@@ -63,25 +58,23 @@ void DependencyEditor::_load_pressed(Object *p_item, int p_cell, int p_button) {
 	for (List<String>::Element *E = ext.front(); E; E = E->next()) {
 		search->add_filter("*" + E->get());
 	}
-	search->popup_centered_ratio();
+	search->popup_centered_ratio(0.65); // So it doesn't completely cover the dialog below it.
 }
 
-void DependencyEditor::_fix_and_find(EditorFileSystemDirectory *efsd, Map<String, Map<String, String> > &candidates) {
-
+void DependencyEditor::_fix_and_find(EditorFileSystemDirectory *efsd, Map<String, Map<String, String>> &candidates) {
 	for (int i = 0; i < efsd->get_subdir_count(); i++) {
 		_fix_and_find(efsd->get_subdir(i), candidates);
 	}
 
 	for (int i = 0; i < efsd->get_file_count(); i++) {
-
 		String file = efsd->get_file(i);
-		if (!candidates.has(file))
+		if (!candidates.has(file)) {
 			continue;
+		}
 
 		String path = efsd->get_file_path(i);
 
 		for (Map<String, String>::Element *E = candidates[file].front(); E; E = E->next()) {
-
 			if (E->get() == String()) {
 				E->get() = path;
 				continue;
@@ -103,7 +96,6 @@ void DependencyEditor::_fix_and_find(EditorFileSystemDirectory *efsd, Map<String
 			int current_score = 0;
 
 			for (int j = 0; j < lostv.size(); j++) {
-
 				if (j < existingv.size() && lostv[j] == existingv[j]) {
 					existing_score++;
 				}
@@ -113,7 +105,6 @@ void DependencyEditor::_fix_and_find(EditorFileSystemDirectory *efsd, Map<String
 			}
 
 			if (current_score > existing_score) {
-
 				//if it was the same, could track distance to new path but..
 
 				E->get() = path; //replace by more accurate
@@ -123,14 +114,13 @@ void DependencyEditor::_fix_and_find(EditorFileSystemDirectory *efsd, Map<String
 }
 
 void DependencyEditor::_fix_all() {
-
-	if (!EditorFileSystem::get_singleton()->get_filesystem())
+	if (!EditorFileSystem::get_singleton()->get_filesystem()) {
 		return;
+	}
 
-	Map<String, Map<String, String> > candidates;
+	Map<String, Map<String, String>> candidates;
 
 	for (List<String>::Element *E = missing.front(); E; E = E->next()) {
-
 		String base = E->get().get_file();
 		if (!candidates.has(base)) {
 			candidates[base] = Map<String, String>();
@@ -143,10 +133,8 @@ void DependencyEditor::_fix_all() {
 
 	Map<String, String> remaps;
 
-	for (Map<String, Map<String, String> >::Element *E = candidates.front(); E; E = E->next()) {
-
+	for (Map<String, Map<String, String>>::Element *E = candidates.front(); E; E = E->next()) {
 		for (Map<String, String>::Element *F = E->get().front(); F; F = F->next()) {
-
 			if (F->get() != String()) {
 				remaps[F->key()] = F->get();
 			}
@@ -154,7 +142,6 @@ void DependencyEditor::_fix_all() {
 	}
 
 	if (remaps.size()) {
-
 		ResourceLoader::rename_dependencies(editing, remaps);
 
 		_update_list();
@@ -163,12 +150,10 @@ void DependencyEditor::_fix_all() {
 }
 
 void DependencyEditor::_update_file() {
-
 	EditorFileSystem::get_singleton()->update_file(editing);
 }
 
 void DependencyEditor::_update_list() {
-
 	List<String> deps;
 	ResourceLoader::get_dependencies(editing, &deps, true);
 
@@ -182,7 +167,6 @@ void DependencyEditor::_update_list() {
 	bool broken = false;
 
 	for (List<String>::Element *E = deps.front(); E; E = E->next()) {
-
 		TreeItem *item = tree->create_item(root);
 
 		String n = E->get();
@@ -198,12 +182,7 @@ void DependencyEditor::_update_list() {
 		}
 		String name = path.get_file();
 
-		Ref<Texture> icon;
-		if (has_icon(type, "EditorIcons")) {
-			icon = get_icon(type, "EditorIcons");
-		} else {
-			icon = get_icon("Object", "EditorIcons");
-		}
+		Ref<Texture> icon = EditorNode::get_singleton()->get_class_icon(type);
 		item->set_text(0, name);
 		item->set_icon(0, icon);
 		item->set_metadata(0, type);
@@ -222,29 +201,26 @@ void DependencyEditor::_update_list() {
 }
 
 void DependencyEditor::edit(const String &p_path) {
-
 	editing = p_path;
 	set_title(TTR("Dependencies For:") + " " + p_path.get_file());
 
 	_update_list();
-	popup_centered_ratio();
+	popup_centered_ratio(0.7); // So it doesn't completely cover the dialog below it.
 
 	if (EditorNode::get_singleton()->is_scene_open(p_path)) {
-		EditorNode::get_singleton()->show_warning(vformat(TTR("Scene '%s' is currently being edited.\nChanges will not take effect unless reloaded."), p_path.get_file()));
+		EditorNode::get_singleton()->show_warning(vformat(TTR("Scene '%s' is currently being edited.\nChanges will only take effect when reloaded."), p_path.get_file()));
 	} else if (ResourceCache::has(p_path)) {
-		EditorNode::get_singleton()->show_warning(vformat(TTR("Resource '%s' is in use.\nChanges will take effect when reloaded."), p_path.get_file()));
+		EditorNode::get_singleton()->show_warning(vformat(TTR("Resource '%s' is in use.\nChanges will only take effect when reloaded."), p_path.get_file()));
 	}
 }
 
 void DependencyEditor::_bind_methods() {
-
 	ClassDB::bind_method(D_METHOD("_searched"), &DependencyEditor::_searched);
 	ClassDB::bind_method(D_METHOD("_load_pressed"), &DependencyEditor::_load_pressed);
 	ClassDB::bind_method(D_METHOD("_fix_all"), &DependencyEditor::_fix_all);
 }
 
 DependencyEditor::DependencyEditor() {
-
 	VBoxContainer *vb = memnew(VBoxContainer);
 	vb->set_name(TTR("Dependencies"));
 	add_child(vb);
@@ -283,7 +259,6 @@ DependencyEditor::DependencyEditor() {
 
 /////////////////////////////////////
 void DependencyEditorOwners::_list_rmb_select(int p_item, const Vector2 &p_pos) {
-
 	file_options->clear();
 	file_options->set_size(Size2(1, 1));
 	if (p_item >= 0) {
@@ -295,7 +270,6 @@ void DependencyEditorOwners::_list_rmb_select(int p_item, const Vector2 &p_pos)
 }
 
 void DependencyEditorOwners::_select_file(int p_idx) {
-
 	String fpath = owners->get_item_text(p_idx);
 
 	if (ResourceLoader::get_resource_type(fpath) == "PackedScene") {
@@ -306,63 +280,52 @@ void DependencyEditorOwners::_select_file(int p_idx) {
 }
 
 void DependencyEditorOwners::_file_option(int p_option) {
-
 	switch (p_option) {
 		case FILE_OPEN: {
 			int idx = owners->get_current();
-			if (idx < 0 || idx >= owners->get_item_count())
+			if (idx < 0 || idx >= owners->get_item_count()) {
 				break;
+			}
 			_select_file(idx);
 		} break;
 	}
 }
 
 void DependencyEditorOwners::_bind_methods() {
-
 	ClassDB::bind_method(D_METHOD("_list_rmb_select"), &DependencyEditorOwners::_list_rmb_select);
 	ClassDB::bind_method(D_METHOD("_file_option"), &DependencyEditorOwners::_file_option);
 	ClassDB::bind_method(D_METHOD("_select_file"), &DependencyEditorOwners::_select_file);
 }
 
 void DependencyEditorOwners::_fill_owners(EditorFileSystemDirectory *efsd) {
-
-	if (!efsd)
+	if (!efsd) {
 		return;
+	}
 
 	for (int i = 0; i < efsd->get_subdir_count(); i++) {
 		_fill_owners(efsd->get_subdir(i));
 	}
 
 	for (int i = 0; i < efsd->get_file_count(); i++) {
-
 		Vector<String> deps = efsd->get_file_deps(i);
-		//print_line(":::"+efsd->get_file_path(i));
 		bool found = false;
 		for (int j = 0; j < deps.size(); j++) {
-			//print_line("\t"+deps[j]+" vs "+editing);
 			if (deps[j] == editing) {
-				//print_line("found");
 				found = true;
 				break;
 			}
 		}
-		if (!found)
+		if (!found) {
 			continue;
-
-		Ref<Texture> icon;
-		String type = efsd->get_file_type(i);
-		if (!has_icon(type, "EditorIcons")) {
-			icon = get_icon("Object", "EditorIcons");
-		} else {
-			icon = get_icon(type, "EditorIcons");
 		}
 
+		Ref<Texture> icon = EditorNode::get_singleton()->get_class_icon(efsd->get_file_type(i));
+
 		owners->add_item(efsd->get_file_path(i), icon);
 	}
 }
 
 void DependencyEditorOwners::show(const String &p_path) {
-
 	editing = p_path;
 	owners->clear();
 	_fill_owners(EditorFileSystem::get_singleton()->get_filesystem());
@@ -372,7 +335,6 @@ void DependencyEditorOwners::show(const String &p_path) {
 }
 
 DependencyEditorOwners::DependencyEditorOwners(EditorNode *p_editor) {
-
 	editor = p_editor;
 
 	file_options = memnew(PopupMenu);
@@ -390,8 +352,9 @@ DependencyEditorOwners::DependencyEditorOwners(EditorNode *p_editor) {
 ///////////////////////
 
 void DependencyRemoveDialog::_find_files_in_removed_folder(EditorFileSystemDirectory *efsd, const String &p_folder) {
-	if (!efsd)
+	if (!efsd) {
 		return;
+	}
 
 	for (int i = 0; i < efsd->get_subdir_count(); ++i) {
 		_find_files_in_removed_folder(efsd->get_subdir(i), p_folder);
@@ -404,8 +367,9 @@ void DependencyRemoveDialog::_find_files_in_removed_folder(EditorFileSystemDirec
 }
 
 void DependencyRemoveDialog::_find_all_removed_dependencies(EditorFileSystemDirectory *efsd, Vector<RemovedDependency> &p_removed) {
-	if (!efsd)
+	if (!efsd) {
 		return;
+	}
 
 	for (int i = 0; i < efsd->get_subdir_count(); i++) {
 		_find_all_removed_dependencies(efsd->get_subdir(i), p_removed);
@@ -415,8 +379,9 @@ void DependencyRemoveDialog::_find_all_removed_dependencies(EditorFileSystemDire
 		const String path = efsd->get_file_path(i);
 
 		//It doesn't matter if a file we are about to delete will have some of its dependencies removed too
-		if (all_remove_files.has(path))
+		if (all_remove_files.has(path)) {
 			continue;
+		}
 
 		Vector<String> all_deps = efsd->get_file_deps(i);
 		for (int j = 0; j < all_deps.size(); ++j) {
@@ -463,7 +428,7 @@ void DependencyRemoveDialog::_build_removed_dependency_tree(const Vector<Removed
 		}
 
 		//List this file under this dependency
-		Ref<Texture> icon = has_icon(rd.file_type, "EditorIcons") ? get_icon(rd.file_type, "EditorIcons") : get_icon("Object", "EditorIcons");
+		Ref<Texture> icon = EditorNode::get_singleton()->get_class_icon(rd.file_type);
 		TreeItem *file_item = owners->create_item(tree_items[rd.dependency]);
 		file_item->set_text(0, rd.file);
 		file_item->set_icon(0, icon);
@@ -472,64 +437,129 @@ void DependencyRemoveDialog::_build_removed_dependency_tree(const Vector<Removed
 
 void DependencyRemoveDialog::show(const Vector<String> &p_folders, const Vector<String> &p_files) {
 	all_remove_files.clear();
-	to_delete.clear();
+	dirs_to_delete.clear();
+	files_to_delete.clear();
 	owners->clear();
 
 	for (int i = 0; i < p_folders.size(); ++i) {
 		String folder = p_folders[i].ends_with("/") ? p_folders[i] : (p_folders[i] + "/");
 		_find_files_in_removed_folder(EditorFileSystem::get_singleton()->get_filesystem_path(folder), folder);
-		to_delete.push_back(folder);
+		dirs_to_delete.push_back(folder);
 	}
 	for (int i = 0; i < p_files.size(); ++i) {
 		all_remove_files[p_files[i]] = String();
-		to_delete.push_back(p_files[i]);
+		files_to_delete.push_back(p_files[i]);
 	}
 
 	Vector<RemovedDependency> removed_deps;
 	_find_all_removed_dependencies(EditorFileSystem::get_singleton()->get_filesystem(), removed_deps);
 	removed_deps.sort();
-
 	if (removed_deps.empty()) {
 		owners->hide();
-		text->set_text(TTR("Remove selected files from the project? (no undo)"));
-		popup_centered_minsize(Size2(400, 100));
+		text->set_text(TTR("Remove the selected files from the project? (Cannot be undone.)\nDepending on your filesystem configuration, the files will either be moved to the system trash or deleted permanently."));
+		set_size(Size2());
+		popup_centered();
 	} else {
 		_build_removed_dependency_tree(removed_deps);
 		owners->show();
-		text->set_text(TTR("The files being removed are required by other resources in order for them to work.\nRemove them anyway? (no undo)"));
-		popup_centered_minsize(Size2(500, 350));
+		text->set_text(TTR("The files being removed are required by other resources in order for them to work.\nRemove them anyway? (Cannot be undone.)\nDepending on your filesystem configuration, the files will either be moved to the system trash or deleted permanently."));
+		popup_centered(Size2(500, 350));
 	}
+	EditorFileSystem::get_singleton()->scan_changes();
 }
 
 void DependencyRemoveDialog::ok_pressed() {
-	bool files_only = true;
-	for (int i = 0; i < to_delete.size(); ++i) {
-		if (to_delete[i].ends_with("/")) {
-			files_only = false;
-		} else if (ResourceCache::has(to_delete[i])) {
-			Resource *res = ResourceCache::get(to_delete[i]);
-			res->set_path(""); //clear reference to path
+	for (int i = 0; i < files_to_delete.size(); ++i) {
+		if (ResourceCache::has(files_to_delete[i])) {
+			Resource *res = ResourceCache::get(files_to_delete[i]);
+			res->set_path("");
 		}
 
-		String path = OS::get_singleton()->get_resource_dir() + to_delete[i].replace_first("res://", "/");
-		print_line("Moving to trash: " + path);
+		// If the file we are deleting for e.g. the main scene, default environment,
+		// or audio bus layout, we must clear its definition in Project Settings.
+		if (files_to_delete[i] == ProjectSettings::get_singleton()->get("application/config/icon")) {
+			ProjectSettings::get_singleton()->set("application/config/icon", "");
+		}
+		if (files_to_delete[i] == ProjectSettings::get_singleton()->get("application/run/main_scene")) {
+			ProjectSettings::get_singleton()->set("application/run/main_scene", "");
+		}
+		if (files_to_delete[i] == ProjectSettings::get_singleton()->get("application/boot_splash/image")) {
+			ProjectSettings::get_singleton()->set("application/boot_splash/image", "");
+		}
+		if (files_to_delete[i] == ProjectSettings::get_singleton()->get("rendering/environment/default_environment")) {
+			ProjectSettings::get_singleton()->set("rendering/environment/default_environment", "");
+		}
+		if (files_to_delete[i] == ProjectSettings::get_singleton()->get("display/mouse_cursor/custom_image")) {
+			ProjectSettings::get_singleton()->set("display/mouse_cursor/custom_image", "");
+		}
+		if (files_to_delete[i] == ProjectSettings::get_singleton()->get("gui/theme/custom")) {
+			ProjectSettings::get_singleton()->set("gui/theme/custom", "");
+		}
+		if (files_to_delete[i] == ProjectSettings::get_singleton()->get("gui/theme/custom_font")) {
+			ProjectSettings::get_singleton()->set("gui/theme/custom_font", "");
+		}
+		if (files_to_delete[i] == ProjectSettings::get_singleton()->get("audio/default_bus_layout")) {
+			ProjectSettings::get_singleton()->set("audio/default_bus_layout", "");
+		}
+
+		String path = OS::get_singleton()->get_resource_dir() + files_to_delete[i].replace_first("res://", "/");
+		print_verbose("Moving to trash: " + path);
 		Error err = OS::get_singleton()->move_to_trash(path);
 		if (err != OK) {
-			EditorNode::get_singleton()->add_io_error(TTR("Cannot remove:") + "\n" + to_delete[i] + "\n");
+			EditorNode::get_singleton()->add_io_error(TTR("Cannot remove:") + "\n" + files_to_delete[i] + "\n");
+		} else {
+			emit_signal("file_removed", files_to_delete[i]);
 		}
 	}
 
-	if (files_only) {
-		//If we only deleted files we should only need to tell the file system about the files we touched.
-		for (int i = 0; i < to_delete.size(); ++i) {
-			EditorFileSystem::get_singleton()->update_file(to_delete[i]);
+	if (dirs_to_delete.size() == 0) {
+		// If we only deleted files we should only need to tell the file system about the files we touched.
+		for (int i = 0; i < files_to_delete.size(); ++i) {
+			EditorFileSystem::get_singleton()->update_file(files_to_delete[i]);
 		}
 	} else {
+		for (int i = 0; i < dirs_to_delete.size(); ++i) {
+			String path = OS::get_singleton()->get_resource_dir() + dirs_to_delete[i].replace_first("res://", "/");
+			print_verbose("Moving to trash: " + path);
+			Error err = OS::get_singleton()->move_to_trash(path);
+			if (err != OK) {
+				EditorNode::get_singleton()->add_io_error(TTR("Cannot remove:") + "\n" + dirs_to_delete[i] + "\n");
+			} else {
+				emit_signal("folder_removed", dirs_to_delete[i]);
+			}
+		}
+
 		EditorFileSystem::get_singleton()->scan_changes();
 	}
+
+	// If some files/dirs would be deleted, favorite dirs need to be updated
+	Vector<String> previous_favorites = EditorSettings::get_singleton()->get_favorites();
+	Vector<String> new_favorites;
+
+	for (int i = 0; i < previous_favorites.size(); ++i) {
+		if (previous_favorites[i].ends_with("/")) {
+			if (dirs_to_delete.find(previous_favorites[i]) < 0) {
+				new_favorites.push_back(previous_favorites[i]);
+			}
+		} else {
+			if (files_to_delete.find(previous_favorites[i]) < 0) {
+				new_favorites.push_back(previous_favorites[i]);
+			}
+		}
+	}
+
+	if (new_favorites.size() < previous_favorites.size()) {
+		EditorSettings::get_singleton()->set_favorites(new_favorites);
+	}
+}
+
+void DependencyRemoveDialog::_bind_methods() {
+	ADD_SIGNAL(MethodInfo("file_removed", PropertyInfo(Variant::STRING, "file")));
+	ADD_SIGNAL(MethodInfo("folder_removed", PropertyInfo(Variant::STRING, "folder")));
 }
 
 DependencyRemoveDialog::DependencyRemoveDialog() {
+	get_ok()->set_text(TTR("Remove"));
 
 	VBoxContainer *vb = memnew(VBoxContainer);
 	add_child(vb);
@@ -541,61 +571,60 @@ DependencyRemoveDialog::DependencyRemoveDialog() {
 	owners->set_hide_root(true);
 	vb->add_child(owners);
 	owners->set_v_size_flags(SIZE_EXPAND_FILL);
-	get_ok()->set_text(TTR("Remove"));
 }
 
 //////////////
 
-void DependencyErrorDialog::show(const String &p_for_file, const Vector<String> &report) {
-
+void DependencyErrorDialog::show(Mode p_mode, const String &p_for_file, const Vector<String> &report) {
+	mode = p_mode;
 	for_file = p_for_file;
 	set_title(TTR("Error loading:") + " " + p_for_file.get_file());
 	files->clear();
 
-	TreeItem *root = files->create_item(NULL);
+	TreeItem *root = files->create_item(nullptr);
 	for (int i = 0; i < report.size(); i++) {
-
 		String dep;
 		String type = "Object";
 		dep = report[i].get_slice("::", 0);
-		if (report[i].get_slice_count("::") > 0)
+		if (report[i].get_slice_count("::") > 0) {
 			type = report[i].get_slice("::", 1);
-
-		Ref<Texture> icon;
-		if (!has_icon(type, "EditorIcons")) {
-			icon = get_icon("Object", "EditorIcons");
-		} else {
-			icon = get_icon(type, "EditorIcons");
 		}
 
+		Ref<Texture> icon = EditorNode::get_singleton()->get_class_icon(type);
+
 		TreeItem *ti = files->create_item(root);
 		ti->set_text(0, dep);
 		ti->set_icon(0, icon);
 	}
 
-	popup_centered_minsize(Size2(500, 220));
+	popup_centered();
 }
 
 void DependencyErrorDialog::ok_pressed() {
-
-	EditorNode::get_singleton()->load_scene(for_file, true);
+	switch (mode) {
+		case MODE_SCENE:
+			EditorNode::get_singleton()->load_scene(for_file, true);
+			break;
+		case MODE_RESOURCE:
+			EditorNode::get_singleton()->load_resource(for_file, true);
+			break;
+	}
 }
 
 void DependencyErrorDialog::custom_action(const String &) {
-
 	EditorNode::get_singleton()->fix_dependencies(for_file);
 }
 
 DependencyErrorDialog::DependencyErrorDialog() {
-
 	VBoxContainer *vb = memnew(VBoxContainer);
 	add_child(vb);
 
 	files = memnew(Tree);
 	files->set_hide_root(true);
-	vb->add_margin_child(TTR("Scene failed to load due to missing dependencies:"), files, true);
+	vb->add_margin_child(TTR("Load failed due to missing dependencies:"), files, true);
 	files->set_v_size_flags(SIZE_EXPAND_FILL);
-	files->set_custom_minimum_size(Size2(1, 200));
+
+	set_custom_minimum_size(Size2(500, 220) * EDSCALE);
 	get_ok()->set_text(TTR("Open Anyway"));
 	get_cancel()->set_text(TTR("Close"));
 
@@ -611,27 +640,26 @@ DependencyErrorDialog::DependencyErrorDialog() {
 //////////////////////////////////////////////////////////////////////
 
 void OrphanResourcesDialog::ok_pressed() {
-
 	paths.clear();
 
 	_find_to_delete(files->get_root(), paths);
-	if (paths.empty())
+	if (paths.empty()) {
 		return;
+	}
 
 	delete_confirm->set_text(vformat(TTR("Permanently delete %d item(s)? (No undo!)"), paths.size()));
-	delete_confirm->popup_centered_minsize();
+	delete_confirm->popup_centered_clamped(delete_confirm->get_minimum_size());
 }
 
 bool OrphanResourcesDialog::_fill_owners(EditorFileSystemDirectory *efsd, HashMap<String, int> &refs, TreeItem *p_parent) {
-
-	if (!efsd)
+	if (!efsd) {
 		return false;
+	}
 
 	bool has_children = false;
 
 	for (int i = 0; i < efsd->get_subdir_count(); i++) {
-
-		TreeItem *dir_item = NULL;
+		TreeItem *dir_item = nullptr;
 		if (p_parent) {
 			dir_item = files->create_item(p_parent);
 			dir_item->set_text(0, efsd->get_subdir(i)->get_name());
@@ -649,18 +677,14 @@ bool OrphanResourcesDialog::_fill_owners(EditorFileSystemDirectory *efsd, HashMa
 	}
 
 	for (int i = 0; i < efsd->get_file_count(); i++) {
-
 		if (!p_parent) {
 			Vector<String> deps = efsd->get_file_deps(i);
-			//print_line(":::"+efsd->get_file_path(i));
 			for (int j = 0; j < deps.size(); j++) {
-
 				if (!refs.has(deps[j])) {
 					refs[deps[j]] = 1;
 				}
 			}
 		} else {
-
 			String path = efsd->get_file_path(i);
 			if (!refs.has(path)) {
 				TreeItem *ti = files->create_item(p_parent);
@@ -670,17 +694,12 @@ bool OrphanResourcesDialog::_fill_owners(EditorFileSystemDirectory *efsd, HashMa
 
 				String type = efsd->get_file_type(i);
 
-				Ref<Texture> icon;
-				if (has_icon(type, "EditorIcons")) {
-					icon = get_icon(type, "EditorIcons");
-				} else {
-					icon = get_icon("Object", "EditorIcons");
-				}
+				Ref<Texture> icon = EditorNode::get_singleton()->get_class_icon(type);
 				ti->set_icon(0, icon);
 				int ds = efsd->get_file_deps(i).size();
 				ti->set_text(1, itos(ds));
 				if (ds) {
-					ti->add_button(1, get_icon("GuiVisibilityVisible", "EditorIcons"));
+					ti->add_button(1, get_icon("GuiVisibilityVisible", "EditorIcons"), -1, false, TTR("Show Dependencies"));
 				}
 				ti->set_metadata(0, path);
 				has_children = true;
@@ -693,22 +712,19 @@ bool OrphanResourcesDialog::_fill_owners(EditorFileSystemDirectory *efsd, HashMa
 
 void OrphanResourcesDialog::refresh() {
 	HashMap<String, int> refs;
-	_fill_owners(EditorFileSystem::get_singleton()->get_filesystem(), refs, NULL);
+	_fill_owners(EditorFileSystem::get_singleton()->get_filesystem(), refs, nullptr);
 	files->clear();
 	TreeItem *root = files->create_item();
 	_fill_owners(EditorFileSystem::get_singleton()->get_filesystem(), refs, root);
 }
 
 void OrphanResourcesDialog::show() {
-
 	refresh();
 	popup_centered_ratio();
 }
 
 void OrphanResourcesDialog::_find_to_delete(TreeItem *p_item, List<String> &paths) {
-
 	while (p_item) {
-
 		if (p_item->get_cell_mode(0) == TreeItem::CELL_MODE_CHECK && p_item->is_checked(0)) {
 			paths.push_back(p_item->get_metadata(0));
 		}
@@ -722,10 +738,8 @@ void OrphanResourcesDialog::_find_to_delete(TreeItem *p_item, List<String> &path
 }
 
 void OrphanResourcesDialog::_delete_confirm() {
-
 	DirAccess *da = DirAccess::create(DirAccess::ACCESS_RESOURCES);
 	for (List<String>::Element *E = paths.front(); E; E = E->next()) {
-
 		da->remove(E->get());
 		EditorFileSystem::get_singleton()->update_file(E->get());
 	}
@@ -734,7 +748,6 @@ void OrphanResourcesDialog::_delete_confirm() {
 }
 
 void OrphanResourcesDialog::_button_pressed(Object *p_item, int p_column, int p_id) {
-
 	TreeItem *ti = Object::cast_to<TreeItem>(p_item);
 
 	String path = ti->get_metadata(0);
@@ -742,12 +755,19 @@ void OrphanResourcesDialog::_button_pressed(Object *p_item, int p_column, int p_
 }
 
 void OrphanResourcesDialog::_bind_methods() {
-
 	ClassDB::bind_method(D_METHOD("_delete_confirm"), &OrphanResourcesDialog::_delete_confirm);
 	ClassDB::bind_method(D_METHOD("_button_pressed"), &OrphanResourcesDialog::_button_pressed);
 }
 
 OrphanResourcesDialog::OrphanResourcesDialog() {
+	set_title(TTR("Orphan Resource Explorer"));
+	delete_confirm = memnew(ConfirmationDialog);
+	get_ok()->set_text(TTR("Delete"));
+	add_child(delete_confirm);
+	dep_edit = memnew(DependencyEditor);
+	add_child(dep_edit);
+	delete_confirm->connect("confirmed", this, "_delete_confirm");
+	set_hide_on_ok(false);
 
 	VBoxContainer *vbc = memnew(VBoxContainer);
 	add_child(vbc);
@@ -762,14 +782,5 @@ OrphanResourcesDialog::OrphanResourcesDialog() {
 	files->set_column_title(1, TTR("Owns"));
 	files->set_hide_root(true);
 	vbc->add_margin_child(TTR("Resources Without Explicit Ownership:"), files, true);
-	set_title(TTR("Orphan Resource Explorer"));
-	delete_confirm = memnew(ConfirmationDialog);
-	delete_confirm->set_text(TTR("Delete selected files?"));
-	get_ok()->set_text(TTR("Delete"));
-	add_child(delete_confirm);
-	dep_edit = memnew(DependencyEditor);
-	add_child(dep_edit);
 	files->connect("button_pressed", this, "_button_pressed");
-	delete_confirm->connect("confirmed", this, "_delete_confirm");
-	set_hide_on_ok(false);
 }
