@@ -5,8 +5,8 @@
 /*                           GODOT ENGINE                                */
 /*                      https://godotengine.org                          */
 /*************************************************************************/
-/* Copyright (c) 2007-2018 Juan Linietsky, Ariel Manzur.                 */
-/* Copyright (c) 2014-2018 Godot Engine contributors (cf. AUTHORS.md)    */
+/* Copyright (c) 2007-2022 Juan Linietsky, Ariel Manzur.                 */
+/* Copyright (c) 2014-2022 Godot Engine contributors (cf. AUTHORS.md).   */
 /*                                                                       */
 /* Permission is hereby granted, free of charge, to any person obtaining */
 /* a copy of this software and associated documentation files (the       */
@@ -31,17 +31,16 @@
 #include "collision_polygon_editor_plugin.h"
 
 #include "canvas_item_editor_plugin.h"
+#include "core/os/file_access.h"
+#include "core/os/input.h"
+#include "core/os/keyboard.h"
 #include "editor/editor_settings.h"
-#include "os/file_access.h"
 #include "scene/3d/camera.h"
 #include "spatial_editor_plugin.h"
 
-void CollisionPolygonEditor::_notification(int p_what) {
-
+void Polygon3DEditor::_notification(int p_what) {
 	switch (p_what) {
-
 		case NOTIFICATION_READY: {
-
 			button_create->set_icon(get_icon("Edit", "EditorIcons"));
 			button_edit->set_icon(get_icon("MovePoint", "EditorIcons"));
 			button_edit->set_pressed(true);
@@ -53,37 +52,33 @@ void CollisionPolygonEditor::_notification(int p_what) {
 				return;
 			}
 
-			if (node->get_depth() != prev_depth) {
+			if (_get_depth() != prev_depth) {
 				_polygon_draw();
-				prev_depth = node->get_depth();
+				prev_depth = _get_depth();
 			}
 
 		} break;
 	}
 }
-void CollisionPolygonEditor::_node_removed(Node *p_node) {
-
+void Polygon3DEditor::_node_removed(Node *p_node) {
 	if (p_node == node) {
-		node = NULL;
-		if (imgeom->get_parent() == p_node)
+		node = nullptr;
+		if (imgeom->get_parent() == p_node) {
 			p_node->remove_child(imgeom);
+		}
 		hide();
 		set_process(false);
 	}
 }
 
-void CollisionPolygonEditor::_menu_option(int p_option) {
-
+void Polygon3DEditor::_menu_option(int p_option) {
 	switch (p_option) {
-
 		case MODE_CREATE: {
-
 			mode = MODE_CREATE;
 			button_create->set_pressed(true);
 			button_edit->set_pressed(false);
 		} break;
 		case MODE_EDIT: {
-
 			mode = MODE_EDIT;
 			button_create->set_pressed(false);
 			button_edit->set_pressed(true);
@@ -91,10 +86,9 @@ void CollisionPolygonEditor::_menu_option(int p_option) {
 	}
 }
 
-void CollisionPolygonEditor::_wip_close() {
-
-	undo_redo->create_action(TTR("Create Poly3D"));
-	undo_redo->add_undo_method(node, "set_polygon", node->get_polygon());
+void Polygon3DEditor::_wip_close() {
+	undo_redo->create_action(TTR("Create Polygon3D"));
+	undo_redo->add_undo_method(node, "set_polygon", node->call("get_polygon"));
 	undo_redo->add_do_method(node, "set_polygon", wip);
 	undo_redo->add_do_method(this, "_polygon_draw");
 	undo_redo->add_undo_method(this, "_polygon_draw");
@@ -107,67 +101,65 @@ void CollisionPolygonEditor::_wip_close() {
 	undo_redo->commit_action();
 }
 
-bool CollisionPolygonEditor::forward_spatial_gui_input(Camera *p_camera, const Ref<InputEvent> &p_event) {
-
-	if (!node)
+bool Polygon3DEditor::forward_spatial_gui_input(Camera *p_camera, const Ref<InputEvent> &p_event) {
+	if (!node) {
 		return false;
+	}
 
 	Transform gt = node->get_global_transform();
 	Transform gi = gt.affine_inverse();
-	float depth = node->get_depth() * 0.5;
+	float depth = _get_depth() * 0.5;
 	Vector3 n = gt.basis.get_axis(2).normalized();
 	Plane p(gt.origin + n * depth, n);
 
 	Ref<InputEventMouseButton> mb = p_event;
 
 	if (mb.is_valid()) {
-
 		Vector2 gpoint = mb->get_position();
 		Vector3 ray_from = p_camera->project_ray_origin(gpoint);
 		Vector3 ray_dir = p_camera->project_ray_normal(gpoint);
 
 		Vector3 spoint;
 
-		if (!p.intersects_ray(ray_from, ray_dir, &spoint))
+		if (!p.intersects_ray(ray_from, ray_dir, &spoint)) {
 			return false;
+		}
 
 		spoint = gi.xform(spoint);
 
 		Vector2 cpoint(spoint.x, spoint.y);
 
-		cpoint = CanvasItemEditor::get_singleton()->snap_point(cpoint);
+		//DO NOT snap here, it's confusing in 3D for adding points.
+		//Let the snap happen when the point is being moved, instead.
+		//cpoint = CanvasItemEditor::get_singleton()->snap_point(cpoint);
 
-		Vector<Vector2> poly = node->get_polygon();
+		Vector<Vector2> poly = node->call("get_polygon");
 
 		//first check if a point is to be added (segment split)
-		real_t grab_threshold = EDITOR_DEF("editors/poly_editor/point_grab_radius", 8);
+		real_t grab_threshold = EDITOR_GET("editors/poly_editor/point_grab_radius");
 
 		switch (mode) {
-
 			case MODE_CREATE: {
-
 				if (mb->get_button_index() == BUTTON_LEFT && mb->is_pressed()) {
-
 					if (!wip_active) {
-
 						wip.clear();
 						wip.push_back(cpoint);
 						wip_active = true;
 						edited_point_pos = cpoint;
+						snap_ignore = false;
 						_polygon_draw();
 						edited_point = 1;
 						return true;
 					} else {
-
 						if (wip.size() > 1 && p_camera->unproject_position(gt.xform(Vector3(wip[0].x, wip[0].y, depth))).distance_to(gpoint) < grab_threshold) {
 							//wip closed
 							_wip_close();
 
 							return true;
 						} else {
-
 							wip.push_back(cpoint);
 							edited_point = wip.size();
+							snap_ignore = false;
 							_polygon_draw();
 							return true;
 						}
@@ -179,14 +171,10 @@ bool CollisionPolygonEditor::forward_spatial_gui_input(Camera *p_camera, const R
 			} break;
 
 			case MODE_EDIT: {
-
 				if (mb->get_button_index() == BUTTON_LEFT) {
 					if (mb->is_pressed()) {
-
 						if (mb->get_control()) {
-
 							if (poly.size() < 3) {
-
 								undo_redo->create_action(TTR("Edit Poly"));
 								undo_redo->add_undo_method(node, "set_polygon", poly);
 								poly.push_back(cpoint);
@@ -202,15 +190,15 @@ bool CollisionPolygonEditor::forward_spatial_gui_input(Camera *p_camera, const R
 							Vector2 closest_pos;
 							real_t closest_dist = 1e10;
 							for (int i = 0; i < poly.size(); i++) {
-
 								Vector2 points[2] = {
 									p_camera->unproject_position(gt.xform(Vector3(poly[i].x, poly[i].y, depth))),
 									p_camera->unproject_position(gt.xform(Vector3(poly[(i + 1) % poly.size()].x, poly[(i + 1) % poly.size()].y, depth)))
 								};
 
 								Vector2 cp = Geometry::get_closest_point_to_segment_2d(gpoint, points);
-								if (cp.distance_squared_to(points[0]) < CMP_EPSILON2 || cp.distance_squared_to(points[1]) < CMP_EPSILON2)
+								if (cp.distance_squared_to(points[0]) < CMP_EPSILON2 || cp.distance_squared_to(points[1]) < CMP_EPSILON2) {
 									continue; //not valid to reuse point
+								}
 
 								real_t d = cp.distance_to(gpoint);
 								if (d < closest_dist && d < grab_threshold) {
@@ -221,24 +209,23 @@ bool CollisionPolygonEditor::forward_spatial_gui_input(Camera *p_camera, const R
 							}
 
 							if (closest_idx >= 0) {
-
 								pre_move_edit = poly;
 								poly.insert(closest_idx + 1, cpoint);
 								edited_point = closest_idx + 1;
 								edited_point_pos = cpoint;
-								node->set_polygon(poly);
+								node->call("set_polygon", poly);
 								_polygon_draw();
+								snap_ignore = true;
+
 								return true;
 							}
 						} else {
-
 							//look for points to move
 
 							int closest_idx = -1;
 							Vector2 closest_pos;
 							real_t closest_dist = 1e10;
 							for (int i = 0; i < poly.size(); i++) {
-
 								Vector2 cp = p_camera->unproject_position(gt.xform(Vector3(poly[i].x, poly[i].y, depth)));
 
 								real_t d = cp.distance_to(gpoint);
@@ -250,22 +237,22 @@ bool CollisionPolygonEditor::forward_spatial_gui_input(Camera *p_camera, const R
 							}
 
 							if (closest_idx >= 0) {
-
 								pre_move_edit = poly;
 								edited_point = closest_idx;
 								edited_point_pos = poly[closest_idx];
 								_polygon_draw();
+								snap_ignore = false;
 								return true;
 							}
 						}
 					} else {
+						snap_ignore = false;
 
 						if (edited_point != -1) {
-
 							//apply
 
 							ERR_FAIL_INDEX_V(edited_point, poly.size(), false);
-							poly[edited_point] = edited_point_pos;
+							poly.write[edited_point] = edited_point_pos;
 							undo_redo->create_action(TTR("Edit Poly"));
 							undo_redo->add_do_method(node, "set_polygon", poly);
 							undo_redo->add_undo_method(node, "set_polygon", pre_move_edit);
@@ -279,12 +266,10 @@ bool CollisionPolygonEditor::forward_spatial_gui_input(Camera *p_camera, const R
 					}
 				}
 				if (mb->get_button_index() == BUTTON_RIGHT && mb->is_pressed() && edited_point == -1) {
-
 					int closest_idx = -1;
 					Vector2 closest_pos;
 					real_t closest_dist = 1e10;
 					for (int i = 0; i < poly.size(); i++) {
-
 						Vector2 cp = p_camera->unproject_position(gt.xform(Vector3(poly[i].x, poly[i].y, depth)));
 
 						real_t d = cp.distance_to(gpoint);
@@ -296,7 +281,6 @@ bool CollisionPolygonEditor::forward_spatial_gui_input(Camera *p_camera, const R
 					}
 
 					if (closest_idx >= 0) {
-
 						undo_redo->create_action(TTR("Edit Poly (Remove Point)"));
 						undo_redo->add_undo_method(node, "set_polygon", poly);
 						poly.remove(closest_idx);
@@ -315,9 +299,7 @@ bool CollisionPolygonEditor::forward_spatial_gui_input(Camera *p_camera, const R
 	Ref<InputEventMouseMotion> mm = p_event;
 
 	if (mm.is_valid()) {
-
 		if (edited_point != -1 && (wip_active || mm->get_button_mask() & BUTTON_MASK_LEFT)) {
-
 			Vector2 gpoint = mm->get_position();
 
 			Vector3 ray_from = p_camera->project_ray_origin(gpoint);
@@ -325,14 +307,23 @@ bool CollisionPolygonEditor::forward_spatial_gui_input(Camera *p_camera, const R
 
 			Vector3 spoint;
 
-			if (!p.intersects_ray(ray_from, ray_dir, &spoint))
+			if (!p.intersects_ray(ray_from, ray_dir, &spoint)) {
 				return false;
+			}
 
 			spoint = gi.xform(spoint);
 
 			Vector2 cpoint(spoint.x, spoint.y);
 
-			cpoint = CanvasItemEditor::get_singleton()->snap_point(cpoint);
+			if (snap_ignore && !Input::get_singleton()->is_key_pressed(KEY_CONTROL)) {
+				snap_ignore = false;
+			}
+
+			if (!snap_ignore && SpatialEditor::get_singleton()->is_snap_enabled()) {
+				cpoint = cpoint.snapped(Vector2(
+						SpatialEditor::get_singleton()->get_translate_snap(),
+						SpatialEditor::get_singleton()->get_translate_snap()));
+			}
 			edited_point_pos = cpoint;
 
 			_polygon_draw();
@@ -341,19 +332,29 @@ bool CollisionPolygonEditor::forward_spatial_gui_input(Camera *p_camera, const R
 
 	return false;
 }
-void CollisionPolygonEditor::_polygon_draw() {
 
-	if (!node)
+float Polygon3DEditor::_get_depth() {
+	if (bool(node->call("_has_editable_3d_polygon_no_depth"))) {
+		return 0;
+	}
+
+	return float(node->call("get_depth"));
+}
+
+void Polygon3DEditor::_polygon_draw() {
+	if (!node) {
 		return;
+	}
 
 	Vector<Vector2> poly;
 
-	if (wip_active)
+	if (wip_active) {
 		poly = wip;
-	else
-		poly = node->get_polygon();
+	} else {
+		poly = node->call("get_polygon");
+	}
 
-	float depth = node->get_depth() * 0.5;
+	float depth = _get_depth() * 0.5;
 
 	imgeom->clear();
 	imgeom->set_material_override(line_material);
@@ -362,18 +363,19 @@ void CollisionPolygonEditor::_polygon_draw() {
 	Rect2 rect;
 
 	for (int i = 0; i < poly.size(); i++) {
-
 		Vector2 p, p2;
 		p = i == edited_point ? edited_point_pos : poly[i];
-		if ((wip_active && i == poly.size() - 1) || (((i + 1) % poly.size()) == edited_point))
+		if ((wip_active && i == poly.size() - 1) || (((i + 1) % poly.size()) == edited_point)) {
 			p2 = edited_point_pos;
-		else
+		} else {
 			p2 = poly[(i + 1) % poly.size()];
+		}
 
-		if (i == 0)
+		if (i == 0) {
 			rect.position = p;
-		else
+		} else {
 			rect.expand_to(p);
+		}
 
 		Vector3 point = Vector3(p.x, p.y, depth);
 		Vector3 next_point = Vector3(p2.x, p2.y, depth);
@@ -440,18 +442,17 @@ void CollisionPolygonEditor::_polygon_draw() {
 		m->surface_remove(0);
 	}
 
-	if (poly.size() == 0)
+	if (poly.size() == 0) {
 		return;
+	}
 
 	Array a;
 	a.resize(Mesh::ARRAY_MAX);
 	PoolVector<Vector3> va;
 	{
-
 		va.resize(poly.size());
 		PoolVector<Vector3>::Write w = va.write();
 		for (int i = 0; i < poly.size(); i++) {
-
 			Vector2 p, p2;
 			p = i == edited_point ? edited_point_pos : poly[i];
 
@@ -464,13 +465,11 @@ void CollisionPolygonEditor::_polygon_draw() {
 	m->surface_set_material(0, handle_material);
 }
 
-void CollisionPolygonEditor::edit(Node *p_collision_polygon) {
-
+void Polygon3DEditor::edit(Node *p_collision_polygon) {
 	if (p_collision_polygon) {
-
-		node = Object::cast_to<CollisionPolygon>(p_collision_polygon);
+		node = Object::cast_to<Spatial>(p_collision_polygon);
 		//Enable the pencil tool if the polygon is empty
-		if (node->get_polygon().size() == 0) {
+		if (Vector<Vector2>(node->call("get_polygon")).size() == 0) {
 			_menu_option(MODE_CREATE);
 		}
 		wip.clear();
@@ -482,27 +481,26 @@ void CollisionPolygonEditor::edit(Node *p_collision_polygon) {
 		prev_depth = -1;
 
 	} else {
-		node = NULL;
+		node = nullptr;
 
-		if (imgeom->get_parent())
+		if (imgeom->get_parent()) {
 			imgeom->get_parent()->remove_child(imgeom);
+		}
 
 		set_process(false);
 	}
 }
 
-void CollisionPolygonEditor::_bind_methods() {
-
-	ClassDB::bind_method(D_METHOD("_menu_option"), &CollisionPolygonEditor::_menu_option);
-	ClassDB::bind_method(D_METHOD("_polygon_draw"), &CollisionPolygonEditor::_polygon_draw);
-	ClassDB::bind_method(D_METHOD("_node_removed"), &CollisionPolygonEditor::_node_removed);
+void Polygon3DEditor::_bind_methods() {
+	ClassDB::bind_method(D_METHOD("_menu_option"), &Polygon3DEditor::_menu_option);
+	ClassDB::bind_method(D_METHOD("_polygon_draw"), &Polygon3DEditor::_polygon_draw);
+	ClassDB::bind_method(D_METHOD("_node_removed"), &Polygon3DEditor::_node_removed);
 }
 
-CollisionPolygonEditor::CollisionPolygonEditor(EditorNode *p_editor) {
-
-	node = NULL;
+Polygon3DEditor::Polygon3DEditor(EditorNode *p_editor) {
+	node = nullptr;
 	editor = p_editor;
-	undo_redo = editor->get_undo_redo();
+	undo_redo = EditorNode::get_undo_redo();
 
 	add_child(memnew(VSeparator));
 	button_create = memnew(ToolButton);
@@ -543,42 +541,38 @@ CollisionPolygonEditor::CollisionPolygonEditor(EditorNode *p_editor) {
 	m.instance();
 	pointsm->set_mesh(m);
 	pointsm->set_transform(Transform(Basis(), Vector3(0, 0, 0.00001)));
+
+	snap_ignore = false;
 }
 
-CollisionPolygonEditor::~CollisionPolygonEditor() {
-
+Polygon3DEditor::~Polygon3DEditor() {
 	memdelete(imgeom);
 }
 
-void CollisionPolygonEditorPlugin::edit(Object *p_object) {
-
+void Polygon3DEditorPlugin::edit(Object *p_object) {
 	collision_polygon_editor->edit(Object::cast_to<Node>(p_object));
 }
 
-bool CollisionPolygonEditorPlugin::handles(Object *p_object) const {
-
-	return p_object->is_class("CollisionPolygon");
+bool Polygon3DEditorPlugin::handles(Object *p_object) const {
+	return Object::cast_to<Spatial>(p_object) && bool(p_object->call("_is_editable_3d_polygon"));
 }
 
-void CollisionPolygonEditorPlugin::make_visible(bool p_visible) {
-
+void Polygon3DEditorPlugin::make_visible(bool p_visible) {
 	if (p_visible) {
 		collision_polygon_editor->show();
 	} else {
-
 		collision_polygon_editor->hide();
-		collision_polygon_editor->edit(NULL);
+		collision_polygon_editor->edit(nullptr);
 	}
 }
 
-CollisionPolygonEditorPlugin::CollisionPolygonEditorPlugin(EditorNode *p_node) {
-
+Polygon3DEditorPlugin::Polygon3DEditorPlugin(EditorNode *p_node) {
 	editor = p_node;
-	collision_polygon_editor = memnew(CollisionPolygonEditor(p_node));
+	collision_polygon_editor = memnew(Polygon3DEditor(p_node));
 	SpatialEditor::get_singleton()->add_control_to_menu_panel(collision_polygon_editor);
 
 	collision_polygon_editor->hide();
 }
 
-CollisionPolygonEditorPlugin::~CollisionPolygonEditorPlugin() {
+Polygon3DEditorPlugin::~Polygon3DEditorPlugin() {
 }
