@@ -246,12 +246,12 @@ void ProjectExportDialog::_edit_preset(int p_index) {
 				}
 				error += " - " + items[i];
 			}
-
 			export_error->set_text(error);
 			export_error->show();
 		} else {
 			export_error->hide();
 		}
+		export_warning->hide();
 		if (needs_templates) {
 			export_templates_error->show();
 		} else {
@@ -262,6 +262,20 @@ void ProjectExportDialog::_edit_preset(int p_index) {
 		get_ok()->set_disabled(true);
 
 	} else {
+		if (error != String()) {
+			Vector<String> items = error.split("\n", false);
+			error = "";
+			for (int i = 0; i < items.size(); i++) {
+				if (i > 0) {
+					error += "\n";
+				}
+				error += " - " + items[i];
+			}
+			export_warning->set_text(error);
+			export_warning->show();
+		} else {
+			export_warning->hide();
+		}
 		export_error->hide();
 		export_templates_error->hide();
 		export_button->set_disabled(false);
@@ -1144,6 +1158,11 @@ ProjectExportDialog::ProjectExportDialog() {
 	export_error->hide();
 	export_error->add_color_override("font_color", EditorNode::get_singleton()->get_gui_base()->get_color("error_color", "Editor"));
 
+	export_warning = memnew(Label);
+	main_vb->add_child(export_warning);
+	export_warning->hide();
+	export_warning->add_color_override("font_color", EditorNode::get_singleton()->get_gui_base()->get_color("warning_color", "Editor"));
+
 	export_templates_error = memnew(HBoxContainer);
 	main_vb->add_child(export_templates_error);
 	export_templates_error->hide();
