@@ -5,8 +5,8 @@
 /*                           GODOT ENGINE                                */
 /*                      https://godotengine.org                          */
 /*************************************************************************/
-/* Copyright (c) 2007-2018 Juan Linietsky, Ariel Manzur.                 */
-/* Copyright (c) 2014-2018 Godot Engine contributors (cf. AUTHORS.md)    */
+/* Copyright (c) 2007-2022 Juan Linietsky, Ariel Manzur.                 */
+/* Copyright (c) 2014-2022 Godot Engine contributors (cf. AUTHORS.md).   */
 /*                                                                       */
 /* Permission is hereby granted, free of charge, to any person obtaining */
 /* a copy of this software and associated documentation files (the       */
@@ -191,6 +191,8 @@ uint8_t GDAPI godot_pool_byte_array_get(const godot_pool_byte_array *p_self, con
 
 godot_int GDAPI godot_pool_byte_array_size(const godot_pool_byte_array *p_self);
 
+godot_bool GDAPI godot_pool_byte_array_empty(const godot_pool_byte_array *p_self);
+
 void GDAPI godot_pool_byte_array_destroy(godot_pool_byte_array *p_self);
 
 // int
@@ -222,6 +224,8 @@ godot_int GDAPI godot_pool_int_array_get(const godot_pool_int_array *p_self, con
 
 godot_int GDAPI godot_pool_int_array_size(const godot_pool_int_array *p_self);
 
+godot_bool GDAPI godot_pool_int_array_empty(const godot_pool_int_array *p_self);
+
 void GDAPI godot_pool_int_array_destroy(godot_pool_int_array *p_self);
 
 // real
@@ -253,6 +257,8 @@ godot_real GDAPI godot_pool_real_array_get(const godot_pool_real_array *p_self,
 
 godot_int GDAPI godot_pool_real_array_size(const godot_pool_real_array *p_self);
 
+godot_bool GDAPI godot_pool_real_array_empty(const godot_pool_real_array *p_self);
+
 void GDAPI godot_pool_real_array_destroy(godot_pool_real_array *p_self);
 
 // string
@@ -284,6 +290,8 @@ godot_string GDAPI godot_pool_string_array_get(const godot_pool_string_array *p_
 
 godot_int GDAPI godot_pool_string_array_size(const godot_pool_string_array *p_self);
 
+godot_bool GDAPI godot_pool_string_array_empty(const godot_pool_string_array *p_self);
+
 void GDAPI godot_pool_string_array_destroy(godot_pool_string_array *p_self);
 
 // vector2
@@ -315,6 +323,8 @@ godot_vector2 GDAPI godot_pool_vector2_array_get(const godot_pool_vector2_array
 
 godot_int GDAPI godot_pool_vector2_array_size(const godot_pool_vector2_array *p_self);
 
+godot_bool GDAPI godot_pool_vector2_array_empty(const godot_pool_vector2_array *p_self);
+
 void GDAPI godot_pool_vector2_array_destroy(godot_pool_vector2_array *p_self);
 
 // vector3
@@ -346,6 +356,8 @@ godot_vector3 GDAPI godot_pool_vector3_array_get(const godot_pool_vector3_array
 
 godot_int GDAPI godot_pool_vector3_array_size(const godot_pool_vector3_array *p_self);
 
+godot_bool GDAPI godot_pool_vector3_array_empty(const godot_pool_vector3_array *p_self);
+
 void GDAPI godot_pool_vector3_array_destroy(godot_pool_vector3_array *p_self);
 
 // color
@@ -377,6 +389,8 @@ godot_color GDAPI godot_pool_color_array_get(const godot_pool_color_array *p_sel
 
 godot_int GDAPI godot_pool_color_array_size(const godot_pool_color_array *p_self);
 
+godot_bool GDAPI godot_pool_color_array_empty(const godot_pool_color_array *p_self);
+
 void GDAPI godot_pool_color_array_destroy(godot_pool_color_array *p_self);
 
 //
