@@ -5,8 +5,8 @@
 /*                           GODOT ENGINE                                */
 /*                      https://godotengine.org                          */
 /*************************************************************************/
-/* Copyright (c) 2007-2018 Juan Linietsky, Ariel Manzur.                 */
-/* Copyright (c) 2014-2018 Godot Engine contributors (cf. AUTHORS.md)    */
+/* Copyright (c) 2007-2022 Juan Linietsky, Ariel Manzur.                 */
+/* Copyright (c) 2014-2022 Godot Engine contributors (cf. AUTHORS.md).   */
 /*                                                                       */
 /* Permission is hereby granted, free of charge, to any person obtaining */
 /* a copy of this software and associated documentation files (the       */
@@ -35,18 +35,17 @@
 #include "core/class_db.h"
 #include "core/engine.h"
 #include "core/global_constants.h"
+#include "core/os/file_access.h"
 #include "core/pair.h"
-#include "os/file_access.h"
 
 // helper stuff
 
 static Error save_file(const String &p_path, const List<String> &p_content) {
-
 	FileAccessRef file = FileAccess::open(p_path, FileAccess::WRITE);
 
 	ERR_FAIL_COND_V(!file, ERR_FILE_CANT_WRITE);
 
-	for (const List<String>::Element *e = p_content.front(); e != NULL; e = e->next()) {
+	for (const List<String>::Element *e = p_content.front(); e != nullptr; e = e->next()) {
 		file->store_string(e->get());
 	}
 
@@ -98,7 +97,7 @@ struct SignalAPI {
 
 struct EnumAPI {
 	String name;
-	List<Pair<int, String> > values;
+	List<Pair<int, String>> values;
 };
 
 struct ClassAPI {
@@ -108,9 +107,9 @@ struct ClassAPI {
 	ClassDB::APIType api_type;
 
 	bool is_singleton;
+	String singleton_name;
 	bool is_instanciable;
 	// @Unclear
-	bool is_creatable;
 	bool is_reference;
 
 	List<MethodAPI> methods;
@@ -139,15 +138,40 @@ static String get_type_name(const PropertyInfo &info) {
 	return Variant::get_type_name(info.type);
 }
 
+/*
+ * Some comparison helper functions we need
+ */
+
+struct MethodInfoComparator {
+	StringName::AlphCompare compare;
+	bool operator()(const MethodInfo &p_a, const MethodInfo &p_b) const {
+		return compare(p_a.name, p_b.name);
+	}
+};
+
+struct PropertyInfoComparator {
+	StringName::AlphCompare compare;
+	bool operator()(const PropertyInfo &p_a, const PropertyInfo &p_b) const {
+		return compare(p_a.name, p_b.name);
+	}
+};
+
+struct ConstantAPIComparator {
+	NoCaseComparator compare;
+	bool operator()(const ConstantAPI &p_a, const ConstantAPI &p_b) const {
+		return compare(p_a.constant_name, p_b.constant_name);
+	}
+};
+
 /*
  * Reads the entire Godot API to a list
  */
 List<ClassAPI> generate_c_api_classes() {
-
 	List<ClassAPI> api;
 
 	List<StringName> classes;
 	ClassDB::get_class_list(&classes);
+	classes.sort_custom<StringName::AlphCompare>();
 
 	// Register global constants as a fake GlobalConstants singleton class
 	{
@@ -155,6 +179,7 @@ List<ClassAPI> generate_c_api_classes() {
 		global_constants_api.class_name = L"GlobalConstants";
 		global_constants_api.api_type = ClassDB::API_CORE;
 		global_constants_api.is_singleton = true;
+		global_constants_api.singleton_name = L"GlobalConstants";
 		global_constants_api.is_instanciable = false;
 		const int constants_count = GlobalConstants::get_global_constant_count();
 		for (int i = 0; i < constants_count; ++i) {
@@ -163,10 +188,11 @@ List<ClassAPI> generate_c_api_classes() {
 			constant_api.constant_value = GlobalConstants::get_global_constant_value(i);
 			global_constants_api.constants.push_back(constant_api);
 		}
+		global_constants_api.constants.sort_custom<ConstantAPIComparator>();
 		api.push_back(global_constants_api);
 	}
 
-	for (List<StringName>::Element *e = classes.front(); e != NULL; e = e->next()) {
+	for (List<StringName>::Element *e = classes.front(); e != nullptr; e = e->next()) {
 		StringName class_name = e->get();
 
 		ClassAPI class_api;
@@ -179,13 +205,16 @@ List<ClassAPI> generate_c_api_classes() {
 				name.remove(0);
 			}
 			class_api.is_singleton = Engine::get_singleton()->has_singleton(name);
+			if (class_api.is_singleton) {
+				class_api.singleton_name = name;
+			}
 		}
 		class_api.is_instanciable = !class_api.is_singleton && ClassDB::can_instance(class_name);
 
 		{
 			List<StringName> inheriters;
 			ClassDB::get_inheriters_from_class("Reference", &inheriters);
-			bool is_reference = !!inheriters.find(class_name);
+			bool is_reference = !!inheriters.find(class_name) || class_name == "Reference";
 			// @Unclear
 			class_api.is_reference = !class_api.is_singleton && is_reference;
 		}
@@ -194,7 +223,8 @@ List<ClassAPI> generate_c_api_classes() {
 		{
 			List<String> constant;
 			ClassDB::get_integer_constant_list(class_name, &constant, true);
-			for (List<String>::Element *c = constant.front(); c != NULL; c = c->next()) {
+			constant.sort_custom<NoCaseComparator>();
+			for (List<String>::Element *c = constant.front(); c != nullptr; c = c->next()) {
 				ConstantAPI constant_api;
 				constant_api.constant_name = c->get();
 				constant_api.constant_value = ClassDB::get_integer_constant(class_name, c->get());
@@ -207,6 +237,7 @@ List<ClassAPI> generate_c_api_classes() {
 		{
 			List<MethodInfo> signals_;
 			ClassDB::get_signal_list(class_name, &signals_, true);
+			signals_.sort_custom<MethodInfoComparator>();
 
 			for (int i = 0; i < signals_.size(); i++) {
 				SignalAPI signal;
@@ -246,8 +277,9 @@ List<ClassAPI> generate_c_api_classes() {
 		{
 			List<PropertyInfo> properties;
 			ClassDB::get_property_list(class_name, &properties, true);
+			properties.sort_custom<PropertyInfoComparator>();
 
-			for (List<PropertyInfo>::Element *p = properties.front(); p != NULL; p = p->next()) {
+			for (List<PropertyInfo>::Element *p = properties.front(); p != nullptr; p = p->next()) {
 				PropertyAPI property_api;
 
 				property_api.name = p->get().name;
@@ -273,14 +305,15 @@ List<ClassAPI> generate_c_api_classes() {
 		{
 			List<MethodInfo> methods;
 			ClassDB::get_method_list(class_name, &methods, true);
+			methods.sort_custom<MethodInfoComparator>();
 
-			for (List<MethodInfo>::Element *m = methods.front(); m != NULL; m = m->next()) {
+			for (List<MethodInfo>::Element *m = methods.front(); m != nullptr; m = m->next()) {
 				MethodAPI method_api;
 				MethodBind *method_bind = ClassDB::get_method(class_name, m->get().name);
 				MethodInfo &method_info = m->get();
 
 				//method name
-				method_api.method_name = m->get().name;
+				method_api.method_name = method_info.name;
 				//method return type
 				if (method_api.method_name.find(":") != -1) {
 					method_api.return_type = method_api.method_name.get_slice(":", 1);
@@ -293,6 +326,7 @@ List<ClassAPI> generate_c_api_classes() {
 				method_api.has_varargs = method_bind && method_bind->is_vararg();
 
 				// Method flags
+				method_api.is_virtual = false;
 				if (method_info.flags) {
 					const uint32_t flags = method_info.flags;
 					method_api.is_editor = flags & METHOD_FLAG_EDITOR;
@@ -321,6 +355,11 @@ List<ClassAPI> generate_c_api_classes() {
 						arg_type = arg_info.hint_string;
 					} else if (arg_info.type == Variant::NIL) {
 						arg_type = "Variant";
+					} else if (arg_info.type == Variant::OBJECT) {
+						arg_type = arg_info.class_name;
+						if (arg_type == "") {
+							arg_type = Variant::get_type_name(arg_info.type);
+						}
 					} else {
 						arg_type = Variant::get_type_name(arg_info.type);
 					}
@@ -348,10 +387,10 @@ List<ClassAPI> generate_c_api_classes() {
 				enum_api.name = E->get();
 				ClassDB::get_enum_constants(class_name, E->get(), &value_names, true);
 				for (List<StringName>::Element *val_e = value_names.front(); val_e; val_e = val_e->next()) {
-					int int_val = ClassDB::get_integer_constant(class_name, val_e->get(), NULL);
+					int int_val = ClassDB::get_integer_constant(class_name, val_e->get(), nullptr);
 					enum_api.values.push_back(Pair<int, String>(int_val, val_e->get()));
 				}
-				enum_api.values.sort_custom<PairSort<int, String> >();
+				enum_api.values.sort_custom<PairSort<int, String>>();
 				class_api.enums.push_back(enum_api);
 			}
 		}
@@ -366,14 +405,13 @@ List<ClassAPI> generate_c_api_classes() {
  * Generates the JSON source from the API in p_api
  */
 static List<String> generate_c_api_json(const List<ClassAPI> &p_api) {
-
 	// I'm sorry for the \t mess
 
 	List<String> source;
 
 	source.push_back("[\n");
 
-	for (const List<ClassAPI>::Element *c = p_api.front(); c != NULL; c = c->next()) {
+	for (const List<ClassAPI>::Element *c = p_api.front(); c != nullptr; c = c->next()) {
 		ClassAPI api = c->get();
 
 		source.push_back("\t{\n");
@@ -382,10 +420,10 @@ static List<String> generate_c_api_json(const List<ClassAPI> &p_api) {
 		source.push_back("\t\t\"base_class\": \"" + api.super_class_name + "\",\n");
 		source.push_back(String("\t\t\"api_type\": \"") + (api.api_type == ClassDB::API_CORE ? "core" : (api.api_type == ClassDB::API_EDITOR ? "tools" : "none")) + "\",\n");
 		source.push_back(String("\t\t\"singleton\": ") + (api.is_singleton ? "true" : "false") + ",\n");
+		source.push_back("\t\t\"singleton_name\": \"" + api.singleton_name + "\",\n");
 		source.push_back(String("\t\t\"instanciable\": ") + (api.is_instanciable ? "true" : "false") + ",\n");
 		source.push_back(String("\t\t\"is_reference\": ") + (api.is_reference ? "true" : "false") + ",\n");
 		// @Unclear
-		// source.push_back(String("\t\t\"createable\": ") + (api.is_creatable ? "true" : "false") + ",\n");
 
 		source.push_back("\t\t\"constants\": {\n");
 		for (List<ConstantAPI>::Element *e = api.constants.front(); e; e = e->next()) {
@@ -414,6 +452,7 @@ static List<String> generate_c_api_json(const List<ClassAPI> &p_api) {
 				source.push_back("\t\t\t\t\t{\n");
 				source.push_back("\t\t\t\t\t\t\"name\": \"" + e->get().argument_names[i] + "\",\n");
 				source.push_back("\t\t\t\t\t\t\"type\": \"" + e->get().argument_types[i] + "\",\n");
+				source.push_back(String("\t\t\t\t\t\t\"has_default_value\": ") + (e->get().default_arguments.has(i) ? "true" : "false") + ",\n");
 				source.push_back("\t\t\t\t\t\t\"default_value\": \"" + (e->get().default_arguments.has(i) ? (String)e->get().default_arguments[i] : "") + "\"\n");
 				source.push_back(String("\t\t\t\t\t}") + ((i < e->get().argument_names.size() - 1) ? "," : "") + "\n");
 			}
@@ -453,7 +492,7 @@ static List<String> generate_c_api_json(const List<ClassAPI> &p_api) {
 			source.push_back("\t\t\t{\n");
 			source.push_back("\t\t\t\t\"name\": \"" + e->get().name + "\",\n");
 			source.push_back("\t\t\t\t\"values\": {\n");
-			for (List<Pair<int, String> >::Element *val_e = e->get().values.front(); val_e; val_e = val_e->next()) {
+			for (List<Pair<int, String>>::Element *val_e = e->get().values.front(); val_e; val_e = val_e->next()) {
 				source.push_back("\t\t\t\t\t\"" + val_e->get().second + "\": " + itos(val_e->get().first));
 				source.push_back(String((val_e->next() ? "," : "")) + "\n");
 			}
@@ -476,7 +515,6 @@ static List<String> generate_c_api_json(const List<ClassAPI> &p_api) {
  *  p_path
  */
 Error generate_c_api(const String &p_path) {
-
 #ifndef TOOLS_ENABLED
 	return ERR_BUG;
 #else
