@@ -5,8 +5,8 @@
 /*                           GODOT ENGINE                                */
 /*                      https://godotengine.org                          */
 /*************************************************************************/
-/* Copyright (c) 2007-2018 Juan Linietsky, Ariel Manzur.                 */
-/* Copyright (c) 2014-2018 Godot Engine contributors (cf. AUTHORS.md)    */
+/* Copyright (c) 2007-2022 Juan Linietsky, Ariel Manzur.                 */
+/* Copyright (c) 2014-2022 Godot Engine contributors (cf. AUTHORS.md).   */
 /*                                                                       */
 /* Permission is hereby granted, free of charge, to any person obtaining */
 /* a copy of this software and associated documentation files (the       */
@@ -29,50 +29,28 @@
 /*************************************************************************/
 
 #include "grid_map.h"
-#include "message_queue.h"
+
+#include "core/io/marshalls.h"
+#include "core/message_queue.h"
 #include "scene/3d/light.h"
+#include "scene/resources/mesh_library.h"
 #include "scene/resources/surface_tool.h"
+#include "scene/scene_string_names.h"
 #include "servers/visual_server.h"
 
-#include "io/marshalls.h"
-#include "os/os.h"
-#include "scene/resources/mesh_library.h"
-#include "scene/scene_string_names.h"
-
 bool GridMap::_set(const StringName &p_name, const Variant &p_value) {
-
 	String name = p_name;
 
-	/*	} else if (name=="cells") {
-		PoolVector<int> cells = p_value;
-		int amount=cells.size();
-		PoolVector<int>::Read r = cells.read();
-		ERR_FAIL_COND_V(amount&1,false); // not even
-		cell_map.clear();
-		for(int i=0;i<amount/3;i++) {
-
-
-			IndexKey ik;
-			ik.key=decode_uint64(&r[i*3]);
-			Cell cell;
-			cell.cell=uint32_t(r[i*+1]);
-			cell_map[ik]=cell;
-
-		}
-		_recreate_octant_data();*/
 	if (name == "data") {
-
 		Dictionary d = p_value;
 
 		if (d.has("cells")) {
-
 			PoolVector<int> cells = d["cells"];
 			int amount = cells.size();
 			PoolVector<int>::Read r = cells.read();
 			ERR_FAIL_COND_V(amount % 3, false); // not even
 			cell_map.clear();
 			for (int i = 0; i < amount / 3; i++) {
-
 				IndexKey ik;
 				ik.key = decode_uint64((const uint8_t *)&r[i * 3]);
 				Cell cell;
@@ -80,9 +58,10 @@ bool GridMap::_set(const StringName &p_name, const Variant &p_value) {
 				cell_map[ik] = cell;
 			}
 		}
-		_recreate_octant_data();
-	} else if (name == "baked_meshes") {
 
+		_recreate_octant_data();
+
+	} else if (name == "baked_meshes") {
 		clear_baked_meshes();
 
 		Array meshes = p_value;
@@ -103,18 +82,17 @@ bool GridMap::_set(const StringName &p_name, const Variant &p_value) {
 
 		_recreate_octant_data();
 
-	} else
+	} else {
 		return false;
+	}
 
 	return true;
 }
 
 bool GridMap::_get(const StringName &p_name, Variant &r_ret) const {
-
 	String name = p_name;
 
 	if (name == "data") {
-
 		Dictionary d;
 
 		PoolVector<int> cells;
@@ -123,7 +101,6 @@ bool GridMap::_get(const StringName &p_name, Variant &r_ret) const {
 			PoolVector<int>::Write w = cells.write();
 			int i = 0;
 			for (Map<IndexKey, Cell>::Element *E = cell_map.front(); E; E = E->next(), i++) {
-
 				encode_uint64(E->key().key, (uint8_t *)&w[i * 3]);
 				encode_uint32(E->get().cell, (uint8_t *)&w[i * 3 + 2]);
 			}
@@ -133,22 +110,21 @@ bool GridMap::_get(const StringName &p_name, Variant &r_ret) const {
 
 		r_ret = d;
 	} else if (name == "baked_meshes") {
-
 		Array ret;
 		ret.resize(baked_meshes.size());
 		for (int i = 0; i < baked_meshes.size(); i++) {
-			ret.push_back(baked_meshes[i].mesh);
+			ret[i] = baked_meshes[i].mesh;
 		}
 		r_ret = ret;
 
-	} else
+	} else {
 		return false;
+	}
 
 	return true;
 }
 
 void GridMap::_get_property_list(List<PropertyInfo> *p_list) const {
-
 	if (baked_meshes.size()) {
 		p_list->push_back(PropertyInfo(Variant::ARRAY, "baked_meshes", PROPERTY_HINT_NONE, "", PROPERTY_USAGE_STORAGE));
 	}
@@ -157,97 +133,100 @@ void GridMap::_get_property_list(List<PropertyInfo> *p_list) const {
 }
 
 void GridMap::set_collision_layer(uint32_t p_layer) {
-
 	collision_layer = p_layer;
 	_reset_physic_bodies_collision_filters();
 }
 
 uint32_t GridMap::get_collision_layer() const {
-
 	return collision_layer;
 }
 
 void GridMap::set_collision_mask(uint32_t p_mask) {
-
 	collision_mask = p_mask;
 	_reset_physic_bodies_collision_filters();
 }
 
 uint32_t GridMap::get_collision_mask() const {
-
 	return collision_mask;
 }
 
 void GridMap::set_collision_mask_bit(int p_bit, bool p_value) {
-
+	ERR_FAIL_INDEX_MSG(p_bit, 32, "Collision mask bit must be between 0 and 31 inclusive.");
 	uint32_t mask = get_collision_mask();
-	if (p_value)
+	if (p_value) {
 		mask |= 1 << p_bit;
-	else
+	} else {
 		mask &= ~(1 << p_bit);
+	}
 	set_collision_mask(mask);
 }
 
 bool GridMap::get_collision_mask_bit(int p_bit) const {
-
+	ERR_FAIL_INDEX_V_MSG(p_bit, 32, false, "Collision mask bit must be between 0 and 31 inclusive.");
 	return get_collision_mask() & (1 << p_bit);
 }
 
 void GridMap::set_collision_layer_bit(int p_bit, bool p_value) {
-
-	uint32_t mask = get_collision_layer();
-	if (p_value)
-		mask |= 1 << p_bit;
-	else
-		mask &= ~(1 << p_bit);
-	set_collision_layer(mask);
+	ERR_FAIL_INDEX_MSG(p_bit, 32, "Collision layer bit must be between 0 and 31 inclusive.");
+	uint32_t layer = get_collision_layer();
+	if (p_value) {
+		layer |= 1 << p_bit;
+	} else {
+		layer &= ~(1 << p_bit);
+	}
+	set_collision_layer(layer);
 }
 
 bool GridMap::get_collision_layer_bit(int p_bit) const {
-
+	ERR_FAIL_INDEX_V_MSG(p_bit, 32, false, "Collision layer bit must be between 0 and 31 inclusive.");
 	return get_collision_layer() & (1 << p_bit);
 }
 
-void GridMap::set_theme(const Ref<MeshLibrary> &p_theme) {
-
-	if (!theme.is_null())
-		theme->unregister_owner(this);
-	theme = p_theme;
-	if (!theme.is_null())
-		theme->register_owner(this);
+void GridMap::set_mesh_library(const Ref<MeshLibrary> &p_mesh_library) {
+	if (!mesh_library.is_null()) {
+		mesh_library->unregister_owner(this);
+	}
+	mesh_library = p_mesh_library;
+	if (!mesh_library.is_null()) {
+		mesh_library->register_owner(this);
+	}
 
 	_recreate_octant_data();
-	_change_notify("theme");
+	_change_notify("mesh_library");
 }
 
-Ref<MeshLibrary> GridMap::get_theme() const {
+Ref<MeshLibrary> GridMap::get_mesh_library() const {
+	return mesh_library;
+}
 
-	return theme;
+void GridMap::set_use_in_baked_light(bool p_use_baked_light) {
+	use_in_baked_light = p_use_baked_light;
+}
+
+bool GridMap::get_use_in_baked_light() const {
+	return use_in_baked_light;
 }
 
 void GridMap::set_cell_size(const Vector3 &p_size) {
-
 	ERR_FAIL_COND(p_size.x < 0.001 || p_size.y < 0.001 || p_size.z < 0.001);
 	cell_size = p_size;
 	_recreate_octant_data();
+	emit_signal("cell_size_changed", cell_size);
 }
 Vector3 GridMap::get_cell_size() const {
-
 	return cell_size;
 }
 
 void GridMap::set_octant_size(int p_size) {
-
+	ERR_FAIL_COND(p_size == 0);
 	octant_size = p_size;
 	_recreate_octant_data();
 }
 int GridMap::get_octant_size() const {
-
 	return octant_size;
 }
 
 void GridMap::set_center_x(bool p_enable) {
-
 	center_x = p_enable;
 	_recreate_octant_data();
 }
@@ -257,7 +236,6 @@ bool GridMap::get_center_x() const {
 }
 
 void GridMap::set_center_y(bool p_enable) {
-
 	center_y = p_enable;
 	_recreate_octant_data();
 }
@@ -267,7 +245,6 @@ bool GridMap::get_center_y() const {
 }
 
 void GridMap::set_center_z(bool p_enable) {
-
 	center_z = p_enable;
 	_recreate_octant_data();
 }
@@ -277,7 +254,6 @@ bool GridMap::get_center_z() const {
 }
 
 void GridMap::set_cell_item(int p_x, int p_y, int p_z, int p_item, int p_rot) {
-
 	if (baked_meshes.size() && !recreating_octants) {
 		//if you set a cell item, baked meshes go good bye
 		clear_baked_meshes();
@@ -326,7 +302,6 @@ void GridMap::set_cell_item(int p_x, int p_y, int p_z, int p_item, int p_rot) {
 		SceneTree *st = SceneTree::get_singleton();
 
 		if (st && st->is_debugging_collisions_hint()) {
-
 			g->collision_debug = VisualServer::get_singleton()->mesh_create();
 			g->collision_debug_instance = VisualServer::get_singleton()->instance_create();
 			VisualServer::get_singleton()->instance_set_base(g->collision_debug_instance, g->collision_debug);
@@ -353,7 +328,6 @@ void GridMap::set_cell_item(int p_x, int p_y, int p_z, int p_item, int p_rot) {
 }
 
 int GridMap::get_cell_item(int p_x, int p_y, int p_z) const {
-
 	ERR_FAIL_INDEX_V(ABS(p_x), 1 << 20, INVALID_CELL_ITEM);
 	ERR_FAIL_INDEX_V(ABS(p_y), 1 << 20, INVALID_CELL_ITEM);
 	ERR_FAIL_INDEX_V(ABS(p_z), 1 << 20, INVALID_CELL_ITEM);
@@ -363,13 +337,13 @@ int GridMap::get_cell_item(int p_x, int p_y, int p_z) const {
 	key.y = p_y;
 	key.z = p_z;
 
-	if (!cell_map.has(key))
+	if (!cell_map.has(key)) {
 		return INVALID_CELL_ITEM;
+	}
 	return cell_map[key].item;
 }
 
 int GridMap::get_cell_item_orientation(int p_x, int p_y, int p_z) const {
-
 	ERR_FAIL_INDEX_V(ABS(p_x), 1 << 20, -1);
 	ERR_FAIL_INDEX_V(ABS(p_y), 1 << 20, -1);
 	ERR_FAIL_INDEX_V(ABS(p_z), 1 << 20, -1);
@@ -379,8 +353,9 @@ int GridMap::get_cell_item_orientation(int p_x, int p_y, int p_z) const {
 	key.y = p_y;
 	key.z = p_z;
 
-	if (!cell_map.has(key))
+	if (!cell_map.has(key)) {
 		return -1;
+	}
 	return cell_map[key].rot;
 }
 
@@ -402,7 +377,6 @@ Vector3 GridMap::map_to_world(int p_x, int p_y, int p_z) const {
 }
 
 void GridMap::_octant_transform(const OctantKey &p_key) {
-
 	ERR_FAIL_COND(!octant_map.has(p_key));
 	Octant &g = *octant_map[p_key];
 	PhysicsServer::get_singleton()->body_set_state(g.static_body, PhysicsServer::BODY_STATE_TRANSFORM, get_global_transform());
@@ -419,15 +393,15 @@ void GridMap::_octant_transform(const OctantKey &p_key) {
 bool GridMap::_octant_update(const OctantKey &p_key) {
 	ERR_FAIL_COND_V(!octant_map.has(p_key), false);
 	Octant &g = *octant_map[p_key];
-	if (!g.dirty)
+	if (!g.dirty) {
 		return false;
+	}
 
 	//erase body shapes
 	PhysicsServer::get_singleton()->body_clear_shapes(g.static_body);
 
 	//erase body shapes debug
 	if (g.collision_debug.is_valid()) {
-
 		VS::get_singleton()->mesh_clear(g.collision_debug);
 	}
 
@@ -442,9 +416,12 @@ bool GridMap::_octant_update(const OctantKey &p_key) {
 	//erase multimeshes
 
 	for (int i = 0; i < g.multimesh_instances.size(); i++) {
-
-		VS::get_singleton()->free(g.multimesh_instances[i].instance);
-		VS::get_singleton()->free(g.multimesh_instances[i].multimesh);
+		if (g.multimesh_instances[i].instance.is_valid()) {
+			VS::get_singleton()->free(g.multimesh_instances[i].instance);
+		}
+		if (g.multimesh_instances[i].multimesh.is_valid()) {
+			VS::get_singleton()->free(g.multimesh_instances[i].multimesh);
+		}
 	}
 	g.multimesh_instances.clear();
 
@@ -462,63 +439,55 @@ bool GridMap::_octant_update(const OctantKey &p_key) {
 	 * and set said multimesh bounding box to one containing all cells which have this item
 	 */
 
-	Map<int, List<Pair<Transform, IndexKey> > > multimesh_items;
+	Map<int, List<Pair<Transform, IndexKey>>> multimesh_items;
 
 	for (Set<IndexKey>::Element *E = g.cells.front(); E; E = E->next()) {
-
 		ERR_CONTINUE(!cell_map.has(E->get()));
 		const Cell &c = cell_map[E->get()];
 
-		if (!theme.is_valid() || !theme->has_item(c.item))
+		if (!mesh_library.is_valid() || !mesh_library->has_item(c.item)) {
 			continue;
-
-		//print_line("OCTANT, CELLS: "+itos(ii.cells.size()));
+		}
 
 		Vector3 cellpos = Vector3(E->get().x, E->get().y, E->get().z);
 		Vector3 ofs = _get_offset();
 
 		Transform xform;
 
-		if (clip && ((clip_above && cellpos[clip_axis] > clip_floor) || (!clip_above && cellpos[clip_axis] < clip_floor))) {
-
-		} else {
-		}
-
 		xform.basis.set_orthogonal_index(c.rot);
 		xform.set_origin(cellpos * cell_size + ofs);
 		xform.basis.scale(Vector3(cell_scale, cell_scale, cell_scale));
 		if (baked_meshes.size() == 0) {
-			if (theme->get_item_mesh(c.item).is_valid()) {
+			if (mesh_library->get_item_mesh(c.item).is_valid()) {
 				if (!multimesh_items.has(c.item)) {
-					multimesh_items[c.item] = List<Pair<Transform, IndexKey> >();
+					multimesh_items[c.item] = List<Pair<Transform, IndexKey>>();
 				}
 
 				Pair<Transform, IndexKey> p;
-				p.first = xform;
+				p.first = xform * mesh_library->get_item_mesh_transform(c.item);
 				p.second = E->get();
 				multimesh_items[c.item].push_back(p);
 			}
 		}
 
-		Vector<MeshLibrary::ShapeData> shapes = theme->get_item_shapes(c.item);
+		Vector<MeshLibrary::ShapeData> shapes = mesh_library->get_item_shapes(c.item);
 		// add the item's shape at given xform to octant's static_body
 		for (int i = 0; i < shapes.size(); i++) {
 			// add the item's shape
-			if (!shapes[i].shape.is_valid())
+			if (!shapes[i].shape.is_valid()) {
 				continue;
+			}
 			PhysicsServer::get_singleton()->body_add_shape(g.static_body, shapes[i].shape->get_rid(), xform * shapes[i].local_transform);
 			if (g.collision_debug.is_valid()) {
-				shapes[i].shape->add_vertices_to_array(col_debug, xform * shapes[i].local_transform);
+				shapes.write[i].shape->add_vertices_to_array(col_debug, xform * shapes[i].local_transform);
 			}
-
-			//print_line("PHIS x: "+xform);
 		}
 
 		// add the item's navmesh at given xform to GridMap's Navigation ancestor
-		Ref<NavigationMesh> navmesh = theme->get_item_navmesh(c.item);
+		Ref<NavigationMesh> navmesh = mesh_library->get_item_navmesh(c.item);
 		if (navmesh.is_valid()) {
 			Octant::NavMesh nm;
-			nm.xform = xform;
+			nm.xform = xform * mesh_library->get_item_navmesh_transform(c.item);
 
 			if (navigation) {
 				nm.id = navigation->navmesh_add(navmesh, xform, this);
@@ -531,16 +500,15 @@ bool GridMap::_octant_update(const OctantKey &p_key) {
 
 	//update multimeshes, only if not baked
 	if (baked_meshes.size() == 0) {
-
-		for (Map<int, List<Pair<Transform, IndexKey> > >::Element *E = multimesh_items.front(); E; E = E->next()) {
+		for (Map<int, List<Pair<Transform, IndexKey>>>::Element *E = multimesh_items.front(); E; E = E->next()) {
 			Octant::MultimeshInstance mmi;
 
 			RID mm = VS::get_singleton()->multimesh_create();
 			VS::get_singleton()->multimesh_allocate(mm, E->get().size(), VS::MULTIMESH_TRANSFORM_3D, VS::MULTIMESH_COLOR_NONE);
-			VS::get_singleton()->multimesh_set_mesh(mm, theme->get_item_mesh(E->key())->get_rid());
+			VS::get_singleton()->multimesh_set_mesh(mm, mesh_library->get_item_mesh(E->key())->get_rid());
 
 			int idx = 0;
-			for (List<Pair<Transform, IndexKey> >::Element *F = E->get().front(); F; F = F->next()) {
+			for (List<Pair<Transform, IndexKey>>::Element *F = E->get().front(); F; F = F->next()) {
 				VS::get_singleton()->multimesh_instance_set_transform(mm, idx, F->get().first);
 #ifdef TOOLS_ENABLED
 
@@ -570,7 +538,6 @@ bool GridMap::_octant_update(const OctantKey &p_key) {
 	}
 
 	if (col_debug.size()) {
-
 		Array arr;
 		arr.resize(VS::ARRAY_MAX);
 		arr[VS::ARRAY_VERTEX] = col_debug;
@@ -595,12 +562,10 @@ void GridMap::_reset_physic_bodies_collision_filters() {
 }
 
 void GridMap::_octant_enter_world(const OctantKey &p_key) {
-
 	ERR_FAIL_COND(!octant_map.has(p_key));
 	Octant &g = *octant_map[p_key];
 	PhysicsServer::get_singleton()->body_set_state(g.static_body, PhysicsServer::BODY_STATE_TRANSFORM, get_global_transform());
 	PhysicsServer::get_singleton()->body_set_space(g.static_body, get_world()->get_space());
-	//print_line("BODYPOS: "+get_global_transform());
 
 	if (g.collision_debug_instance.is_valid()) {
 		VS::get_singleton()->instance_set_scenario(g.collision_debug_instance, get_world()->get_scenario());
@@ -612,11 +577,10 @@ void GridMap::_octant_enter_world(const OctantKey &p_key) {
 		VS::get_singleton()->instance_set_transform(g.multimesh_instances[i].instance, get_global_transform());
 	}
 
-	if (navigation && theme.is_valid()) {
+	if (navigation && mesh_library.is_valid()) {
 		for (Map<IndexKey, Octant::NavMesh>::Element *F = g.navmesh_ids.front(); F; F = F->next()) {
-
 			if (cell_map.has(F->key()) && F->get().id < 0) {
-				Ref<NavigationMesh> nm = theme->get_item_navmesh(cell_map[F->key()].item);
+				Ref<NavigationMesh> nm = mesh_library->get_item_navmesh(cell_map[F->key()].item);
 				if (nm.is_valid()) {
 					F->get().id = navigation->navmesh_add(nm, F->get().xform, this);
 				}
@@ -626,14 +590,12 @@ void GridMap::_octant_enter_world(const OctantKey &p_key) {
 }
 
 void GridMap::_octant_exit_world(const OctantKey &p_key) {
-
 	ERR_FAIL_COND(!octant_map.has(p_key));
 	Octant &g = *octant_map[p_key];
 	PhysicsServer::get_singleton()->body_set_state(g.static_body, PhysicsServer::BODY_STATE_TRANSFORM, get_global_transform());
 	PhysicsServer::get_singleton()->body_set_space(g.static_body, RID());
 
 	if (g.collision_debug_instance.is_valid()) {
-
 		VS::get_singleton()->instance_set_scenario(g.collision_debug_instance, RID());
 	}
 
@@ -643,7 +605,6 @@ void GridMap::_octant_exit_world(const OctantKey &p_key) {
 
 	if (navigation) {
 		for (Map<IndexKey, Octant::NavMesh>::Element *F = g.navmesh_ids.front(); F; F = F->next()) {
-
 			if (F->get().id >= 0) {
 				navigation->navmesh_remove(F->get().id);
 				F->get().id = -1;
@@ -653,16 +614,23 @@ void GridMap::_octant_exit_world(const OctantKey &p_key) {
 }
 
 void GridMap::_octant_clean_up(const OctantKey &p_key) {
-
 	ERR_FAIL_COND(!octant_map.has(p_key));
 	Octant &g = *octant_map[p_key];
 
-	if (g.collision_debug.is_valid())
+	if (g.collision_debug.is_valid()) {
 		VS::get_singleton()->free(g.collision_debug);
-	if (g.collision_debug_instance.is_valid())
-		VS::get_singleton()->free(g.collision_debug_instance);
+		g.collision_debug = RID();
+	}
 
-	PhysicsServer::get_singleton()->free(g.static_body);
+	if (g.collision_debug_instance.is_valid()) {
+		VS::get_singleton()->free(g.collision_debug_instance);
+		g.collision_debug_instance = RID();
+	}
+
+	if (g.static_body.is_valid()) {
+		PhysicsServer::get_singleton()->free(g.static_body);
+		g.static_body = RID();
+	}
 
 	//erase navigation
 	if (navigation) {
@@ -675,19 +643,19 @@ void GridMap::_octant_clean_up(const OctantKey &p_key) {
 	//erase multimeshes
 
 	for (int i = 0; i < g.multimesh_instances.size(); i++) {
-
-		VS::get_singleton()->free(g.multimesh_instances[i].instance);
-		VS::get_singleton()->free(g.multimesh_instances[i].multimesh);
+		if (g.multimesh_instances[i].instance.is_valid()) {
+			VS::get_singleton()->free(g.multimesh_instances[i].instance);
+		}
+		if (g.multimesh_instances[i].multimesh.is_valid()) {
+			VS::get_singleton()->free(g.multimesh_instances[i].multimesh);
+		}
 	}
 	g.multimesh_instances.clear();
 }
 
 void GridMap::_notification(int p_what) {
-
 	switch (p_what) {
-
 		case NOTIFICATION_ENTER_WORLD: {
-
 			Spatial *c = this;
 			while (c) {
 				navigation = Object::cast_to<Navigation>(c);
@@ -711,10 +679,10 @@ void GridMap::_notification(int p_what) {
 
 		} break;
 		case NOTIFICATION_TRANSFORM_CHANGED: {
-
 			Transform new_xform = get_global_transform();
-			if (new_xform == last_transform)
+			if (new_xform == last_transform) {
 				break;
+			}
 			//update run
 			for (Map<OctantKey, Octant *>::Element *E = octant_map.front(); E; E = E->next()) {
 				_octant_transform(E->key());
@@ -728,12 +696,11 @@ void GridMap::_notification(int p_what) {
 
 		} break;
 		case NOTIFICATION_EXIT_WORLD: {
-
 			for (Map<OctantKey, Octant *>::Element *E = octant_map.front(); E; E = E->next()) {
 				_octant_exit_world(E->key());
 			}
 
-			navigation = NULL;
+			navigation = nullptr;
 
 			//_queue_octants_dirty(MAP_DIRTY_INSTANCES|MAP_DIRTY_TRANSFORMS);
 			//_update_octants_callback();
@@ -750,46 +717,49 @@ void GridMap::_notification(int p_what) {
 }
 
 void GridMap::_update_visibility() {
-	if (!is_inside_tree())
+	if (!is_inside_tree()) {
 		return;
+	}
 
 	_change_notify("visible");
 
 	for (Map<OctantKey, Octant *>::Element *e = octant_map.front(); e; e = e->next()) {
 		Octant *octant = e->value();
 		for (int i = 0; i < octant->multimesh_instances.size(); i++) {
-			Octant::MultimeshInstance &mi = octant->multimesh_instances[i];
-			VS::get_singleton()->instance_set_visible(mi.instance, is_visible());
+			const Octant::MultimeshInstance &mi = octant->multimesh_instances[i];
+			VS::get_singleton()->instance_set_visible(mi.instance, is_visible_in_tree());
 		}
 	}
+
+	for (int i = 0; i < baked_meshes.size(); i++) {
+		VS::get_singleton()->instance_set_visible(baked_meshes[i].instance, is_visible_in_tree());
+	}
 }
 
 void GridMap::_queue_octants_dirty() {
-
-	if (awaiting_update)
+	if (awaiting_update) {
 		return;
+	}
 
 	MessageQueue::get_singleton()->push_call(this, "_update_octants_callback");
 	awaiting_update = true;
 }
 
 void GridMap::_recreate_octant_data() {
-
 	recreating_octants = true;
 	Map<IndexKey, Cell> cell_copy = cell_map;
 	_clear_internal();
 	for (Map<IndexKey, Cell>::Element *E = cell_copy.front(); E; E = E->next()) {
-
 		set_cell_item(E->key().x, E->key().y, E->key().z, E->get().item, E->get().rot);
 	}
 	recreating_octants = false;
 }
 
 void GridMap::_clear_internal() {
-
 	for (Map<OctantKey, Octant *>::Element *E = octant_map.front(); E; E = E->next()) {
-		if (is_inside_world())
+		if (is_inside_world()) {
 			_octant_exit_world(E->key());
+		}
 
 		_octant_clean_up(E->key());
 		memdelete(E->get());
@@ -800,32 +770,30 @@ void GridMap::_clear_internal() {
 }
 
 void GridMap::clear() {
-
 	_clear_internal();
 	clear_baked_meshes();
 }
 
 void GridMap::resource_changed(const RES &p_res) {
-
 	_recreate_octant_data();
 }
 
 void GridMap::_update_octants_callback() {
-
-	if (!awaiting_update)
+	if (!awaiting_update) {
 		return;
+	}
 
 	List<OctantKey> to_delete;
 	for (Map<OctantKey, Octant *>::Element *E = octant_map.front(); E; E = E->next()) {
-
 		if (_octant_update(E->key())) {
 			to_delete.push_back(E->key());
 		}
 	}
 
 	while (to_delete.front()) {
+		memdelete(octant_map[to_delete.front()->get()]);
 		octant_map.erase(to_delete.front()->get());
-		to_delete.pop_back();
+		to_delete.pop_front();
 	}
 
 	_update_visibility();
@@ -833,7 +801,6 @@ void GridMap::_update_octants_callback() {
 }
 
 void GridMap::_bind_methods() {
-
 	ClassDB::bind_method(D_METHOD("set_collision_layer", "layer"), &GridMap::set_collision_layer);
 	ClassDB::bind_method(D_METHOD("get_collision_layer"), &GridMap::get_collision_layer);
 
@@ -846,8 +813,8 @@ void GridMap::_bind_methods() {
 	ClassDB::bind_method(D_METHOD("set_collision_layer_bit", "bit", "value"), &GridMap::set_collision_layer_bit);
 	ClassDB::bind_method(D_METHOD("get_collision_layer_bit", "bit"), &GridMap::get_collision_layer_bit);
 
-	ClassDB::bind_method(D_METHOD("set_theme", "theme"), &GridMap::set_theme);
-	ClassDB::bind_method(D_METHOD("get_theme"), &GridMap::get_theme);
+	ClassDB::bind_method(D_METHOD("set_mesh_library", "mesh_library"), &GridMap::set_mesh_library);
+	ClassDB::bind_method(D_METHOD("get_mesh_library"), &GridMap::get_mesh_library);
 
 	ClassDB::bind_method(D_METHOD("set_cell_size", "size"), &GridMap::set_cell_size);
 	ClassDB::bind_method(D_METHOD("get_cell_size"), &GridMap::get_cell_size);
@@ -865,7 +832,6 @@ void GridMap::_bind_methods() {
 	ClassDB::bind_method(D_METHOD("world_to_map", "pos"), &GridMap::world_to_map);
 	ClassDB::bind_method(D_METHOD("map_to_world", "x", "y", "z"), &GridMap::map_to_world);
 
-	//ClassDB::bind_method(D_METHOD("_recreate_octants"),&GridMap::_recreate_octants);
 	ClassDB::bind_method(D_METHOD("_update_octants_callback"), &GridMap::_update_octants_callback);
 	ClassDB::bind_method(D_METHOD("resource_changed", "resource"), &GridMap::resource_changed);
 
@@ -889,7 +855,11 @@ void GridMap::_bind_methods() {
 	ClassDB::bind_method(D_METHOD("clear_baked_meshes"), &GridMap::clear_baked_meshes);
 	ClassDB::bind_method(D_METHOD("make_baked_meshes", "gen_lightmap_uv", "lightmap_uv_texel_size"), &GridMap::make_baked_meshes, DEFVAL(false), DEFVAL(0.1));
 
-	ADD_PROPERTY(PropertyInfo(Variant::OBJECT, "theme", PROPERTY_HINT_RESOURCE_TYPE, "MeshLibrary"), "set_theme", "get_theme");
+	ClassDB::bind_method(D_METHOD("set_use_in_baked_light", "use_in_baked_light"), &GridMap::set_use_in_baked_light);
+	ClassDB::bind_method(D_METHOD("get_use_in_baked_light"), &GridMap::get_use_in_baked_light);
+
+	ADD_PROPERTY(PropertyInfo(Variant::OBJECT, "mesh_library", PROPERTY_HINT_RESOURCE_TYPE, "MeshLibrary"), "set_mesh_library", "get_mesh_library");
+	ADD_PROPERTY(PropertyInfo(Variant::BOOL, "use_in_baked_light"), "set_use_in_baked_light", "get_use_in_baked_light");
 	ADD_GROUP("Cell", "cell_");
 	ADD_PROPERTY(PropertyInfo(Variant::VECTOR3, "cell_size"), "set_cell_size", "get_cell_size");
 	ADD_PROPERTY(PropertyInfo(Variant::INT, "cell_octant_size", PROPERTY_HINT_RANGE, "1,1024,1"), "set_octant_size", "get_octant_size");
@@ -902,14 +872,17 @@ void GridMap::_bind_methods() {
 	ADD_PROPERTY(PropertyInfo(Variant::INT, "collision_mask", PROPERTY_HINT_LAYERS_3D_PHYSICS), "set_collision_mask", "get_collision_mask");
 
 	BIND_CONSTANT(INVALID_CELL_ITEM);
+
+	ADD_SIGNAL(MethodInfo("cell_size_changed", PropertyInfo(Variant::VECTOR3, "cell_size")));
 }
 
 void GridMap::set_clip(bool p_enabled, bool p_clip_above, int p_floor, Vector3::Axis p_axis) {
-
-	if (!p_enabled && !clip)
+	if (!p_enabled && !clip) {
 		return;
-	if (clip && p_enabled && clip_floor == p_floor && p_clip_above == clip_above && p_axis == clip_axis)
+	}
+	if (clip && p_enabled && clip_floor == p_floor && p_clip_above == clip_above && p_axis == clip_axis) {
 		return;
+	}
 
 	clip = p_enabled;
 	clip_floor = p_floor;
@@ -918,7 +891,6 @@ void GridMap::set_clip(bool p_enabled, bool p_clip_above, int p_floor, Vector3::
 
 	//make it all update
 	for (Map<OctantKey, Octant *>::Element *E = octant_map.front(); E; E = E->next()) {
-
 		Octant *g = E->get();
 		g->dirty = true;
 	}
@@ -927,18 +899,15 @@ void GridMap::set_clip(bool p_enabled, bool p_clip_above, int p_floor, Vector3::
 }
 
 void GridMap::set_cell_scale(float p_scale) {
-
 	cell_scale = p_scale;
 	_recreate_octant_data();
 }
 
 float GridMap::get_cell_scale() const {
-
 	return cell_scale;
 }
 
 Array GridMap::get_used_cells() const {
-
 	Array a;
 	a.resize(cell_map.size());
 	int i = 0;
@@ -951,21 +920,22 @@ Array GridMap::get_used_cells() const {
 }
 
 Array GridMap::get_meshes() {
-
-	if (theme.is_null())
+	if (mesh_library.is_null()) {
 		return Array();
+	}
 
 	Vector3 ofs = _get_offset();
 	Array meshes;
 
 	for (Map<IndexKey, Cell>::Element *E = cell_map.front(); E; E = E->next()) {
-
 		int id = E->get().item;
-		if (!theme->has_item(id))
+		if (!mesh_library->has_item(id)) {
 			continue;
-		Ref<Mesh> mesh = theme->get_item_mesh(id);
-		if (mesh.is_null())
+		}
+		Ref<Mesh> mesh = mesh_library->get_item_mesh(id);
+		if (mesh.is_null()) {
 			continue;
+		}
 
 		IndexKey ik = E->key();
 
@@ -993,9 +963,10 @@ Vector3 GridMap::_get_offset() const {
 }
 
 void GridMap::clear_baked_meshes() {
-
 	for (int i = 0; i < baked_meshes.size(); i++) {
-		VS::get_singleton()->free(baked_meshes[i].instance);
+		if (baked_meshes[i].instance.is_valid()) {
+			VS::get_singleton()->free(baked_meshes[i].instance);
+		}
 	}
 	baked_meshes.clear();
 
@@ -1003,24 +974,25 @@ void GridMap::clear_baked_meshes() {
 }
 
 void GridMap::make_baked_meshes(bool p_gen_lightmap_uv, float p_lightmap_uv_texel_size) {
-
-	if (!theme.is_valid())
+	if (!mesh_library.is_valid()) {
 		return;
+	}
 
 	//generate
-	Map<OctantKey, Map<Ref<Material>, Ref<SurfaceTool> > > surface_map;
+	Map<OctantKey, Map<Ref<Material>, Ref<SurfaceTool>>> surface_map;
 
 	for (Map<IndexKey, Cell>::Element *E = cell_map.front(); E; E = E->next()) {
-
 		IndexKey key = E->key();
 
 		int item = E->get().item;
-		if (!theme->has_item(item))
+		if (!mesh_library->has_item(item)) {
 			continue;
+		}
 
-		Ref<Mesh> mesh = theme->get_item_mesh(item);
-		if (!mesh.is_valid())
+		Ref<Mesh> mesh = mesh_library->get_item_mesh(item);
+		if (!mesh.is_valid()) {
 			continue;
+		}
 
 		Vector3 cellpos = Vector3(key.x, key.y, key.z);
 		Vector3 ofs = _get_offset();
@@ -1037,15 +1009,15 @@ void GridMap::make_baked_meshes(bool p_gen_lightmap_uv, float p_lightmap_uv_texe
 		ok.z = key.z / octant_size;
 
 		if (!surface_map.has(ok)) {
-			surface_map[ok] = Map<Ref<Material>, Ref<SurfaceTool> >();
+			surface_map[ok] = Map<Ref<Material>, Ref<SurfaceTool>>();
 		}
 
-		Map<Ref<Material>, Ref<SurfaceTool> > &mat_map = surface_map[ok];
+		Map<Ref<Material>, Ref<SurfaceTool>> &mat_map = surface_map[ok];
 
 		for (int i = 0; i < mesh->get_surface_count(); i++) {
-
-			if (mesh->surface_get_primitive_type(i) != Mesh::PRIMITIVE_TRIANGLES)
+			if (mesh->surface_get_primitive_type(i) != Mesh::PRIMITIVE_TRIANGLES) {
 				continue;
+			}
 
 			Ref<Material> surf_mat = mesh->surface_get_material(i);
 			if (!mat_map.has(surf_mat)) {
@@ -1060,14 +1032,10 @@ void GridMap::make_baked_meshes(bool p_gen_lightmap_uv, float p_lightmap_uv_texe
 		}
 	}
 
-	int ofs = 0;
-
-	for (Map<OctantKey, Map<Ref<Material>, Ref<SurfaceTool> > >::Element *E = surface_map.front(); E; E = E->next()) {
-
-		print_line("generating mesh " + itos(ofs++) + "/" + itos(surface_map.size()));
+	for (Map<OctantKey, Map<Ref<Material>, Ref<SurfaceTool>>>::Element *E = surface_map.front(); E; E = E->next()) {
 		Ref<ArrayMesh> mesh;
 		mesh.instance();
-		for (Map<Ref<Material>, Ref<SurfaceTool> >::Element *F = E->get().front(); F; F = F->next()) {
+		for (Map<Ref<Material>, Ref<SurfaceTool>>::Element *F = E->get().front(); F; F = F->next()) {
 			F->get()->commit(mesh);
 		}
 
@@ -1091,6 +1059,9 @@ void GridMap::make_baked_meshes(bool p_gen_lightmap_uv, float p_lightmap_uv_texe
 }
 
 Array GridMap::get_bake_meshes() {
+	if (!use_in_baked_light) {
+		return Array();
+	}
 
 	if (!baked_meshes.size()) {
 		make_baked_meshes(true);
@@ -1106,13 +1077,11 @@ Array GridMap::get_bake_meshes() {
 }
 
 RID GridMap::get_bake_mesh_instance(int p_idx) {
-
 	ERR_FAIL_INDEX_V(p_idx, baked_meshes.size(), RID());
 	return baked_meshes[p_idx].instance;
 }
 
 GridMap::GridMap() {
-
 	collision_layer = 1;
 	collision_mask = 1;
 
@@ -1130,15 +1099,17 @@ GridMap::GridMap() {
 	clip_above = true;
 	cell_scale = 1.0;
 
-	navigation = NULL;
+	navigation = nullptr;
 	set_notify_transform(true);
 	recreating_octants = false;
+
+	use_in_baked_light = false;
 }
 
 GridMap::~GridMap() {
-
-	if (!theme.is_null())
-		theme->unregister_owner(this);
+	if (!mesh_library.is_null()) {
+		mesh_library->unregister_owner(this);
+	}
 
 	clear();
 }
