@@ -5,8 +5,8 @@
 /*                           GODOT ENGINE                                */
 /*                      https://godotengine.org                          */
 /*************************************************************************/
-/* Copyright (c) 2007-2018 Juan Linietsky, Ariel Manzur.                 */
-/* Copyright (c) 2014-2018 Godot Engine contributors (cf. AUTHORS.md)    */
+/* Copyright (c) 2007-2022 Juan Linietsky, Ariel Manzur.                 */
+/* Copyright (c) 2014-2022 Godot Engine contributors (cf. AUTHORS.md).   */
 /*                                                                       */
 /* Permission is hereby granted, free of charge, to any person obtaining */
 /* a copy of this software and associated documentation files (the       */
@@ -30,13 +30,13 @@
 
 #include "visual_script_builtin_funcs.h"
 
-#include "class_db.h"
-#include "func_ref.h"
-#include "io/marshalls.h"
-#include "math_funcs.h"
-#include "os/os.h"
-#include "reference.h"
-#include "variant_parser.h"
+#include "core/class_db.h"
+#include "core/func_ref.h"
+#include "core/io/marshalls.h"
+#include "core/math/math_funcs.h"
+#include "core/os/os.h"
+#include "core/reference.h"
+#include "core/variant_parser.h"
 
 const char *VisualScriptBuiltinFunc::func_name[VisualScriptBuiltinFunc::FUNC_MAX] = {
 	"sin",
@@ -68,6 +68,7 @@ const char *VisualScriptBuiltinFunc::func_name[VisualScriptBuiltinFunc::FUNC_MAX
 	"lerp",
 	"inverse_lerp",
 	"range_lerp",
+	"move_toward",
 	"dectime",
 	"randomize",
 	"randi",
@@ -102,37 +103,38 @@ const char *VisualScriptBuiltinFunc::func_name[VisualScriptBuiltinFunc::FUNC_MAX
 	"var2bytes",
 	"bytes2var",
 	"color_named",
+	"smoothstep",
+	"posmod",
+	"lerp_angle",
+	"ord",
 };
 
 VisualScriptBuiltinFunc::BuiltinFunc VisualScriptBuiltinFunc::find_function(const String &p_string) {
-
 	for (int i = 0; i < FUNC_MAX; i++) {
-		if (p_string == func_name[i])
+		if (p_string == func_name[i]) {
 			return BuiltinFunc(i);
+		}
 	}
 
 	return FUNC_MAX;
 }
 
 String VisualScriptBuiltinFunc::get_func_name(BuiltinFunc p_func) {
-
 	ERR_FAIL_INDEX_V(p_func, FUNC_MAX, String());
 	return func_name[p_func];
 }
 
 int VisualScriptBuiltinFunc::get_output_sequence_port_count() const {
-
 	return has_input_sequence_port() ? 1 : 0;
 }
 
 bool VisualScriptBuiltinFunc::has_input_sequence_port() const {
-
 	switch (func) {
-
 		case MATH_RANDOMIZE:
 		case TEXT_PRINT:
 		case TEXT_PRINTERR:
 		case TEXT_PRINTRAW:
+		case MATH_SEED:
 			return true;
 		default:
 			return false;
@@ -140,9 +142,7 @@ bool VisualScriptBuiltinFunc::has_input_sequence_port() const {
 }
 
 int VisualScriptBuiltinFunc::get_func_argument_count(BuiltinFunc p_func) {
-
 	switch (p_func) {
-
 		case MATH_RANDOMIZE:
 		case MATH_RAND:
 		case MATH_RANDF:
@@ -177,19 +177,21 @@ int VisualScriptBuiltinFunc::get_func_argument_count(BuiltinFunc p_func) {
 		case OBJ_WEAKREF:
 		case TYPE_OF:
 		case TEXT_CHAR:
+		case TEXT_ORD:
 		case TEXT_STR:
 		case TEXT_PRINT:
 		case TEXT_PRINTERR:
 		case TEXT_PRINTRAW:
 		case VAR_TO_STR:
 		case STR_TO_VAR:
-		case VAR_TO_BYTES:
-		case BYTES_TO_VAR:
 		case TYPE_EXISTS:
 			return 1;
+		case VAR_TO_BYTES:
+		case BYTES_TO_VAR:
 		case MATH_ATAN2:
 		case MATH_FMOD:
 		case MATH_FPOSMOD:
+		case MATH_POSMOD:
 		case MATH_POW:
 		case MATH_EASE:
 		case MATH_STEPIFY:
@@ -203,7 +205,10 @@ int VisualScriptBuiltinFunc::get_func_argument_count(BuiltinFunc p_func) {
 		case COLORN:
 			return 2;
 		case MATH_LERP:
+		case MATH_LERP_ANGLE:
 		case MATH_INVERSE_LERP:
+		case MATH_SMOOTHSTEP:
+		case MATH_MOVE_TOWARD:
 		case MATH_DECTIME:
 		case MATH_WRAP:
 		case MATH_WRAPF:
@@ -218,11 +223,9 @@ int VisualScriptBuiltinFunc::get_func_argument_count(BuiltinFunc p_func) {
 }
 
 int VisualScriptBuiltinFunc::get_input_value_port_count() const {
-
 	return get_func_argument_count(func);
 }
 int VisualScriptBuiltinFunc::get_output_value_port_count() const {
-
 	switch (func) {
 		case MATH_RANDOMIZE:
 		case TEXT_PRINT:
@@ -240,14 +243,11 @@ int VisualScriptBuiltinFunc::get_output_value_port_count() const {
 }
 
 String VisualScriptBuiltinFunc::get_output_sequence_port_text(int p_port) const {
-
 	return String();
 }
 
 PropertyInfo VisualScriptBuiltinFunc::get_input_value_port_info(int p_idx) const {
-
 	switch (func) {
-
 		case MATH_SIN:
 		case MATH_COS:
 		case MATH_TAN:
@@ -257,107 +257,121 @@ PropertyInfo VisualScriptBuiltinFunc::get_input_value_port_info(int p_idx) const
 		case MATH_ASIN:
 		case MATH_ACOS:
 		case MATH_ATAN:
-		case MATH_SQRT: {
-			return PropertyInfo(Variant::REAL, "s");
-		} break;
-		case MATH_ATAN2:
-		case MATH_FMOD:
-		case MATH_FPOSMOD: {
-			if (p_idx == 0)
-				return PropertyInfo(Variant::REAL, "x");
-			else
-				return PropertyInfo(Variant::REAL, "y");
-		} break;
+		case MATH_SQRT:
 		case MATH_FLOOR:
 		case MATH_CEIL:
 		case MATH_ROUND:
 		case MATH_ABS:
-		case MATH_SIGN: {
-			return PropertyInfo(Variant::REAL, "s");
-
-		} break;
-
-		case MATH_POW: {
-			if (p_idx == 0)
-				return PropertyInfo(Variant::REAL, "x");
-			else
-				return PropertyInfo(Variant::REAL, "y");
-		} break;
+		case MATH_SIGN:
 		case MATH_LOG:
 		case MATH_EXP:
 		case MATH_ISNAN:
 		case MATH_ISINF: {
 			return PropertyInfo(Variant::REAL, "s");
 		} break;
+		case MATH_ATAN2: {
+			if (p_idx == 0) {
+				return PropertyInfo(Variant::REAL, "y");
+			} else {
+				return PropertyInfo(Variant::REAL, "x");
+			}
+		} break;
+		case MATH_FMOD:
+		case MATH_FPOSMOD:
+		case LOGIC_MAX:
+		case LOGIC_MIN: {
+			if (p_idx == 0) {
+				return PropertyInfo(Variant::REAL, "a");
+			} else {
+				return PropertyInfo(Variant::REAL, "b");
+			}
+		} break;
+		case MATH_POSMOD: {
+			if (p_idx == 0) {
+				return PropertyInfo(Variant::INT, "a");
+			} else {
+				return PropertyInfo(Variant::INT, "b");
+			}
+		} break;
+		case MATH_POW: {
+			if (p_idx == 0) {
+				return PropertyInfo(Variant::REAL, "base");
+			} else {
+				return PropertyInfo(Variant::REAL, "exp");
+			}
+		} break;
 		case MATH_EASE: {
-			if (p_idx == 0)
+			if (p_idx == 0) {
 				return PropertyInfo(Variant::REAL, "s");
-			else
+			} else {
 				return PropertyInfo(Variant::REAL, "curve");
+			}
 		} break;
 		case MATH_DECIMALS: {
 			return PropertyInfo(Variant::REAL, "step");
 		} break;
 		case MATH_STEPIFY: {
-			if (p_idx == 0)
+			if (p_idx == 0) {
 				return PropertyInfo(Variant::REAL, "s");
-			else
+			} else {
 				return PropertyInfo(Variant::REAL, "steps");
+			}
 		} break;
-		case MATH_LERP: {
-			if (p_idx == 0)
+		case MATH_LERP:
+		case MATH_LERP_ANGLE:
+		case MATH_INVERSE_LERP:
+		case MATH_SMOOTHSTEP: {
+			if (p_idx == 0) {
 				return PropertyInfo(Variant::REAL, "from");
-			else if (p_idx == 1)
+			} else if (p_idx == 1) {
 				return PropertyInfo(Variant::REAL, "to");
-			else
-				return PropertyInfo(Variant::REAL, "weight");
-		} break;
-		case MATH_INVERSE_LERP: {
-			if (p_idx == 0)
-				return PropertyInfo(Variant::REAL, "from");
-			else if (p_idx == 1)
-				return PropertyInfo(Variant::REAL, "to");
-			else
+			} else {
 				return PropertyInfo(Variant::REAL, "weight");
+			}
 		} break;
 		case MATH_RANGE_LERP: {
-			if (p_idx == 0)
+			if (p_idx == 0) {
 				return PropertyInfo(Variant::REAL, "value");
-			else if (p_idx == 1)
+			} else if (p_idx == 1) {
 				return PropertyInfo(Variant::REAL, "istart");
-			else if (p_idx == 2)
+			} else if (p_idx == 2) {
 				return PropertyInfo(Variant::REAL, "istop");
-			else if (p_idx == 3)
+			} else if (p_idx == 3) {
 				return PropertyInfo(Variant::REAL, "ostart");
-			else
+			} else {
 				return PropertyInfo(Variant::REAL, "ostop");
+			}
+		} break;
+		case MATH_MOVE_TOWARD: {
+			if (p_idx == 0) {
+				return PropertyInfo(Variant::REAL, "from");
+			} else if (p_idx == 1) {
+				return PropertyInfo(Variant::REAL, "to");
+			} else {
+				return PropertyInfo(Variant::REAL, "delta");
+			}
 		} break;
 		case MATH_DECTIME: {
-			if (p_idx == 0)
+			if (p_idx == 0) {
 				return PropertyInfo(Variant::REAL, "value");
-			else if (p_idx == 1)
+			} else if (p_idx == 1) {
 				return PropertyInfo(Variant::REAL, "amount");
-			else
+			} else {
 				return PropertyInfo(Variant::REAL, "step");
+			}
 		} break;
-		case MATH_RANDOMIZE: {
-
-		} break;
-		case MATH_RAND: {
-
-		} break;
+		case MATH_RANDOMIZE:
+		case MATH_RAND:
 		case MATH_RANDF: {
-
 		} break;
 		case MATH_RANDOM: {
-			if (p_idx == 0)
+			if (p_idx == 0) {
 				return PropertyInfo(Variant::REAL, "from");
-			else
+			} else {
 				return PropertyInfo(Variant::REAL, "to");
+			}
 		} break;
-		case MATH_SEED: {
-			return PropertyInfo(Variant::INT, "seed");
-		} break;
+		case MATH_SEED:
 		case MATH_RANDSEED: {
 			return PropertyInfo(Variant::INT, "seed");
 		} break;
@@ -374,132 +388,101 @@ PropertyInfo VisualScriptBuiltinFunc::get_input_value_port_info(int p_idx) const
 			return PropertyInfo(Variant::REAL, "db");
 		} break;
 		case MATH_POLAR2CARTESIAN: {
-			if (p_idx == 0)
+			if (p_idx == 0) {
 				return PropertyInfo(Variant::REAL, "r");
-			else
+			} else {
 				return PropertyInfo(Variant::REAL, "th");
+			}
 		} break;
 		case MATH_CARTESIAN2POLAR: {
-			if (p_idx == 0)
+			if (p_idx == 0) {
 				return PropertyInfo(Variant::REAL, "x");
-			else
+			} else {
 				return PropertyInfo(Variant::REAL, "y");
+			}
 		} break;
 		case MATH_WRAP: {
-			if (p_idx == 0)
+			if (p_idx == 0) {
 				return PropertyInfo(Variant::INT, "value");
-			else if (p_idx == 1)
+			} else if (p_idx == 1) {
 				return PropertyInfo(Variant::INT, "min");
-			else
+			} else {
 				return PropertyInfo(Variant::INT, "max");
+			}
 		} break;
-		case MATH_WRAPF: {
-			if (p_idx == 0)
-				return PropertyInfo(Variant::REAL, "value");
-			else if (p_idx == 1)
-				return PropertyInfo(Variant::REAL, "min");
-			else
-				return PropertyInfo(Variant::REAL, "max");
-		} break;
-		case LOGIC_MAX: {
-			if (p_idx == 0)
-				return PropertyInfo(Variant::REAL, "a");
-			else
-				return PropertyInfo(Variant::REAL, "b");
-		} break;
-		case LOGIC_MIN: {
-			if (p_idx == 0)
-				return PropertyInfo(Variant::REAL, "a");
-			else
-				return PropertyInfo(Variant::REAL, "b");
-		} break;
+		case MATH_WRAPF:
 		case LOGIC_CLAMP: {
-			if (p_idx == 0)
+			if (p_idx == 0) {
 				return PropertyInfo(Variant::REAL, "value");
-			else if (p_idx == 1)
+			} else if (p_idx == 1) {
 				return PropertyInfo(Variant::REAL, "min");
-			else
+			} else {
 				return PropertyInfo(Variant::REAL, "max");
+			}
 		} break;
 		case LOGIC_NEAREST_PO2: {
 			return PropertyInfo(Variant::INT, "value");
 		} break;
 		case OBJ_WEAKREF: {
-
 			return PropertyInfo(Variant::OBJECT, "source");
-
 		} break;
 		case FUNC_FUNCREF: {
-
-			if (p_idx == 0)
+			if (p_idx == 0) {
 				return PropertyInfo(Variant::OBJECT, "instance");
-			else
+			} else {
 				return PropertyInfo(Variant::STRING, "funcname");
-
+			}
 		} break;
 		case TYPE_CONVERT: {
-
-			if (p_idx == 0)
+			if (p_idx == 0) {
 				return PropertyInfo(Variant::NIL, "what");
-			else
+			} else {
 				return PropertyInfo(Variant::STRING, "type");
+			}
 		} break;
 		case TYPE_OF: {
 			return PropertyInfo(Variant::NIL, "what");
-
 		} break;
 		case TYPE_EXISTS: {
-
 			return PropertyInfo(Variant::STRING, "type");
-
+		} break;
+		case TEXT_ORD: {
+			return PropertyInfo(Variant::STRING, "character");
 		} break;
 		case TEXT_CHAR: {
-
 			return PropertyInfo(Variant::INT, "ascii");
-
-		} break;
-		case TEXT_STR: {
-
-			return PropertyInfo(Variant::NIL, "value");
-
-		} break;
-		case TEXT_PRINT: {
-
-			return PropertyInfo(Variant::NIL, "value");
-
-		} break;
-		case TEXT_PRINTERR: {
-			return PropertyInfo(Variant::NIL, "value");
-
 		} break;
+		case TEXT_STR:
+		case TEXT_PRINT:
+		case TEXT_PRINTERR:
 		case TEXT_PRINTRAW: {
-
 			return PropertyInfo(Variant::NIL, "value");
-
-		} break;
-		case VAR_TO_STR: {
-			return PropertyInfo(Variant::NIL, "var");
-
 		} break;
 		case STR_TO_VAR: {
-
 			return PropertyInfo(Variant::STRING, "string");
 		} break;
+		case VAR_TO_STR:
 		case VAR_TO_BYTES: {
-			return PropertyInfo(Variant::NIL, "var");
+			if (p_idx == 0) {
+				return PropertyInfo(Variant::NIL, "var");
+			} else {
+				return PropertyInfo(Variant::BOOL, "full_objects");
+			}
 
 		} break;
 		case BYTES_TO_VAR: {
-
-			return PropertyInfo(Variant::POOL_BYTE_ARRAY, "bytes");
+			if (p_idx == 0) {
+				return PropertyInfo(Variant::POOL_BYTE_ARRAY, "bytes");
+			} else {
+				return PropertyInfo(Variant::BOOL, "allow_objects");
+			}
 		} break;
 		case COLORN: {
-
-			if (p_idx == 0)
+			if (p_idx == 0) {
 				return PropertyInfo(Variant::STRING, "name");
-			else
+			} else {
 				return PropertyInfo(Variant::REAL, "alpha");
-
+			}
 		} break;
 		case FUNC_MAX: {
 		}
@@ -509,10 +492,8 @@ PropertyInfo VisualScriptBuiltinFunc::get_input_value_port_info(int p_idx) const
 }
 
 PropertyInfo VisualScriptBuiltinFunc::get_output_value_port_info(int p_idx) const {
-
 	Variant::Type t = Variant::NIL;
 	switch (func) {
-
 		case MATH_SIN:
 		case MATH_COS:
 		case MATH_TAN:
@@ -530,9 +511,12 @@ PropertyInfo VisualScriptBuiltinFunc::get_output_value_port_info(int p_idx) cons
 		case MATH_CEIL: {
 			t = Variant::REAL;
 		} break;
-		case MATH_ROUND: {
+		case MATH_POSMOD: {
 			t = Variant::INT;
 		} break;
+		case MATH_ROUND: {
+			t = Variant::REAL;
+		} break;
 		case MATH_ABS: {
 			t = Variant::NIL;
 		} break;
@@ -556,17 +540,18 @@ PropertyInfo VisualScriptBuiltinFunc::get_output_value_port_info(int p_idx) cons
 		} break;
 		case MATH_STEPIFY:
 		case MATH_LERP:
+		case MATH_LERP_ANGLE:
 		case MATH_INVERSE_LERP:
 		case MATH_RANGE_LERP:
+		case MATH_SMOOTHSTEP:
+		case MATH_MOVE_TOWARD:
 		case MATH_DECTIME: {
 			t = Variant::REAL;
 
 		} break;
 		case MATH_RANDOMIZE: {
-
 		} break;
 		case MATH_RAND: {
-
 			t = Variant::INT;
 		} break;
 		case MATH_RANDF:
@@ -574,14 +559,13 @@ PropertyInfo VisualScriptBuiltinFunc::get_output_value_port_info(int p_idx) cons
 			t = Variant::REAL;
 		} break;
 		case MATH_SEED: {
-
 		} break;
 		case MATH_RANDSEED: {
-
-			if (p_idx == 0)
+			if (p_idx == 0) {
 				return PropertyInfo(Variant::INT, "rnd");
-			else
+			} else {
 				return PropertyInfo(Variant::INT, "seed");
+			}
 		} break;
 		case MATH_DEG2RAD:
 		case MATH_RAD2DEG:
@@ -600,61 +584,58 @@ PropertyInfo VisualScriptBuiltinFunc::get_output_value_port_info(int p_idx) cons
 		case LOGIC_MAX:
 		case LOGIC_MIN:
 		case LOGIC_CLAMP: {
-
 		} break;
 
 		case LOGIC_NEAREST_PO2: {
 			t = Variant::NIL;
 		} break;
 		case OBJ_WEAKREF: {
-
 			t = Variant::OBJECT;
 
 		} break;
 		case FUNC_FUNCREF: {
-
 			t = Variant::OBJECT;
 
 		} break;
 		case TYPE_CONVERT: {
-
 		} break;
+		case TEXT_ORD:
 		case TYPE_OF: {
 			t = Variant::INT;
 
 		} break;
 		case TYPE_EXISTS: {
-
 			t = Variant::BOOL;
 
 		} break;
 		case TEXT_CHAR:
 		case TEXT_STR: {
-
 			t = Variant::STRING;
 
 		} break;
 		case TEXT_PRINT: {
-
 		} break;
 		case TEXT_PRINTERR: {
-
 		} break;
 		case TEXT_PRINTRAW: {
-
 		} break;
 		case VAR_TO_STR: {
 			t = Variant::STRING;
 		} break;
 		case STR_TO_VAR: {
-
 		} break;
 		case VAR_TO_BYTES: {
-			t = Variant::POOL_BYTE_ARRAY;
+			if (p_idx == 0) {
+				t = Variant::POOL_BYTE_ARRAY;
+			} else {
+				t = Variant::BOOL;
+			}
 
 		} break;
 		case BYTES_TO_VAR: {
-
+			if (p_idx == 1) {
+				t = Variant::BOOL;
+			}
 		} break;
 		case COLORN: {
 			t = Variant::COLOR;
@@ -666,18 +647,18 @@ PropertyInfo VisualScriptBuiltinFunc::get_output_value_port_info(int p_idx) cons
 	return PropertyInfo(t, "");
 }
 
+/*
 String VisualScriptBuiltinFunc::get_caption() const {
 
 	return "BuiltinFunc";
 }
+*/
 
-String VisualScriptBuiltinFunc::get_text() const {
-
+String VisualScriptBuiltinFunc::get_caption() const {
 	return func_name[func];
 }
 
 void VisualScriptBuiltinFunc::set_func(BuiltinFunc p_which) {
-
 	ERR_FAIL_INDEX(p_which, FUNC_MAX);
 	func = p_which;
 	_change_notify();
@@ -697,184 +678,159 @@ VisualScriptBuiltinFunc::BuiltinFunc VisualScriptBuiltinFunc::get_func() {
 	}
 
 void VisualScriptBuiltinFunc::exec_func(BuiltinFunc p_func, const Variant **p_inputs, Variant *r_return, Variant::CallError &r_error, String &r_error_str) {
-
 	switch (p_func) {
 		case VisualScriptBuiltinFunc::MATH_SIN: {
-
 			VALIDATE_ARG_NUM(0);
 			*r_return = Math::sin((double)*p_inputs[0]);
 		} break;
 		case VisualScriptBuiltinFunc::MATH_COS: {
-
 			VALIDATE_ARG_NUM(0);
 			*r_return = Math::cos((double)*p_inputs[0]);
 		} break;
 		case VisualScriptBuiltinFunc::MATH_TAN: {
-
 			VALIDATE_ARG_NUM(0);
 			*r_return = Math::tan((double)*p_inputs[0]);
 		} break;
 		case VisualScriptBuiltinFunc::MATH_SINH: {
-
 			VALIDATE_ARG_NUM(0);
 			*r_return = Math::sinh((double)*p_inputs[0]);
 		} break;
 		case VisualScriptBuiltinFunc::MATH_COSH: {
-
 			VALIDATE_ARG_NUM(0);
 			*r_return = Math::cosh((double)*p_inputs[0]);
 		} break;
 		case VisualScriptBuiltinFunc::MATH_TANH: {
-
 			VALIDATE_ARG_NUM(0);
 			*r_return = Math::tanh((double)*p_inputs[0]);
 		} break;
 		case VisualScriptBuiltinFunc::MATH_ASIN: {
-
 			VALIDATE_ARG_NUM(0);
 			*r_return = Math::asin((double)*p_inputs[0]);
 		} break;
 		case VisualScriptBuiltinFunc::MATH_ACOS: {
-
 			VALIDATE_ARG_NUM(0);
 			*r_return = Math::acos((double)*p_inputs[0]);
 		} break;
 		case VisualScriptBuiltinFunc::MATH_ATAN: {
-
 			VALIDATE_ARG_NUM(0);
 			*r_return = Math::atan((double)*p_inputs[0]);
 		} break;
 		case VisualScriptBuiltinFunc::MATH_ATAN2: {
-
 			VALIDATE_ARG_NUM(0);
 			VALIDATE_ARG_NUM(1);
 			*r_return = Math::atan2((double)*p_inputs[0], (double)*p_inputs[1]);
 		} break;
 		case VisualScriptBuiltinFunc::MATH_SQRT: {
-
 			VALIDATE_ARG_NUM(0);
 			*r_return = Math::sqrt((double)*p_inputs[0]);
 		} break;
 		case VisualScriptBuiltinFunc::MATH_FMOD: {
-
 			VALIDATE_ARG_NUM(0);
 			VALIDATE_ARG_NUM(1);
 			*r_return = Math::fmod((double)*p_inputs[0], (double)*p_inputs[1]);
 		} break;
 		case VisualScriptBuiltinFunc::MATH_FPOSMOD: {
-
 			VALIDATE_ARG_NUM(0);
 			VALIDATE_ARG_NUM(1);
 			*r_return = Math::fposmod((double)*p_inputs[0], (double)*p_inputs[1]);
 		} break;
+		case VisualScriptBuiltinFunc::MATH_POSMOD: {
+			VALIDATE_ARG_NUM(0);
+			VALIDATE_ARG_NUM(1);
+			*r_return = Math::posmod((int64_t)*p_inputs[0], (int64_t)*p_inputs[1]);
+		} break;
 		case VisualScriptBuiltinFunc::MATH_FLOOR: {
-
 			VALIDATE_ARG_NUM(0);
 			*r_return = Math::floor((double)*p_inputs[0]);
 		} break;
 		case VisualScriptBuiltinFunc::MATH_CEIL: {
-
 			VALIDATE_ARG_NUM(0);
 			*r_return = Math::ceil((double)*p_inputs[0]);
 		} break;
 		case VisualScriptBuiltinFunc::MATH_ROUND: {
-
 			VALIDATE_ARG_NUM(0);
 			*r_return = Math::round((double)*p_inputs[0]);
 		} break;
 		case VisualScriptBuiltinFunc::MATH_ABS: {
-
 			if (p_inputs[0]->get_type() == Variant::INT) {
-
 				int64_t i = *p_inputs[0];
 				*r_return = ABS(i);
 			} else if (p_inputs[0]->get_type() == Variant::REAL) {
-
 				real_t r = *p_inputs[0];
 				*r_return = Math::abs(r);
 			} else {
-
 				r_error.error = Variant::CallError::CALL_ERROR_INVALID_ARGUMENT;
 				r_error.argument = 0;
 				r_error.expected = Variant::REAL;
 			}
 		} break;
 		case VisualScriptBuiltinFunc::MATH_SIGN: {
-
 			if (p_inputs[0]->get_type() == Variant::INT) {
-
 				int64_t i = *p_inputs[0];
 				*r_return = i < 0 ? -1 : (i > 0 ? +1 : 0);
 			} else if (p_inputs[0]->get_type() == Variant::REAL) {
-
 				real_t r = *p_inputs[0];
 				*r_return = r < 0.0 ? -1.0 : (r > 0.0 ? +1.0 : 0.0);
 			} else {
-
 				r_error.error = Variant::CallError::CALL_ERROR_INVALID_ARGUMENT;
 				r_error.argument = 0;
 				r_error.expected = Variant::REAL;
 			}
 		} break;
 		case VisualScriptBuiltinFunc::MATH_POW: {
-
 			VALIDATE_ARG_NUM(0);
 			VALIDATE_ARG_NUM(1);
 			*r_return = Math::pow((double)*p_inputs[0], (double)*p_inputs[1]);
 		} break;
 		case VisualScriptBuiltinFunc::MATH_LOG: {
-
 			VALIDATE_ARG_NUM(0);
 			*r_return = Math::log((double)*p_inputs[0]);
 		} break;
 		case VisualScriptBuiltinFunc::MATH_EXP: {
-
 			VALIDATE_ARG_NUM(0);
 			*r_return = Math::exp((double)*p_inputs[0]);
 		} break;
 		case VisualScriptBuiltinFunc::MATH_ISNAN: {
-
 			VALIDATE_ARG_NUM(0);
 			*r_return = Math::is_nan((double)*p_inputs[0]);
 		} break;
 		case VisualScriptBuiltinFunc::MATH_ISINF: {
-
 			VALIDATE_ARG_NUM(0);
 			*r_return = Math::is_inf((double)*p_inputs[0]);
 		} break;
 		case VisualScriptBuiltinFunc::MATH_EASE: {
-
 			VALIDATE_ARG_NUM(0);
 			VALIDATE_ARG_NUM(1);
 			*r_return = Math::ease((double)*p_inputs[0], (double)*p_inputs[1]);
 		} break;
 		case VisualScriptBuiltinFunc::MATH_DECIMALS: {
-
 			VALIDATE_ARG_NUM(0);
 			*r_return = Math::step_decimals((double)*p_inputs[0]);
 		} break;
 		case VisualScriptBuiltinFunc::MATH_STEPIFY: {
-
 			VALIDATE_ARG_NUM(0);
 			VALIDATE_ARG_NUM(1);
 			*r_return = Math::stepify((double)*p_inputs[0], (double)*p_inputs[1]);
 		} break;
 		case VisualScriptBuiltinFunc::MATH_LERP: {
-
 			VALIDATE_ARG_NUM(0);
 			VALIDATE_ARG_NUM(1);
 			VALIDATE_ARG_NUM(2);
 			*r_return = Math::lerp((double)*p_inputs[0], (double)*p_inputs[1], (double)*p_inputs[2]);
 		} break;
+		case VisualScriptBuiltinFunc::MATH_LERP_ANGLE: {
+			VALIDATE_ARG_NUM(0);
+			VALIDATE_ARG_NUM(1);
+			VALIDATE_ARG_NUM(2);
+			*r_return = Math::lerp_angle((double)*p_inputs[0], (double)*p_inputs[1], (double)*p_inputs[2]);
+		} break;
 		case VisualScriptBuiltinFunc::MATH_INVERSE_LERP: {
-
 			VALIDATE_ARG_NUM(0);
 			VALIDATE_ARG_NUM(1);
 			VALIDATE_ARG_NUM(2);
 			*r_return = Math::inverse_lerp((double)*p_inputs[0], (double)*p_inputs[1], (double)*p_inputs[2]);
 		} break;
 		case VisualScriptBuiltinFunc::MATH_RANGE_LERP: {
-
 			VALIDATE_ARG_NUM(0);
 			VALIDATE_ARG_NUM(1);
 			VALIDATE_ARG_NUM(2);
@@ -882,8 +838,19 @@ void VisualScriptBuiltinFunc::exec_func(BuiltinFunc p_func, const Variant **p_in
 			VALIDATE_ARG_NUM(4);
 			*r_return = Math::range_lerp((double)*p_inputs[0], (double)*p_inputs[1], (double)*p_inputs[2], (double)*p_inputs[3], (double)*p_inputs[4]);
 		} break;
+		case VisualScriptBuiltinFunc::MATH_SMOOTHSTEP: {
+			VALIDATE_ARG_NUM(0);
+			VALIDATE_ARG_NUM(1);
+			VALIDATE_ARG_NUM(2);
+			*r_return = Math::smoothstep((double)*p_inputs[0], (double)*p_inputs[1], (double)*p_inputs[2]);
+		} break;
+		case VisualScriptBuiltinFunc::MATH_MOVE_TOWARD: {
+			VALIDATE_ARG_NUM(0);
+			VALIDATE_ARG_NUM(1);
+			VALIDATE_ARG_NUM(2);
+			*r_return = Math::move_toward((double)*p_inputs[0], (double)*p_inputs[1], (double)*p_inputs[2]);
+		} break;
 		case VisualScriptBuiltinFunc::MATH_DECTIME: {
-
 			VALIDATE_ARG_NUM(0);
 			VALIDATE_ARG_NUM(1);
 			VALIDATE_ARG_NUM(2);
@@ -900,20 +867,17 @@ void VisualScriptBuiltinFunc::exec_func(BuiltinFunc p_func, const Variant **p_in
 			*r_return = Math::randf();
 		} break;
 		case VisualScriptBuiltinFunc::MATH_RANDOM: {
-
 			VALIDATE_ARG_NUM(0);
 			VALIDATE_ARG_NUM(1);
 			*r_return = Math::random((double)*p_inputs[0], (double)*p_inputs[1]);
 		} break;
 		case VisualScriptBuiltinFunc::MATH_SEED: {
-
 			VALIDATE_ARG_NUM(0);
 			uint64_t seed = *p_inputs[0];
 			Math::seed(seed);
 
 		} break;
 		case VisualScriptBuiltinFunc::MATH_RANDSEED: {
-
 			VALIDATE_ARG_NUM(0);
 			uint64_t seed = *p_inputs[0];
 			int ret = Math::rand_from_seed(&seed);
@@ -924,22 +888,18 @@ void VisualScriptBuiltinFunc::exec_func(BuiltinFunc p_func, const Variant **p_in
 
 		} break;
 		case VisualScriptBuiltinFunc::MATH_DEG2RAD: {
-
 			VALIDATE_ARG_NUM(0);
 			*r_return = Math::deg2rad((double)*p_inputs[0]);
 		} break;
 		case VisualScriptBuiltinFunc::MATH_RAD2DEG: {
-
 			VALIDATE_ARG_NUM(0);
 			*r_return = Math::rad2deg((double)*p_inputs[0]);
 		} break;
 		case VisualScriptBuiltinFunc::MATH_LINEAR2DB: {
-
 			VALIDATE_ARG_NUM(0);
 			*r_return = Math::linear2db((double)*p_inputs[0]);
 		} break;
 		case VisualScriptBuiltinFunc::MATH_DB2LINEAR: {
-
 			VALIDATE_ARG_NUM(0);
 			*r_return = Math::db2linear((double)*p_inputs[0]);
 		} break;
@@ -970,9 +930,7 @@ void VisualScriptBuiltinFunc::exec_func(BuiltinFunc p_func, const Variant **p_in
 			*r_return = Math::wrapf((double)*p_inputs[0], (double)*p_inputs[1], (double)*p_inputs[2]);
 		} break;
 		case VisualScriptBuiltinFunc::LOGIC_MAX: {
-
 			if (p_inputs[0]->get_type() == Variant::INT && p_inputs[1]->get_type() == Variant::INT) {
-
 				int64_t a = *p_inputs[0];
 				int64_t b = *p_inputs[1];
 				*r_return = MAX(a, b);
@@ -988,9 +946,7 @@ void VisualScriptBuiltinFunc::exec_func(BuiltinFunc p_func, const Variant **p_in
 
 		} break;
 		case VisualScriptBuiltinFunc::LOGIC_MIN: {
-
 			if (p_inputs[0]->get_type() == Variant::INT && p_inputs[1]->get_type() == Variant::INT) {
-
 				int64_t a = *p_inputs[0];
 				int64_t b = *p_inputs[1];
 				*r_return = MIN(a, b);
@@ -1005,9 +961,7 @@ void VisualScriptBuiltinFunc::exec_func(BuiltinFunc p_func, const Variant **p_in
 			}
 		} break;
 		case VisualScriptBuiltinFunc::LOGIC_CLAMP: {
-
 			if (p_inputs[0]->get_type() == Variant::INT && p_inputs[1]->get_type() == Variant::INT && p_inputs[2]->get_type() == Variant::INT) {
-
 				int64_t a = *p_inputs[0];
 				int64_t b = *p_inputs[1];
 				int64_t c = *p_inputs[2];
@@ -1025,15 +979,12 @@ void VisualScriptBuiltinFunc::exec_func(BuiltinFunc p_func, const Variant **p_in
 			}
 		} break;
 		case VisualScriptBuiltinFunc::LOGIC_NEAREST_PO2: {
-
 			VALIDATE_ARG_NUM(0);
 			int64_t num = *p_inputs[0];
 			*r_return = next_power_of_2(num);
 		} break;
 		case VisualScriptBuiltinFunc::OBJ_WEAKREF: {
-
 			if (p_inputs[0]->get_type() != Variant::OBJECT) {
-
 				r_error.error = Variant::CallError::CALL_ERROR_INVALID_ARGUMENT;
 				r_error.argument = 0;
 				r_error.expected = Variant::OBJECT;
@@ -1042,10 +993,8 @@ void VisualScriptBuiltinFunc::exec_func(BuiltinFunc p_func, const Variant **p_in
 			}
 
 			if (p_inputs[0]->is_ref()) {
-
 				REF r = *p_inputs[0];
 				if (!r.is_valid()) {
-
 					return;
 				}
 
@@ -1055,7 +1004,6 @@ void VisualScriptBuiltinFunc::exec_func(BuiltinFunc p_func, const Variant **p_in
 			} else {
 				Object *obj = *p_inputs[0];
 				if (!obj) {
-
 					return;
 				}
 				Ref<WeakRef> wref = memnew(WeakRef);
@@ -1065,9 +1013,7 @@ void VisualScriptBuiltinFunc::exec_func(BuiltinFunc p_func, const Variant **p_in
 
 		} break;
 		case VisualScriptBuiltinFunc::FUNC_FUNCREF: {
-
 			if (p_inputs[0]->get_type() != Variant::OBJECT) {
-
 				r_error.error = Variant::CallError::CALL_ERROR_INVALID_ARGUMENT;
 				r_error.argument = 0;
 				r_error.expected = Variant::OBJECT;
@@ -1075,7 +1021,6 @@ void VisualScriptBuiltinFunc::exec_func(BuiltinFunc p_func, const Variant **p_in
 				return;
 			}
 			if (p_inputs[1]->get_type() != Variant::STRING && p_inputs[1]->get_type() != Variant::NODE_PATH) {
-
 				r_error.error = Variant::CallError::CALL_ERROR_INVALID_ARGUMENT;
 				r_error.argument = 1;
 				r_error.expected = Variant::STRING;
@@ -1092,11 +1037,9 @@ void VisualScriptBuiltinFunc::exec_func(BuiltinFunc p_func, const Variant **p_in
 
 		} break;
 		case VisualScriptBuiltinFunc::TYPE_CONVERT: {
-
 			VALIDATE_ARG_NUM(1);
 			int type = *p_inputs[1];
 			if (type < 0 || type >= Variant::VARIANT_MAX) {
-
 				r_error_str = RTR("Invalid type argument to convert(), use TYPE_* constants.");
 				r_error.error = Variant::CallError::CALL_ERROR_INVALID_ARGUMENT;
 				r_error.argument = 0;
@@ -1104,64 +1047,74 @@ void VisualScriptBuiltinFunc::exec_func(BuiltinFunc p_func, const Variant **p_in
 				return;
 
 			} else {
-
 				*r_return = Variant::construct(Variant::Type(type), p_inputs, 1, r_error);
 			}
 		} break;
 		case VisualScriptBuiltinFunc::TYPE_OF: {
-
 			*r_return = p_inputs[0]->get_type();
 
 		} break;
 		case VisualScriptBuiltinFunc::TYPE_EXISTS: {
-
 			*r_return = ClassDB::class_exists(*p_inputs[0]);
 
 		} break;
 		case VisualScriptBuiltinFunc::TEXT_CHAR: {
-
 			CharType result[2] = { *p_inputs[0], 0 };
 
 			*r_return = String(result);
 
 		} break;
-		case VisualScriptBuiltinFunc::TEXT_STR: {
+		case VisualScriptBuiltinFunc::TEXT_ORD: {
+			if (p_inputs[0]->get_type() != Variant::STRING) {
+				r_error.error = Variant::CallError::CALL_ERROR_INVALID_ARGUMENT;
+				r_error.argument = 0;
+				r_error.expected = Variant::STRING;
 
+				return;
+			}
+
+			String str = p_inputs[0]->operator String();
+
+			if (str.length() != 1) {
+				r_error.error = Variant::CallError::CALL_ERROR_INVALID_ARGUMENT;
+				r_error.argument = 0;
+				r_error.expected = Variant::STRING;
+				*r_return = "Expected a string of length 1 (a character).";
+
+				return;
+			}
+
+			*r_return = str.get(0);
+
+		} break;
+		case VisualScriptBuiltinFunc::TEXT_STR: {
 			String str = *p_inputs[0];
 
 			*r_return = str;
 
 		} break;
 		case VisualScriptBuiltinFunc::TEXT_PRINT: {
-
 			String str = *p_inputs[0];
 			print_line(str);
 
 		} break;
 
 		case VisualScriptBuiltinFunc::TEXT_PRINTERR: {
-
 			String str = *p_inputs[0];
-
-			//str+="\n";
 			print_error(str);
 
 		} break;
 		case VisualScriptBuiltinFunc::TEXT_PRINTRAW: {
 			String str = *p_inputs[0];
-
-			//str+="\n";
 			OS::get_singleton()->print("%s", str.utf8().get_data());
 
 		} break;
 		case VisualScriptBuiltinFunc::VAR_TO_STR: {
-
 			String vars;
 			VariantWriter::write_to_string(*p_inputs[0], vars);
 			*r_return = vars;
 		} break;
 		case VisualScriptBuiltinFunc::STR_TO_VAR: {
-
 			if (p_inputs[0]->get_type() != Variant::STRING) {
 				r_error.error = Variant::CallError::CALL_ERROR_INVALID_ARGUMENT;
 				r_error.argument = 0;
@@ -1187,10 +1140,16 @@ void VisualScriptBuiltinFunc::exec_func(BuiltinFunc p_func, const Variant **p_in
 
 		} break;
 		case VisualScriptBuiltinFunc::VAR_TO_BYTES: {
-
+			if (p_inputs[1]->get_type() != Variant::BOOL) {
+				r_error.error = Variant::CallError::CALL_ERROR_INVALID_ARGUMENT;
+				r_error.argument = 1;
+				r_error.expected = Variant::BOOL;
+				return;
+			}
 			PoolByteArray barr;
 			int len;
-			Error err = encode_variant(*p_inputs[0], NULL, len);
+			bool full_objects = *p_inputs[1];
+			Error err = encode_variant(*p_inputs[0], nullptr, len, full_objects);
 			if (err) {
 				r_error.error = Variant::CallError::CALL_ERROR_INVALID_ARGUMENT;
 				r_error.argument = 0;
@@ -1202,25 +1161,30 @@ void VisualScriptBuiltinFunc::exec_func(BuiltinFunc p_func, const Variant **p_in
 			barr.resize(len);
 			{
 				PoolByteArray::Write w = barr.write();
-				encode_variant(*p_inputs[0], w.ptr(), len);
+				encode_variant(*p_inputs[0], w.ptr(), len, full_objects);
 			}
 			*r_return = barr;
 		} break;
 		case VisualScriptBuiltinFunc::BYTES_TO_VAR: {
-
 			if (p_inputs[0]->get_type() != Variant::POOL_BYTE_ARRAY) {
 				r_error.error = Variant::CallError::CALL_ERROR_INVALID_ARGUMENT;
 				r_error.argument = 0;
 				r_error.expected = Variant::POOL_BYTE_ARRAY;
-
+				return;
+			}
+			if (p_inputs[1]->get_type() != Variant::BOOL) {
+				r_error.error = Variant::CallError::CALL_ERROR_INVALID_ARGUMENT;
+				r_error.argument = 1;
+				r_error.expected = Variant::BOOL;
 				return;
 			}
 
 			PoolByteArray varr = *p_inputs[0];
+			bool allow_objects = *p_inputs[1];
 			Variant ret;
 			{
 				PoolByteArray::Read r = varr.read();
-				Error err = decode_variant(ret, r.ptr(), varr.size(), NULL);
+				Error err = decode_variant(ret, r.ptr(), varr.size(), nullptr, allow_objects);
 				if (err != OK) {
 					r_error_str = RTR("Not enough bytes for decoding bytes, or invalid format.");
 					r_error.error = Variant::CallError::CALL_ERROR_INVALID_ARGUMENT;
@@ -1234,7 +1198,6 @@ void VisualScriptBuiltinFunc::exec_func(BuiltinFunc p_func, const Variant **p_in
 
 		} break;
 		case VisualScriptBuiltinFunc::COLORN: {
-
 			VALIDATE_ARG_NUM(1);
 
 			Color color = Color::named(*p_inputs[0]);
@@ -1243,7 +1206,8 @@ void VisualScriptBuiltinFunc::exec_func(BuiltinFunc p_func, const Variant **p_in
 			*r_return = String(color);
 
 		} break;
-		default: {}
+		default: {
+		}
 	}
 }
 
@@ -1259,14 +1223,12 @@ public:
 	//virtual bool get_output_port_unsequenced(int p_idx,Variant* r_value,Variant* p_working_mem,String &r_error) const { return true; }
 
 	virtual int step(const Variant **p_inputs, Variant **p_outputs, StartMode p_start_mode, Variant *p_working_mem, Variant::CallError &r_error, String &r_error_str) {
-
 		VisualScriptBuiltinFunc::exec_func(func, p_inputs, p_outputs[0], r_error, r_error_str);
 		return 0;
 	}
 };
 
 VisualScriptNodeInstance *VisualScriptBuiltinFunc::instance(VisualScriptInstance *p_instance) {
-
 	VisualScriptNodeInstanceBuiltinFunc *instance = memnew(VisualScriptNodeInstanceBuiltinFunc);
 	instance->node = this;
 	instance->instance = p_instance;
@@ -1275,16 +1237,15 @@ VisualScriptNodeInstance *VisualScriptBuiltinFunc::instance(VisualScriptInstance
 }
 
 void VisualScriptBuiltinFunc::_bind_methods() {
-
 	ClassDB::bind_method(D_METHOD("set_func", "which"), &VisualScriptBuiltinFunc::set_func);
 	ClassDB::bind_method(D_METHOD("get_func"), &VisualScriptBuiltinFunc::get_func);
 
 	String cc;
 
 	for (int i = 0; i < FUNC_MAX; i++) {
-
-		if (i > 0)
+		if (i > 0) {
 			cc += ",";
+		}
 		cc += func_name[i];
 	}
 	ADD_PROPERTY(PropertyInfo(Variant::INT, "function", PROPERTY_HINT_ENUM, cc), "set_func", "get_func");
@@ -1318,6 +1279,7 @@ void VisualScriptBuiltinFunc::_bind_methods() {
 	BIND_ENUM_CONSTANT(MATH_LERP);
 	BIND_ENUM_CONSTANT(MATH_INVERSE_LERP);
 	BIND_ENUM_CONSTANT(MATH_RANGE_LERP);
+	BIND_ENUM_CONSTANT(MATH_MOVE_TOWARD);
 	BIND_ENUM_CONSTANT(MATH_DECTIME);
 	BIND_ENUM_CONSTANT(MATH_RANDOMIZE);
 	BIND_ENUM_CONSTANT(MATH_RAND);
@@ -1352,28 +1314,28 @@ void VisualScriptBuiltinFunc::_bind_methods() {
 	BIND_ENUM_CONSTANT(VAR_TO_BYTES);
 	BIND_ENUM_CONSTANT(BYTES_TO_VAR);
 	BIND_ENUM_CONSTANT(COLORN);
+	BIND_ENUM_CONSTANT(MATH_SMOOTHSTEP);
+	BIND_ENUM_CONSTANT(MATH_POSMOD);
+	BIND_ENUM_CONSTANT(MATH_LERP_ANGLE);
+	BIND_ENUM_CONSTANT(TEXT_ORD);
 	BIND_ENUM_CONSTANT(FUNC_MAX);
 }
 
 VisualScriptBuiltinFunc::VisualScriptBuiltinFunc(VisualScriptBuiltinFunc::BuiltinFunc func) {
-
 	this->func = func;
 }
 
 VisualScriptBuiltinFunc::VisualScriptBuiltinFunc() {
-
 	func = MATH_SIN;
 }
 
 template <VisualScriptBuiltinFunc::BuiltinFunc func>
 static Ref<VisualScriptNode> create_builtin_func_node(const String &p_name) {
-
 	Ref<VisualScriptBuiltinFunc> node = memnew(VisualScriptBuiltinFunc(func));
 	return node;
 }
 
 void register_visual_script_builtin_func_node() {
-
 	VisualScriptLanguage::singleton->add_register_func("functions/built_in/sin", create_builtin_func_node<VisualScriptBuiltinFunc::MATH_SIN>);
 	VisualScriptLanguage::singleton->add_register_func("functions/built_in/cos", create_builtin_func_node<VisualScriptBuiltinFunc::MATH_COS>);
 	VisualScriptLanguage::singleton->add_register_func("functions/built_in/tan", create_builtin_func_node<VisualScriptBuiltinFunc::MATH_TAN>);
@@ -1389,6 +1351,7 @@ void register_visual_script_builtin_func_node() {
 	VisualScriptLanguage::singleton->add_register_func("functions/built_in/sqrt", create_builtin_func_node<VisualScriptBuiltinFunc::MATH_SQRT>);
 	VisualScriptLanguage::singleton->add_register_func("functions/built_in/fmod", create_builtin_func_node<VisualScriptBuiltinFunc::MATH_FMOD>);
 	VisualScriptLanguage::singleton->add_register_func("functions/built_in/fposmod", create_builtin_func_node<VisualScriptBuiltinFunc::MATH_FPOSMOD>);
+	VisualScriptLanguage::singleton->add_register_func("functions/built_in/posmod", create_builtin_func_node<VisualScriptBuiltinFunc::MATH_POSMOD>);
 	VisualScriptLanguage::singleton->add_register_func("functions/built_in/floor", create_builtin_func_node<VisualScriptBuiltinFunc::MATH_FLOOR>);
 	VisualScriptLanguage::singleton->add_register_func("functions/built_in/ceil", create_builtin_func_node<VisualScriptBuiltinFunc::MATH_CEIL>);
 	VisualScriptLanguage::singleton->add_register_func("functions/built_in/round", create_builtin_func_node<VisualScriptBuiltinFunc::MATH_ROUND>);
@@ -1404,8 +1367,11 @@ void register_visual_script_builtin_func_node() {
 	VisualScriptLanguage::singleton->add_register_func("functions/built_in/decimals", create_builtin_func_node<VisualScriptBuiltinFunc::MATH_DECIMALS>);
 	VisualScriptLanguage::singleton->add_register_func("functions/built_in/stepify", create_builtin_func_node<VisualScriptBuiltinFunc::MATH_STEPIFY>);
 	VisualScriptLanguage::singleton->add_register_func("functions/built_in/lerp", create_builtin_func_node<VisualScriptBuiltinFunc::MATH_LERP>);
+	VisualScriptLanguage::singleton->add_register_func("functions/built_in/lerp_angle", create_builtin_func_node<VisualScriptBuiltinFunc::MATH_LERP_ANGLE>);
 	VisualScriptLanguage::singleton->add_register_func("functions/built_in/inverse_lerp", create_builtin_func_node<VisualScriptBuiltinFunc::MATH_INVERSE_LERP>);
 	VisualScriptLanguage::singleton->add_register_func("functions/built_in/range_lerp", create_builtin_func_node<VisualScriptBuiltinFunc::MATH_RANGE_LERP>);
+	VisualScriptLanguage::singleton->add_register_func("functions/built_in/smoothstep", create_builtin_func_node<VisualScriptBuiltinFunc::MATH_SMOOTHSTEP>);
+	VisualScriptLanguage::singleton->add_register_func("functions/built_in/move_toward", create_builtin_func_node<VisualScriptBuiltinFunc::MATH_MOVE_TOWARD>);
 	VisualScriptLanguage::singleton->add_register_func("functions/built_in/dectime", create_builtin_func_node<VisualScriptBuiltinFunc::MATH_DECTIME>);
 	VisualScriptLanguage::singleton->add_register_func("functions/built_in/randomize", create_builtin_func_node<VisualScriptBuiltinFunc::MATH_RANDOMIZE>);
 	VisualScriptLanguage::singleton->add_register_func("functions/built_in/rand", create_builtin_func_node<VisualScriptBuiltinFunc::MATH_RAND>);
@@ -1434,6 +1400,7 @@ void register_visual_script_builtin_func_node() {
 	VisualScriptLanguage::singleton->add_register_func("functions/built_in/typeof", create_builtin_func_node<VisualScriptBuiltinFunc::TYPE_OF>);
 	VisualScriptLanguage::singleton->add_register_func("functions/built_in/type_exists", create_builtin_func_node<VisualScriptBuiltinFunc::TYPE_EXISTS>);
 	VisualScriptLanguage::singleton->add_register_func("functions/built_in/char", create_builtin_func_node<VisualScriptBuiltinFunc::TEXT_CHAR>);
+	VisualScriptLanguage::singleton->add_register_func("functions/built_in/ord", create_builtin_func_node<VisualScriptBuiltinFunc::TEXT_ORD>);
 	VisualScriptLanguage::singleton->add_register_func("functions/built_in/str", create_builtin_func_node<VisualScriptBuiltinFunc::TEXT_STR>);
 	VisualScriptLanguage::singleton->add_register_func("functions/built_in/print", create_builtin_func_node<VisualScriptBuiltinFunc::TEXT_PRINT>);
 	VisualScriptLanguage::singleton->add_register_func("functions/built_in/printerr", create_builtin_func_node<VisualScriptBuiltinFunc::TEXT_PRINTERR>);
