@@ -5,8 +5,8 @@
 /*                           GODOT ENGINE                                */
 /*                      https://godotengine.org                          */
 /*************************************************************************/
-/* Copyright (c) 2007-2018 Juan Linietsky, Ariel Manzur.                 */
-/* Copyright (c) 2014-2018 Godot Engine contributors (cf. AUTHORS.md)    */
+/* Copyright (c) 2007-2022 Juan Linietsky, Ariel Manzur.                 */
+/* Copyright (c) 2014-2022 Godot Engine contributors (cf. AUTHORS.md).   */
 /*                                                                       */
 /* Permission is hereby granted, free of charge, to any person obtaining */
 /* a copy of this software and associated documentation files (the       */
@@ -79,12 +79,11 @@ public:
 };
 
 void VehicleWheel::_notification(int p_what) {
-
 	if (p_what == NOTIFICATION_ENTER_TREE) {
-
 		VehicleBody *cb = Object::cast_to<VehicleBody>(get_parent());
-		if (!cb)
+		if (!cb) {
 			return;
+		}
 		body = cb;
 		local_xform = get_transform();
 		cb->wheels.push_back(this);
@@ -94,25 +93,28 @@ void VehicleWheel::_notification(int p_what) {
 		m_wheelAxleCS = get_transform().basis.get_axis(Vector3::AXIS_X).normalized();
 	}
 	if (p_what == NOTIFICATION_EXIT_TREE) {
-
 		VehicleBody *cb = Object::cast_to<VehicleBody>(get_parent());
-		if (!cb)
+		if (!cb) {
 			return;
+		}
 		cb->wheels.erase(this);
-		body = NULL;
+		body = nullptr;
 	}
 }
 
 String VehicleWheel::get_configuration_warning() const {
+	String warning = Spatial::get_configuration_warning();
 	if (!Object::cast_to<VehicleBody>(get_parent())) {
-		return TTR("VehicleWheel serves to provide a wheel system to a VehicleBody. Please use it as a child of a VehicleBody.");
+		if (warning != String()) {
+			warning += "\n\n";
+		}
+		warning += TTR("VehicleWheel serves to provide a wheel system to a VehicleBody. Please use it as a child of a VehicleBody.");
 	}
 
-	return String();
+	return warning;
 }
 
 void VehicleWheel::_update(PhysicsDirectBodyState *s) {
-
 	if (m_raycastInfo.m_isInContact)
 
 	{
@@ -121,7 +123,7 @@ void VehicleWheel::_update(PhysicsDirectBodyState *s) {
 		Vector3 relpos = m_raycastInfo.m_contactPointWS - s->get_transform().origin;
 
 		chassis_velocity_at_contactPoint = s->get_linear_velocity() +
-										   (s->get_angular_velocity()).cross(relpos); // * mPos);
+				(s->get_angular_velocity()).cross(relpos); // * mPos);
 
 		real_t projVel = m_raycastInfo.m_contactNormalWS.dot(chassis_velocity_at_contactPoint);
 		if (project >= real_t(-0.1)) {
@@ -145,77 +147,61 @@ void VehicleWheel::_update(PhysicsDirectBodyState *s) {
 }
 
 void VehicleWheel::set_radius(float p_radius) {
-
 	m_wheelRadius = p_radius;
 	update_gizmo();
 }
 
 float VehicleWheel::get_radius() const {
-
 	return m_wheelRadius;
 }
 
 void VehicleWheel::set_suspension_rest_length(float p_length) {
-
 	m_suspensionRestLength = p_length;
 	update_gizmo();
 }
 float VehicleWheel::get_suspension_rest_length() const {
-
 	return m_suspensionRestLength;
 }
 
 void VehicleWheel::set_suspension_travel(float p_length) {
-
 	m_maxSuspensionTravelCm = p_length / 0.01;
 }
 float VehicleWheel::get_suspension_travel() const {
-
 	return m_maxSuspensionTravelCm * 0.01;
 }
 
 void VehicleWheel::set_suspension_stiffness(float p_value) {
-
 	m_suspensionStiffness = p_value;
 }
 float VehicleWheel::get_suspension_stiffness() const {
-
 	return m_suspensionStiffness;
 }
 
 void VehicleWheel::set_suspension_max_force(float p_value) {
-
 	m_maxSuspensionForce = p_value;
 }
 float VehicleWheel::get_suspension_max_force() const {
-
 	return m_maxSuspensionForce;
 }
 
 void VehicleWheel::set_damping_compression(float p_value) {
-
 	m_wheelsDampingCompression = p_value;
 }
 float VehicleWheel::get_damping_compression() const {
-
 	return m_wheelsDampingCompression;
 }
 
 void VehicleWheel::set_damping_relaxation(float p_value) {
-
 	m_wheelsDampingRelaxation = p_value;
 }
 float VehicleWheel::get_damping_relaxation() const {
-
 	return m_wheelsDampingRelaxation;
 }
 
 void VehicleWheel::set_friction_slip(float p_value) {
-
 	m_frictionSlip = p_value;
 }
 float VehicleWheel::get_friction_slip() const {
-
 	return m_frictionSlip;
 }
 
@@ -232,7 +218,6 @@ bool VehicleWheel::is_in_contact() const {
 }
 
 void VehicleWheel::_bind_methods() {
-
 	ClassDB::bind_method(D_METHOD("set_radius", "length"), &VehicleWheel::set_radius);
 	ClassDB::bind_method(D_METHOD("get_radius"), &VehicleWheel::get_radius);
 
@@ -270,6 +255,22 @@ void VehicleWheel::_bind_methods() {
 
 	ClassDB::bind_method(D_METHOD("get_skidinfo"), &VehicleWheel::get_skidinfo);
 
+	ClassDB::bind_method(D_METHOD("get_rpm"), &VehicleWheel::get_rpm);
+
+	ClassDB::bind_method(D_METHOD("set_engine_force", "engine_force"), &VehicleWheel::set_engine_force);
+	ClassDB::bind_method(D_METHOD("get_engine_force"), &VehicleWheel::get_engine_force);
+
+	ClassDB::bind_method(D_METHOD("set_brake", "brake"), &VehicleWheel::set_brake);
+	ClassDB::bind_method(D_METHOD("get_brake"), &VehicleWheel::get_brake);
+
+	ClassDB::bind_method(D_METHOD("set_steering", "steering"), &VehicleWheel::set_steering);
+	ClassDB::bind_method(D_METHOD("get_steering"), &VehicleWheel::get_steering);
+
+	ADD_GROUP("Per-Wheel Motion", "");
+	ADD_PROPERTY(PropertyInfo(Variant::REAL, "engine_force", PROPERTY_HINT_RANGE, "-1024,1024.0,0.01,or_greater"), "set_engine_force", "get_engine_force");
+	ADD_PROPERTY(PropertyInfo(Variant::REAL, "brake", PROPERTY_HINT_RANGE, "0.0,1.0,0.01"), "set_brake", "get_brake");
+	ADD_PROPERTY(PropertyInfo(Variant::REAL, "steering", PROPERTY_HINT_RANGE, "-180,180.0,0.01"), "set_steering", "get_steering");
+	ADD_GROUP("VehicleBody Motion", "");
 	ADD_PROPERTY(PropertyInfo(Variant::BOOL, "use_as_traction"), "set_use_as_traction", "is_used_as_traction");
 	ADD_PROPERTY(PropertyInfo(Variant::BOOL, "use_as_steering"), "set_use_as_steering", "is_used_as_steering");
 	ADD_GROUP("Wheel", "wheel_");
@@ -286,38 +287,58 @@ void VehicleWheel::_bind_methods() {
 	ADD_PROPERTY(PropertyInfo(Variant::REAL, "damping_relaxation"), "set_damping_relaxation", "get_damping_relaxation");
 }
 
-void VehicleWheel::set_use_as_traction(bool p_enable) {
+void VehicleWheel::set_engine_force(float p_engine_force) {
+	m_engineForce = p_engine_force;
+}
 
+float VehicleWheel::get_engine_force() const {
+	return m_engineForce;
+}
+
+void VehicleWheel::set_brake(float p_brake) {
+	m_brake = p_brake;
+}
+float VehicleWheel::get_brake() const {
+	return m_brake;
+}
+
+void VehicleWheel::set_steering(float p_steering) {
+	m_steering = p_steering;
+}
+float VehicleWheel::get_steering() const {
+	return m_steering;
+}
+
+void VehicleWheel::set_use_as_traction(bool p_enable) {
 	engine_traction = p_enable;
 }
 
 bool VehicleWheel::is_used_as_traction() const {
-
 	return engine_traction;
 }
 
 void VehicleWheel::set_use_as_steering(bool p_enabled) {
-
 	steers = p_enabled;
 }
 
 bool VehicleWheel::is_used_as_steering() const {
-
 	return steers;
 }
 
 float VehicleWheel::get_skidinfo() const {
-
 	return m_skidInfo;
 }
 
-VehicleWheel::VehicleWheel() {
+float VehicleWheel::get_rpm() const {
+	return m_rpm;
+}
 
+VehicleWheel::VehicleWheel() {
 	steers = false;
 	engine_traction = false;
 
 	m_steering = real_t(0.);
-	//m_engineForce = real_t(0.);
+	m_engineForce = real_t(0.);
 	m_rotation = real_t(0.);
 	m_deltaRotation = real_t(0.);
 	m_brake = real_t(0.);
@@ -336,12 +357,12 @@ VehicleWheel::VehicleWheel() {
 	m_suspensionRelativeVelocity = 0;
 	m_clippedInvContactDotSuspension = 1.0;
 	m_raycastInfo.m_isInContact = false;
+	m_raycastInfo.m_suspensionLength = 0.0;
 
-	body = NULL;
+	body = nullptr;
 }
 
 void VehicleBody::_update_wheel_transform(VehicleWheel &wheel, PhysicsDirectBodyState *s) {
-
 	wheel.m_raycastInfo.m_isInContact = false;
 
 	Transform chassisTrans = s->get_transform();
@@ -358,7 +379,6 @@ void VehicleBody::_update_wheel_transform(VehicleWheel &wheel, PhysicsDirectBody
 }
 
 void VehicleBody::_update_wheel(int p_idx, PhysicsDirectBodyState *s) {
-
 	VehicleWheel &wheel = *wheels[p_idx];
 	_update_wheel_transform(wheel, s);
 
@@ -366,22 +386,11 @@ void VehicleBody::_update_wheel(int p_idx, PhysicsDirectBodyState *s) {
 	const Vector3 &right = wheel.m_raycastInfo.m_wheelAxleWS;
 	Vector3 fwd = up.cross(right);
 	fwd = fwd.normalized();
-	//up = right.cross(fwd);
-	//up.normalize();
 
-	//rotate around steering over de wheelAxleWS
-	real_t steering = wheel.steers ? m_steeringValue : 0.0;
-	//print_line(itos(p_idx)+": "+rtos(steering));
-
-	Basis steeringMat(up, steering);
+	Basis steeringMat(up, wheel.m_steering);
 
 	Basis rotatingMat(right, wheel.m_rotation);
 
-	/*
-	if (p_idx==1)
-		print_line("steeringMat " +steeringMat);
-	*/
-
 	Basis basis2(
 			right[0], up[0], fwd[0],
 			right[1], up[1], fwd[1],
@@ -394,7 +403,6 @@ void VehicleBody::_update_wheel(int p_idx, PhysicsDirectBodyState *s) {
 }
 
 real_t VehicleBody::_ray_cast(int p_idx, PhysicsDirectBodyState *s) {
-
 	VehicleWheel &wheel = *wheels[p_idx];
 
 	_update_wheel_transform(wheel, s);
@@ -415,20 +423,19 @@ real_t VehicleBody::_ray_cast(int p_idx, PhysicsDirectBodyState *s) {
 
 	PhysicsDirectSpaceState *ss = s->get_space_state();
 
-	bool col = ss->intersect_ray(source, target, rr, exclude);
+	bool col = ss->intersect_ray(source, target, rr, exclude, get_collision_mask());
 
-	wheel.m_raycastInfo.m_groundObject = 0;
+	wheel.m_raycastInfo.m_groundObject = nullptr;
 
 	if (col) {
-		//print_line("WHEEL "+itos(p_idx)+" FROM "+source+" TO: "+target);
-		//print_line("WHEEL "+itos(p_idx)+" COLLIDE? "+itos(col));
 		param = source.distance_to(rr.position) / source.distance_to(target);
 		depth = raylen * param;
 		wheel.m_raycastInfo.m_contactNormalWS = rr.normal;
 
 		wheel.m_raycastInfo.m_isInContact = true;
-		if (rr.collider)
+		if (rr.collider) {
 			wheel.m_raycastInfo.m_groundObject = Object::cast_to<PhysicsBody>(rr.collider);
+		}
 
 		real_t hitDistance = param * raylen;
 		wheel.m_raycastInfo.m_suspensionLength = hitDistance - wheel.m_wheelRadius;
@@ -453,7 +460,7 @@ real_t VehicleBody::_ray_cast(int p_idx, PhysicsDirectBodyState *s) {
 		//chassis_velocity_at_contactPoint = getRigidBody()->getVelocityInLocalPoint(relpos);
 
 		chassis_velocity_at_contactPoint = s->get_linear_velocity() +
-										   (s->get_angular_velocity()).cross(wheel.m_raycastInfo.m_contactPointWS - s->get_transform().origin); // * mPos);
+				(s->get_angular_velocity()).cross(wheel.m_raycastInfo.m_contactPointWS - s->get_transform().origin); // * mPos);
 
 		real_t projVel = wheel.m_raycastInfo.m_contactNormalWS.dot(chassis_velocity_at_contactPoint);
 
@@ -479,7 +486,6 @@ real_t VehicleBody::_ray_cast(int p_idx, PhysicsDirectBodyState *s) {
 }
 
 void VehicleBody::_update_suspension(PhysicsDirectBodyState *s) {
-
 	real_t chassisMass = mass;
 
 	for (int w_it = 0; w_it < wheels.size(); w_it++) {
@@ -524,8 +530,7 @@ void VehicleBody::_update_suspension(PhysicsDirectBodyState *s) {
 
 //bilateral constraint between two dynamic objects
 void VehicleBody::_resolve_single_bilateral(PhysicsDirectBodyState *s, const Vector3 &pos1,
-		PhysicsBody *body2, const Vector3 &pos2, const Vector3 &normal, real_t &impulse, real_t p_rollInfluence) {
-
+		PhysicsBody *body2, const Vector3 &pos2, const Vector3 &normal, real_t &impulse, const real_t p_rollInfluence) {
 	real_t normalLenSqr = normal.length_squared();
 	//ERR_FAIL_COND( normalLenSqr < real_t(1.1));
 
@@ -536,15 +541,17 @@ void VehicleBody::_resolve_single_bilateral(PhysicsDirectBodyState *s, const Vec
 
 	Vector3 rel_pos1 = pos1 - s->get_transform().origin;
 	Vector3 rel_pos2;
-	if (body2)
+	if (body2) {
 		rel_pos2 = pos2 - body2->get_global_transform().origin;
+	}
 	//this jacobian entry could be re-used for all iterations
 
 	Vector3 vel1 = s->get_linear_velocity() + (s->get_angular_velocity()).cross(rel_pos1); // * mPos);
 	Vector3 vel2;
 
-	if (body2)
+	if (body2) {
 		vel2 = body2->get_linear_velocity() + body2->get_angular_velocity().cross(rel_pos2);
+	}
 
 	Vector3 vel = vel1 - vel2;
 
@@ -572,7 +579,7 @@ void VehicleBody::_resolve_single_bilateral(PhysicsDirectBodyState *s, const Vec
 			b2invmass);
 
 	// FIXME: rel_vel assignment here is overwritten by the following assignment.
-	// What seemes to be intended in the next next assignment is: rel_vel = normal.dot(rel_vel);
+	// What seems to be intended in the next next assignment is: rel_vel = normal.dot(rel_vel);
 	// Investigate why.
 	real_t rel_vel = jac.getRelativeVelocity(
 			s->get_linear_velocity(),
@@ -588,7 +595,7 @@ void VehicleBody::_resolve_single_bilateral(PhysicsDirectBodyState *s, const Vec
 	if (p_rollInfluence > 0.0) {
 		// !BAS! But seeing we apply this frame by frame, makes more sense to me to make this time based
 		// keeping in mind our anti roll factor if it is set
-		contactDamping = s->get_step() / p_rollInfluence;
+		contactDamping = MIN(contactDamping, s->get_step() / p_rollInfluence);
 	}
 
 #define ONLY_USE_LINEAR_MASS
@@ -633,15 +640,15 @@ VehicleBody::btVehicleWheelContactPoint::btVehicleWheelContactPoint(PhysicsDirec
 }
 
 real_t VehicleBody::_calc_rolling_friction(btVehicleWheelContactPoint &contactPoint) {
-
 	real_t j1 = 0.f;
 
 	const Vector3 &contactPosWorld = contactPoint.m_frictionPositionWorld;
 
 	Vector3 rel_pos1 = contactPosWorld - contactPoint.m_s->get_transform().origin;
 	Vector3 rel_pos2;
-	if (contactPoint.m_body1)
+	if (contactPoint.m_body1) {
 		rel_pos2 = contactPosWorld - contactPoint.m_body1->get_global_transform().origin;
+	}
 
 	real_t maxImpulse = contactPoint.m_maxImpulse;
 
@@ -664,56 +671,48 @@ real_t VehicleBody::_calc_rolling_friction(btVehicleWheelContactPoint &contactPo
 
 static const real_t sideFrictionStiffness2 = real_t(1.0);
 void VehicleBody::_update_friction(PhysicsDirectBodyState *s) {
-
 	//calculate the impulse, so that the wheels don't move sidewards
 	int numWheel = wheels.size();
-	if (!numWheel)
+	if (!numWheel) {
 		return;
+	}
 
 	m_forwardWS.resize(numWheel);
 	m_axle.resize(numWheel);
 	m_forwardImpulse.resize(numWheel);
 	m_sideImpulse.resize(numWheel);
 
-	int numWheelsOnGround = 0;
-
 	//collapse all those loops into one!
 	for (int i = 0; i < wheels.size(); i++) {
-		VehicleWheel &wheelInfo = *wheels[i];
-		if (wheelInfo.m_raycastInfo.m_isInContact)
-			numWheelsOnGround++;
-		m_sideImpulse[i] = real_t(0.);
-		m_forwardImpulse[i] = real_t(0.);
+		m_sideImpulse.write[i] = real_t(0.);
+		m_forwardImpulse.write[i] = real_t(0.);
 	}
 
 	{
-
 		for (int i = 0; i < wheels.size(); i++) {
-
 			VehicleWheel &wheelInfo = *wheels[i];
 
 			if (wheelInfo.m_raycastInfo.m_isInContact) {
-
 				//const btTransform& wheelTrans = getWheelTransformWS( i );
 
 				Basis wheelBasis0 = wheelInfo.m_worldTransform.basis; //get_global_transform().basis;
 
-				m_axle[i] = wheelBasis0.get_axis(Vector3::AXIS_X);
+				m_axle.write[i] = wheelBasis0.get_axis(Vector3::AXIS_X);
 				//m_axle[i] = wheelInfo.m_raycastInfo.m_wheelAxleWS;
 
 				const Vector3 &surfNormalWS = wheelInfo.m_raycastInfo.m_contactNormalWS;
 				real_t proj = m_axle[i].dot(surfNormalWS);
-				m_axle[i] -= surfNormalWS * proj;
-				m_axle[i] = m_axle[i].normalized();
+				m_axle.write[i] -= surfNormalWS * proj;
+				m_axle.write[i] = m_axle[i].normalized();
 
-				m_forwardWS[i] = surfNormalWS.cross(m_axle[i]);
-				m_forwardWS[i].normalize();
+				m_forwardWS.write[i] = surfNormalWS.cross(m_axle[i]);
+				m_forwardWS.write[i].normalize();
 
 				_resolve_single_bilateral(s, wheelInfo.m_raycastInfo.m_contactPointWS,
 						wheelInfo.m_raycastInfo.m_groundObject, wheelInfo.m_raycastInfo.m_contactPointWS,
-						m_axle[i], m_sideImpulse[i], wheelInfo.m_rollInfluence);
+						m_axle[i], m_sideImpulse.write[i], wheelInfo.m_rollInfluence);
 
-				m_sideImpulse[i] *= sideFrictionStiffness2;
+				m_sideImpulse.write[i] *= sideFrictionStiffness2;
 			}
 		}
 	}
@@ -731,12 +730,11 @@ void VehicleBody::_update_friction(PhysicsDirectBodyState *s) {
 			real_t rollingFriction = 0.f;
 
 			if (wheelInfo.m_raycastInfo.m_isInContact) {
-				if (engine_force != 0.f) {
-					rollingFriction = -engine_force * s->get_step();
+				if (wheelInfo.m_engineForce != 0.f) {
+					rollingFriction = -wheelInfo.m_engineForce * s->get_step();
 				} else {
 					real_t defaultRollingFrictionImpulse = 0.f;
-					float cbrake = MAX(wheelInfo.m_brake, brake);
-					real_t maxImpulse = cbrake ? cbrake : defaultRollingFrictionImpulse;
+					real_t maxImpulse = wheelInfo.m_brake ? wheelInfo.m_brake : defaultRollingFrictionImpulse;
 					btVehicleWheelContactPoint contactPt(s, wheelInfo.m_raycastInfo.m_groundObject, wheelInfo.m_raycastInfo.m_contactPointWS, m_forwardWS[wheel], maxImpulse);
 					rollingFriction = _calc_rolling_friction(contactPt);
 				}
@@ -744,7 +742,7 @@ void VehicleBody::_update_friction(PhysicsDirectBodyState *s) {
 
 			//switch between active rolling (throttle), braking and non-active rolling friction (no throttle/break)
 
-			m_forwardImpulse[wheel] = real_t(0.);
+			m_forwardImpulse.write[wheel] = real_t(0.);
 			wheelInfo.m_skidInfo = real_t(1.);
 
 			if (wheelInfo.m_raycastInfo.m_isInContact) {
@@ -755,7 +753,7 @@ void VehicleBody::_update_friction(PhysicsDirectBodyState *s) {
 
 				real_t maximpSquared = maximp * maximpSide;
 
-				m_forwardImpulse[wheel] = rollingFriction; //wheelInfo.m_engineForce* timeStep;
+				m_forwardImpulse.write[wheel] = rollingFriction; //wheelInfo.m_engineForce* timeStep;
 
 				real_t x = (m_forwardImpulse[wheel]) * fwdFactor;
 				real_t y = (m_sideImpulse[wheel]) * sideFactor;
@@ -777,8 +775,8 @@ void VehicleBody::_update_friction(PhysicsDirectBodyState *s) {
 		for (int wheel = 0; wheel < wheels.size(); wheel++) {
 			if (m_sideImpulse[wheel] != real_t(0.)) {
 				if (wheels[wheel]->m_skidInfo < real_t(1.)) {
-					m_forwardImpulse[wheel] *= wheels[wheel]->m_skidInfo;
-					m_sideImpulse[wheel] *= wheels[wheel]->m_skidInfo;
+					m_forwardImpulse.write[wheel] *= wheels[wheel]->m_skidInfo;
+					m_sideImpulse.write[wheel] *= wheels[wheel]->m_skidInfo;
 				}
 			}
 		}
@@ -790,7 +788,7 @@ void VehicleBody::_update_friction(PhysicsDirectBodyState *s) {
 			VehicleWheel &wheelInfo = *wheels[wheel];
 
 			Vector3 rel_pos = wheelInfo.m_raycastInfo.m_contactPointWS -
-							  s->get_transform().origin;
+					s->get_transform().origin;
 
 			if (m_forwardImpulse[wheel] != real_t(0.)) {
 				s->apply_impulse(rel_pos, m_forwardWS[wheel] * (m_forwardImpulse[wheel]));
@@ -822,20 +820,18 @@ void VehicleBody::_update_friction(PhysicsDirectBodyState *s) {
 }
 
 void VehicleBody::_direct_state_changed(Object *p_state) {
-
 	RigidBody::_direct_state_changed(p_state);
 
 	state = Object::cast_to<PhysicsDirectBodyState>(p_state);
+	ERR_FAIL_COND_MSG(!state, "Method '_direct_state_changed' must receive a valid PhysicsDirectBodyState object as argument");
 
 	float step = state->get_step();
 
 	for (int i = 0; i < wheels.size(); i++) {
-
 		_update_wheel(i, state);
 	}
 
 	for (int i = 0; i < wheels.size(); i++) {
-
 		_ray_cast(i, state);
 		wheels[i]->set_transform(state->get_transform().inverse() * wheels[i]->m_worldTransform);
 	}
@@ -843,7 +839,6 @@ void VehicleBody::_direct_state_changed(Object *p_state) {
 	_update_suspension(state);
 
 	for (int i = 0; i < wheels.size(); i++) {
-
 		//apply suspension force
 		VehicleWheel &wheel = *wheels[i];
 
@@ -880,48 +875,56 @@ void VehicleBody::_direct_state_changed(Object *p_state) {
 			real_t proj2 = fwd.dot(vel);
 
 			wheel.m_deltaRotation = (proj2 * step) / (wheel.m_wheelRadius);
-			wheel.m_rotation += wheel.m_deltaRotation;
-
-		} else {
-			wheel.m_rotation += wheel.m_deltaRotation;
 		}
 
+		wheel.m_rotation += wheel.m_deltaRotation;
+		wheel.m_rpm = ((wheel.m_deltaRotation / step) * 60) / Math_TAU;
+
 		wheel.m_deltaRotation *= real_t(0.99); //damping of rotation when not in contact
 	}
 
-	state = NULL;
+	state = nullptr;
 }
 
 void VehicleBody::set_engine_force(float p_engine_force) {
-
 	engine_force = p_engine_force;
+	for (int i = 0; i < wheels.size(); i++) {
+		VehicleWheel &wheelInfo = *wheels[i];
+		if (wheelInfo.engine_traction) {
+			wheelInfo.m_engineForce = p_engine_force;
+		}
+	}
 }
 
 float VehicleBody::get_engine_force() const {
-
 	return engine_force;
 }
 
 void VehicleBody::set_brake(float p_brake) {
-
 	brake = p_brake;
+	for (int i = 0; i < wheels.size(); i++) {
+		VehicleWheel &wheelInfo = *wheels[i];
+		wheelInfo.m_brake = p_brake;
+	}
 }
 float VehicleBody::get_brake() const {
-
 	return brake;
 }
 
 void VehicleBody::set_steering(float p_steering) {
-
 	m_steeringValue = p_steering;
+	for (int i = 0; i < wheels.size(); i++) {
+		VehicleWheel &wheelInfo = *wheels[i];
+		if (wheelInfo.steers) {
+			wheelInfo.m_steering = p_steering;
+		}
+	}
 }
 float VehicleBody::get_steering() const {
-
 	return m_steeringValue;
 }
 
 void VehicleBody::_bind_methods() {
-
 	ClassDB::bind_method(D_METHOD("set_engine_force", "engine_force"), &VehicleBody::set_engine_force);
 	ClassDB::bind_method(D_METHOD("get_engine_force"), &VehicleBody::get_engine_force);
 
@@ -932,14 +935,12 @@ void VehicleBody::_bind_methods() {
 	ClassDB::bind_method(D_METHOD("get_steering"), &VehicleBody::get_steering);
 
 	ADD_GROUP("Motion", "");
-	ADD_PROPERTY(PropertyInfo(Variant::REAL, "engine_force", PROPERTY_HINT_RANGE, "0.00,1024.0,0.01"), "set_engine_force", "get_engine_force");
+	ADD_PROPERTY(PropertyInfo(Variant::REAL, "engine_force", PROPERTY_HINT_RANGE, "-1024,1024.0,0.01,or_greater"), "set_engine_force", "get_engine_force");
 	ADD_PROPERTY(PropertyInfo(Variant::REAL, "brake", PROPERTY_HINT_RANGE, "0.0,1.0,0.01"), "set_brake", "get_brake");
 	ADD_PROPERTY(PropertyInfo(Variant::REAL, "steering", PROPERTY_HINT_RANGE, "-180,180.0,0.01"), "set_steering", "get_steering");
 }
 
-VehicleBody::VehicleBody() :
-		RigidBody() {
-
+VehicleBody::VehicleBody() {
 	m_pitchControl = 0;
 	m_currentVehicleSpeedKmHour = real_t(0.);
 	m_steeringValue = real_t(0.);
@@ -947,9 +948,7 @@ VehicleBody::VehicleBody() :
 	engine_force = 0;
 	brake = 0;
 
-	friction = 1;
-
-	state = NULL;
+	state = nullptr;
 	ccd = false;
 
 	exclude.insert(get_rid());
