@@ -5,8 +5,8 @@
 /*                           GODOT ENGINE                                */
 /*                      https://godotengine.org                          */
 /*************************************************************************/
-/* Copyright (c) 2007-2018 Juan Linietsky, Ariel Manzur.                 */
-/* Copyright (c) 2014-2018 Godot Engine contributors (cf. AUTHORS.md)    */
+/* Copyright (c) 2007-2022 Juan Linietsky, Ariel Manzur.                 */
+/* Copyright (c) 2014-2022 Godot Engine contributors (cf. AUTHORS.md).   */
 /*                                                                       */
 /* Permission is hereby granted, free of charge, to any person obtaining */
 /* a copy of this software and associated documentation files (the       */
@@ -57,7 +57,6 @@ public:
 };
 
 class Shape2DSW : public RID_Data {
-
 	RID self;
 	Rect2 aabb;
 	bool configured;
@@ -100,13 +99,12 @@ public:
 	const Map<ShapeOwner2DSW *, int> &get_owners() const;
 
 	_FORCE_INLINE_ void get_supports_transformed_cast(const Vector2 &p_cast, const Vector2 &p_normal, const Transform2D &p_xform, Vector2 *r_supports, int &r_amount) const {
-
 		get_supports(p_xform.basis_xform_inv(p_normal).normalized(), r_supports, r_amount);
-		for (int i = 0; i < r_amount; i++)
+		for (int i = 0; i < r_amount; i++) {
 			r_supports[i] = p_xform.xform(r_supports[i]);
+		}
 
 		if (r_amount == 1) {
-
 			if (Math::abs(p_normal.dot(p_cast.normalized())) < (1.0 - _SEGMENT_IS_VALID_SUPPORT_THRESHOLD)) {
 				//make line because they are parallel
 				r_amount = 2;
@@ -117,7 +115,6 @@ public:
 			}
 
 		} else {
-
 			if (Math::abs(p_normal.dot(p_cast.normalized())) < (1.0 - _SEGMENT_IS_VALID_SUPPORT_THRESHOLD)) {
 				//optimize line and make it larger because they are parallel
 				if ((r_supports[1] - r_supports[0]).dot(p_cast) > 0) {
@@ -145,7 +142,6 @@ public:
 		project_range_cast(p_cast, p_normal, p_transform, r_min, r_max);                                                                                         \
 	}                                                                                                                                                            \
 	_FORCE_INLINE_ void project_range_cast(const Vector2 &p_cast, const Vector2 &p_normal, const Transform2D &p_transform, real_t &r_min, real_t &r_max) const { \
-                                                                                                                                                                 \
 		real_t mina, maxa;                                                                                                                                       \
 		real_t minb, maxb;                                                                                                                                       \
 		Transform2D ofsb = p_transform;                                                                                                                          \
@@ -157,7 +153,6 @@ public:
 	}
 
 class LineShape2DSW : public Shape2DSW {
-
 	Vector2 normal;
 	real_t d;
 
@@ -195,11 +190,12 @@ public:
 };
 
 class RayShape2DSW : public Shape2DSW {
-
 	real_t length;
+	bool slips_on_slope;
 
 public:
 	_FORCE_INLINE_ real_t get_length() const { return length; }
+	_FORCE_INLINE_ bool get_slips_on_slope() const { return slips_on_slope; }
 
 	virtual Physics2DServer::ShapeType get_type() const { return Physics2DServer::SHAPE_RAY; }
 
@@ -218,7 +214,6 @@ public:
 		r_max = p_normal.dot(p_transform.get_origin());
 		r_min = p_normal.dot(p_transform.xform(Vector2(0, length)));
 		if (r_max < r_min) {
-
 			SWAP(r_max, r_min);
 		}
 	}
@@ -230,7 +225,6 @@ public:
 };
 
 class SegmentShape2DSW : public Shape2DSW {
-
 	Vector2 a;
 	Vector2 b;
 	Vector2 n;
@@ -243,7 +237,6 @@ public:
 	virtual Physics2DServer::ShapeType get_type() const { return Physics2DServer::SHAPE_SEGMENT; }
 
 	_FORCE_INLINE_ Vector2 get_xformed_normal(const Transform2D &p_xform) const {
-
 		return (p_xform.xform(b) - p_xform.xform(a)).normalized().tangent();
 	}
 	virtual void project_rangev(const Vector2 &p_normal, const Transform2D &p_transform, real_t &r_min, real_t &r_max) const { project_range(p_normal, p_transform, r_min, r_max); }
@@ -261,7 +254,6 @@ public:
 		r_max = p_normal.dot(p_transform.xform(a));
 		r_min = p_normal.dot(p_transform.xform(b));
 		if (r_max < r_min) {
-
 			SWAP(r_max, r_min);
 		}
 	}
@@ -277,7 +269,6 @@ public:
 };
 
 class CircleShape2DSW : public Shape2DSW {
-
 	real_t radius;
 
 public:
@@ -311,7 +302,6 @@ public:
 };
 
 class RectangleShape2DSW : public Shape2DSW {
-
 	Vector2 half_extents;
 
 public:
@@ -334,18 +324,18 @@ public:
 		r_max = -1e20;
 		r_min = 1e20;
 		for (int i = 0; i < 4; i++) {
-
 			real_t d = p_normal.dot(p_transform.xform(Vector2(((i & 1) * 2 - 1) * half_extents.x, ((i >> 1) * 2 - 1) * half_extents.y)));
 
-			if (d > r_max)
+			if (d > r_max) {
 				r_max = d;
-			if (d < r_min)
+			}
+			if (d < r_min) {
 				r_min = d;
+			}
 		}
 	}
 
 	_FORCE_INLINE_ Vector2 get_circle_axis(const Transform2D &p_xform, const Transform2D &p_xform_inv, const Vector2 &p_circle) const {
-
 		Vector2 local_v = p_xform_inv.xform(p_circle);
 
 		Vector2 he(
@@ -356,7 +346,6 @@ public:
 	}
 
 	_FORCE_INLINE_ Vector2 get_box_axis(const Transform2D &p_xform, const Transform2D &p_xform_inv, const RectangleShape2DSW *p_B, const Transform2D &p_B_xform, const Transform2D &p_B_xform_inv) const {
-
 		Vector2 a, b;
 
 		{
@@ -385,7 +374,6 @@ public:
 };
 
 class CapsuleShape2DSW : public Shape2DSW {
-
 	real_t radius;
 	real_t height;
 
@@ -417,7 +405,6 @@ public:
 		r_min = p_normal.dot(p_transform.xform(-n));
 
 		if (r_max < r_min) {
-
 			SWAP(r_max, r_min);
 		}
 
@@ -428,9 +415,7 @@ public:
 };
 
 class ConvexPolygonShape2DSW : public Shape2DSW {
-
 	struct Point {
-
 		Vector2 pos;
 		Vector2 normal; //normal to next segment
 	};
@@ -443,7 +428,6 @@ public:
 	_FORCE_INLINE_ const Vector2 &get_point(int p_idx) const { return points[p_idx].pos; }
 	_FORCE_INLINE_ const Vector2 &get_segment_normal(int p_idx) const { return points[p_idx].normal; }
 	_FORCE_INLINE_ Vector2 get_xformed_segment_normal(const Transform2D &p_xform, int p_idx) const {
-
 		Vector2 a = points[p_idx].pos;
 		p_idx++;
 		Vector2 b = points[p_idx == point_count ? 0 : p_idx].pos;
@@ -463,16 +447,20 @@ public:
 	virtual Variant get_data() const;
 
 	_FORCE_INLINE_ void project_range(const Vector2 &p_normal, const Transform2D &p_transform, real_t &r_min, real_t &r_max) const {
-		// no matter the angle, the box is mirrored anyway
+		if (!points || point_count <= 0) {
+			r_min = r_max = 0;
+			return;
+		}
 
 		r_min = r_max = p_normal.dot(p_transform.xform(points[0].pos));
 		for (int i = 1; i < point_count; i++) {
-
 			real_t d = p_normal.dot(p_transform.xform(points[i].pos));
-			if (d > r_max)
+			if (d > r_max) {
 				r_max = d;
-			if (d < r_min)
+			}
+			if (d < r_min) {
 				r_min = d;
+			}
 		}
 	}
 
@@ -483,18 +471,16 @@ public:
 };
 
 class ConcaveShape2DSW : public Shape2DSW {
-
 public:
-	virtual bool is_concave() const { return true; }
-	typedef void (*Callback)(void *p_userdata, Shape2DSW *p_convex);
+	// Returns true to stop the query.
+	typedef bool (*QueryCallback)(void *p_userdata, Shape2DSW *p_convex);
 
-	virtual void cull(const Rect2 &p_local_aabb, Callback p_callback, void *p_userdata) const = 0;
+	virtual bool is_concave() const { return true; }
+	virtual void cull(const Rect2 &p_local_aabb, QueryCallback p_callback, void *p_userdata) const = 0;
 };
 
 class ConcavePolygonShape2DSW : public ConcaveShape2DSW {
-
 	struct Segment {
-
 		int points[2];
 	};
 
@@ -502,7 +488,6 @@ class ConcavePolygonShape2DSW : public ConcaveShape2DSW {
 	Vector<Point2> points;
 
 	struct BVH {
-
 		Rect2 aabb;
 		int left, right;
 	};
@@ -511,17 +496,13 @@ class ConcavePolygonShape2DSW : public ConcaveShape2DSW {
 	int bvh_depth;
 
 	struct BVH_CompareX {
-
 		_FORCE_INLINE_ bool operator()(const BVH &a, const BVH &b) const {
-
 			return (a.aabb.position.x + a.aabb.size.x * 0.5) < (b.aabb.position.x + b.aabb.size.x * 0.5);
 		}
 	};
 
 	struct BVH_CompareY {
-
 		_FORCE_INLINE_ bool operator()(const BVH &a, const BVH &b) const {
-
 			return (a.aabb.position.y + a.aabb.size.y * 0.5) < (b.aabb.position.y + b.aabb.size.y * 0.5);
 		}
 	};
@@ -545,7 +526,7 @@ public:
 	virtual void set_data(const Variant &p_data);
 	virtual Variant get_data() const;
 
-	virtual void cull(const Rect2 &p_local_aabb, Callback p_callback, void *p_userdata) const;
+	virtual void cull(const Rect2 &p_local_aabb, QueryCallback p_callback, void *p_userdata) const;
 
 	DEFAULT_PROJECT_RANGE_CAST
 };
